import json
import re
import argparse
from pathlib import Path
from typing import Union, Set
from typing import List, Tuple, Dict

'''
This python file parses/converts a simple language file format into a JSON file
The (simple) language file format is a key-value mapping with some additional features (like multi-line strings)
The .slf files can be found in the `lang/` folder
This parser could be much simpler, but it is written to handle edge cases and provide good error messages

The reason that the .slf files exist in the first place is to provide a way for non-programmers to edit the language, and json is not very user-friendly

For a description of the format see below (_HELP_MESSAGE) or run 'python3 language_parser.py --help'
'''

_COMMENT_START = "//"
_WHITESPACE_CHARS = " \t\r\n"
_ONLY_SPACE_PATTERN = re.compile(" *")
_INHERIT_START = "INHERITS"
_REFERENCE_START = "#"
_LINE_QUOTE_INDICATOR = '"'
_BLOCK_QUOTE_INDICATOR = '"""'
_ESCAPED_STRING_INDICATOR = "E"
_ESCAPE_CHARACTER = "\\"
_ESCAPE_CHARACTER_MAPPINGS = {
    "\\": "\\",
    "\"": "\"",
    "t": "\t",
    "r": "\r",
    "n": "\n"
}
_ESCAPE_CHARACTER_READABLE_LIST = ', '.join([f"'{x}'" for x in _ESCAPE_CHARACTER_MAPPINGS.keys()])

_LANG_KEY_REGEX = "[a-z0-9_.]+"
_LANG_KEY_PATTERN = re.compile(_LANG_KEY_REGEX)

_EXTENDS_FILE_REGEX = "[a-z0-9]+"
_EXTENDS_FILE_PATTERN = re.compile(_EXTENDS_FILE_REGEX) 

_LANGUAGE_FILE_ENDING = ".slf"


_HELP_MESSAGE = f"""
This script parses Simple Language Files.
Simple Language Files are a mapping of keys to strings.
Keys must match the following pattern/regex: {_LANG_KEY_REGEX}

A Simple Language File is a list of entries on separate lines (Note that whitespace between elements is mostly ignored):
- An empty line is ignored
- A line starting with {_COMMENT_START} is ignored
- A line of the format '{_INHERIT_START} lang_name' specifies the parent language (see below)
- All other lines must start with a key, followed by an equal sign '=', followed by a value
    Possible values are:
  - '{_REFERENCE_START}' and a language key. This is a reference to another key, whose value will be used for this entry
  - A string value inbetween two '{_LINE_QUOTE_INDICATOR}' (single line quote). The key is than mapped to this value.
  - A '{_BLOCK_QUOTE_INDICATOR}' to indicate a block quote. The value for this key are all following lines until a \
closing '{_BLOCK_QUOTE_INDICATOR}'. The closing '{_BLOCK_QUOTE_INDICATOR}' must be alone on a line, that is \
there are no characters before (except spaces) or after on that line. Equally, there may be no characters after the opening \
'{_BLOCK_QUOTE_INDICATOR}' on the same line.
    - If there are spaces before the closing '{_BLOCK_QUOTE_INDICATOR}', all lines must start with this amount of spaces.
      These spaces are not included in the final text
    - If a line ends with a '\\', the following line break is not included in the final text.
  - Quotes (single or block) are parsed as an escaped string if prefixed with a '{_ESCAPED_STRING_INDICATOR}'. The \
escape character is '{_ESCAPE_CHARACTER}' and the following characters may be escaped: \
{_ESCAPE_CHARACTER_READABLE_LIST}

Short example file:

block_quote.name = "A quote"
block_quote.description = \"""
This is a block quote.
You can have as many lines as you want:


Everything is kept as is
\"""
single_quote.name = #block_quote.name
single_quote.description = E"Above me is a reference to the key \\"block_quote.name\\" <- and these are escaped quotes."

escaped.single = "This is \\n a line break in a single quote"
escaped = E\"""
I am an escaping block quote.
You can even write \\\""" here which will be displayed as three "
You can also \\
use a \\ at the end \\
of a line to not display \\
the line break
\"""
tabbed = \"""
    I have spaces on the beginning of each line.
    <- These spaces will not be included in the final text
    \"""

Inheritance:

A language file may also inherit all entries from another language and only override the values it needs to change.
To specify the parent language, write: '{_INHERIT_START} language_name' on a line. 'language_name' specifies the name of
the language to inherit. A language's name must match the key pattern {_LANG_KEY_REGEX}. The parent language's file must
have the name 'language_name{_LANGUAGE_FILE_ENDING}' and be next to the child language's file.

"""


class ParseError(Exception):
    
    def __init__(self, line_index: Union[int, None], message: str, language: Union[str, None] = None):
        super().__init__(f"{'' if language is None else language + ': '}At line {line_index + 1}: {message}"
                         if line_index is not None else
                         ('' if language is None else language + ': ') + message)


class Reference:
    
    def __init__(self, language_name: str, line_index: int, key: str, target_key: str):
        super().__init__()
        self.language_name = language_name
        self.line_index = line_index
        self.key = key
        self.target_key = target_key
    
    def __repr__(self):
        return f"Reference[{self.language_name}:{self.line_index + 1}: {self.key}={_REFERENCE_START}{self.target_key}]"


class Language:
    
    def __init__(self,
                 name: str,
                 parent_name: Union[str, None],
                 raw_entries: Dict[str, Union[str, Reference]]):
        super().__init__()
        self._name = name
        self._parent_name = parent_name
        self._raw_entries = raw_entries
        # raw_entries with inherited raw entries
        self._all_raw_entries: Union[Dict[str, Union[str, Reference]], None] = None
        self._entries: Union[Dict[str, str], None] = None
        self._inherited_string_keys: Set[str] = set()
    
    def get_name(self) -> str:
        return self._name
    
    def get_parent_name(self) -> Union[str, None]:
        return self._parent_name
    
    def is_resolved(self) -> bool:
        return self._entries is not None
    
    def _resolve_entry(self, key: str, dependents_stack: dict[str, Reference]):
        """Only call from resolve()"""
        value = self._all_raw_entries[key]
        if isinstance(value, str):
            self._entries[key] = value
            if key not in self._raw_entries:
                self._inherited_string_keys.add(key)
        else:
            assert isinstance(value, Reference)
            if value.target_key not in self._entries:
                if value.target_key not in self._all_raw_entries:
                    raise ParseError(value.line_index, f"Unknown reference to {value.target_key}", value.language_name)
                if key in dependents_stack:
                    raise ParseError(value.line_index, f"Circular dependency in references: {dependents_stack}", value.language_name)
                self._resolve_entry(
                    value.target_key,
                    dependents_stack | {
                        value.target_key: value
                    }
                )
            assert value.target_key in self._entries
            self._entries[key] = self._entries[value.target_key]
    
    def resolve(self, parent_language: Union["Language", None]):
        if self._parent_name is not None:
            if parent_language is None:
                raise RuntimeError("Parent language is None")
            self._all_raw_entries = parent_language.get_all_raw_entries() | self._raw_entries
        else:
            if parent_language is not None:
                raise RuntimeError("Parent language is not None")
            self._all_raw_entries = self._raw_entries
        
        self._entries = {}
        for key, value in self._all_raw_entries.items():
            self._resolve_entry(key, {})
    
    def get_entries(self) -> Dict[str, str]:
        if self._entries is None:
            raise RuntimeError("Not resolved")
        return self._entries
    
    def get_all_raw_entries(self) -> Dict[str, str]:
        if self._all_raw_entries is None:
            raise RuntimeError("Not resolved")
        return self._all_raw_entries
    
    def get_inherited_entry_count(self) -> int:
        if self._entries is None:
            raise RuntimeError("Not resolved")
        return len(self._entries) - len(self._raw_entries)
    
    def get_inherited_keys(self) -> Set[str]:
        if self._entries is None:
            raise RuntimeError("Not resolved")
        return self._entries.keys() - self._raw_entries.keys()
    
    def get_inherited_string_entry_count(self) -> int:
        if self._entries is None:
            raise RuntimeError("Not resolved")
        return len(self._inherited_string_keys)
    
    def get_inherited_string_keys(self) -> Set[str]:
        if self._entries is None:
            raise RuntimeError("Not resolved")
        return self._inherited_string_keys


def _parse(language_name: str, data: str) -> Language:
    lines = data.splitlines(keepends=True)
    
    parent_language: Union[str, None] = None
    raw_entries: Dict[str, Union[str, Reference]] = {}
    
    line_index: int = -1
    while line_index + 1 < len(lines):
        line_index += 1
        line = lines[line_index]
        stripped_line = line.strip(_WHITESPACE_CHARS)
        if stripped_line.startswith(_COMMENT_START)\
           or len(stripped_line) == 0:
            continue
        
        if stripped_line.startswith(_INHERIT_START):
            if parent_language is not None:
                raise ParseError(line_index, "The parent language was already defined")
            parent_language = stripped_line[len(_INHERIT_START):]
            if not parent_language.startswith(" "):
                raise ParseError(line_index, f"Missing whitespace after {_INHERIT_START}")
            parent_language = parent_language.strip(_WHITESPACE_CHARS)
            if _LANG_KEY_PATTERN.fullmatch(parent_language) is None:
                raise ParseError(line_index, f"Parent language name '{parent_language}' does not match pattern {_LANG_KEY_REGEX}")
            continue
        
        index_equal = line.find("=")
        if index_equal == -1:
            raise ParseError(line_index, "No '=' character in line")
        key = stripped_line[:index_equal].strip(_WHITESPACE_CHARS)
        if _LANG_KEY_PATTERN.fullmatch(key) is None:
            raise ParseError(line_index, f"Key '{key}' does not match pattern {_LANG_KEY_REGEX}")
        
        if key in raw_entries:
            raise ParseError(line_index, f"An entry for key '{key}' already exists")
        
        second_part_stripped = line[(index_equal + 1):].strip(_WHITESPACE_CHARS)
        escaped = False
        if second_part_stripped.startswith(_ESCAPED_STRING_INDICATOR):
            escaped = True
            second_part_stripped = second_part_stripped[1:]
        
        if second_part_stripped.startswith(_REFERENCE_START):
            if escaped:
                raise ParseError(line_index, "Reference cannot be escaped")
            _parse_reference(language_name, raw_entries, line_index, key, second_part_stripped)
            
        elif second_part_stripped.startswith(_BLOCK_QUOTE_INDICATOR):
            line_index, value = _parse_block_line_quote(lines, line_index, escaped, line)
            raw_entries[key] = value
            
        elif second_part_stripped.startswith(_LINE_QUOTE_INDICATOR):
            raw_entries[key] = _parse_single_line_quote(line_index, escaped, second_part_stripped)
        else:
            raise ParseError(line_index, f"Expected Reference ({_REFERENCE_START}), "
                                         f"single line quote ({_LINE_QUOTE_INDICATOR}) or "
                                         f"block quote ({_BLOCK_QUOTE_INDICATOR})")
    
    return Language(language_name, parent_language, raw_entries)


def _parse_reference(
        language_name: str,
        raw_entries: Dict[str, Union[str, Reference]],
        line_index: int,
        key: str,
        content: str):
    """
    :param content: Starts with _REFERENCE_START. End is stripped
    """
    target_key = content[len(_REFERENCE_START):]  # No whitespace between _REFERENCE_START and key allowed
    if _LANG_KEY_PATTERN.fullmatch(target_key) is None:
        raise ParseError(line_index, f"Reference Key '{key}' does not match pattern {_LANG_KEY_REGEX}")
    raw_entries[key] = Reference(language_name, line_index, key, target_key)


def _parse_single_line_quote(
        line_index: int,
        escaped: bool,
        content: str) -> str:
    """
    :param content: Starts with _LINE_QUOTE_INDICATOR. End is stripped
    """
    content = content[len(_LINE_QUOTE_INDICATOR):]
    
    if escaped:
        value, after_quote = _read_escaped_string(line_index, content, _LINE_QUOTE_INDICATOR)
        if after_quote is None:
            raise ParseError(line_index, f"Missing closing {_LINE_QUOTE_INDICATOR} for single line quote")
    else:
        quote_end = content.find(_LINE_QUOTE_INDICATOR)
        if quote_end == -1:
            raise ParseError(line_index, f"Missing closing {_LINE_QUOTE_INDICATOR} for single line quote")
        value = content[:quote_end]
        after_quote = content[(quote_end + len(_LINE_QUOTE_INDICATOR)):]
    
    if len(after_quote) > 0:
        raise ParseError(line_index, f"Unexpected characters after line quote: {after_quote} "
                                     f"(Note that an escaped string must be prefixed with "
                                     f"'{_ESCAPED_STRING_INDICATOR}')")
    return value


def _parse_block_line_quote(
        lines: List[str],
        line_index: int,
        escaped: bool,
        start_line: str) -> Tuple[int, str]:
    """
    :param start_line: Contains _BLOCK_QUOTE_INDICATOR
    """
    start_line_index = line_index
    after_quote_start = (start_line[start_line.index(_BLOCK_QUOTE_INDICATOR) + len(_BLOCK_QUOTE_INDICATOR):]
                         .strip("\r\n"))
    if len(after_quote_start) > 0:
        raise ParseError(line_index, f"Unexpected characters after block quote start (No characters ("
                         f"including whitespace) are allowed after the quote start on the same line): "
                         f"{after_quote_start}")
    
    all_line_values: List[str] = []
    while True:
        line_index += 1
        if line_index >= len(lines):
            raise ParseError(start_line_index, f"Missing closing {_BLOCK_QUOTE_INDICATOR} for block quote: "
                                               f"Reached end of file")
        line = lines[line_index]
        if escaped:
            line_value, after_quote_end = _read_escaped_string(line_index, line, _BLOCK_QUOTE_INDICATOR)
            if after_quote_end is not None:
                break
        else:
            quote_end = line.find(_BLOCK_QUOTE_INDICATOR)
            if quote_end != -1:
                line_value = line[:quote_end]
                after_quote_end = line[quote_end + len(_BLOCK_QUOTE_INDICATOR):]
                break
            line_value = line
        
        if line_value.endswith("\r\n"):
            line_end = "\r\n"
        elif line_value.endswith("\n"):
            line_end = "\n"
        else:
            line_end = ""
        if line_value.endswith("\\" + line_end):
            line_value = line_value[:-(len(line_end) + 1)]
        
        all_line_values.append(line_value)
    
    if _ONLY_SPACE_PATTERN.fullmatch(line_value) is None:
        raise ParseError(line_index, f"Closing {_BLOCK_QUOTE_INDICATOR} for block quote must be alone on "
                                     f"a line (Only spaces before)"
                                     f" (Note that an escaped string must be prefixed with "
                                     f"'{_ESCAPED_STRING_INDICATOR}')")
    block_space_prefix = line_value
    if len(after_quote_end.strip(_WHITESPACE_CHARS)) > 0:
        raise ParseError(line_index, f"Closing {_BLOCK_QUOTE_INDICATOR} for block quote must be alone on "
                                     f"a line (No non-whitespace characters after)"
                                     f" (Note that an escaped string must be prefixed with "
                                     f"'{_ESCAPED_STRING_INDICATOR}')")
    
    value = ""
    loop_line_index = start_line_index
    for line_value in all_line_values:
        loop_line_index += 1
        if not line_value.startswith(block_space_prefix):
            raise ParseError(loop_line_index, f"Missing {len(block_space_prefix)} spaces at beginning of line "
                                              f"(All lines must start with the same amount of spaces as the closing "
                                              f"{_BLOCK_QUOTE_INDICATOR} has before it. These spaces are removed from "
                                              f"the final text)")
        value += line_value[len(block_space_prefix):]
    
    if value.endswith("\r\n"):
        value = value[:-2]
    elif value.endswith("\n"):
        value = value[:-1]
    else:
        # Last line might end with a \ (although that is useless)
        pass
    
    return line_index, value


def _read_escaped_string(line_index, content: str, end_indicator: str) -> Tuple[str, Union[str, None]]:
    """String must not contain start indicator"""
    string = ""
    i = -1
    while i + 1 < len(content):
        i += 1
        char = content[i]
        if char == _ESCAPE_CHARACTER:
            if i + 1 == len(content) or content[i + 1] not in _ESCAPE_CHARACTER_MAPPINGS:
                raise ParseError(line_index, f"Invalid escape sequence. Only "
                                             + _ESCAPE_CHARACTER_READABLE_LIST
                                             + f" can be escaped with '{_ESCAPE_CHARACTER}'")
            string += _ESCAPE_CHARACTER_MAPPINGS[content[i + 1]]
            i += 1  # Skip next character
            continue
        if content.startswith(end_indicator, i):
            return string, content[i + len(end_indicator):]
        string += content[i]  # This is a bit inefficient
    return string, None


def _resolve_language(all_languages: Dict[str, Language], language: Language, dependents_stack: List[str]):
    if language.is_resolved():
        return
    
    parent_name = language.get_parent_name()
    if parent_name is not None:
        if parent_name not in all_languages:
            raise ParseError(-1, f"{language.get_name()}: Unknown parent language: {parent_name}")
        if parent_name in dependents_stack:
            raise ParseError(-1, f"{language.get_name()}: Loop in language dependencies: {dependents_stack}")
        _resolve_language(
            all_languages,
            all_languages[parent_name],
            dependents_stack + [language.get_name()])
    
    language.resolve(None if parent_name is None else all_languages[parent_name])


def _parse_all_files_in_dir(source_path: Path, dest_path: Union[Path, None], show_inherited: bool) -> int:
    if not source_path.is_dir():
        print("Source does not exist or is not a directory")
        return -1
    if dest_path is not None and dest_path.exists() and not dest_path.is_dir():
        print("Destination already exist but is not a directory")
        return -1
    
    all_languages: Dict[str, Language] = {}
    for file_path in source_path.iterdir():
        if not file_path.name.endswith(_LANGUAGE_FILE_ENDING):
            continue
        print(f"Parsing {file_path.name}")
        language_name = file_path.name[:-len(_LANGUAGE_FILE_ENDING)]
        all_languages[language_name] = _parse(language_name, file_path.read_text("UTF-8"))
    
    if len(all_languages) == 0:
        print("Found no input files")
        return 0
    
    first_no_parent_language = None
    for _, language in all_languages.items():
        _resolve_language(all_languages, language, [])
        if language.get_parent_name() is None:
            first_no_parent_language = language
    
    assert first_no_parent_language is not None
    
    differences_in_languages = False
    for _, language in all_languages.items():
        entries = language.get_entries()
        too_many_keys = entries.keys() - first_no_parent_language.get_entries().keys()
        missing_keys = first_no_parent_language.get_entries().keys() - entries.keys()
        if len(too_many_keys) > 0:
            differences_in_languages = True
            print(
                f"{language.get_name()}: Got too many keys (compared to {first_no_parent_language.get_name()}): {too_many_keys}")
        if len(missing_keys) > 0:
            differences_in_languages = True
            print(
                f"{language.get_name()}: Missing keys (compared to {first_no_parent_language.get_name()}): {missing_keys}")
        inherited_entry_count = language.get_inherited_entry_count()
        inherited_string_entry_count = language.get_inherited_string_entry_count()
        if len(entries) > 0:
            print(f"{language.get_name()}: Got {len(entries)} entries, "
                  f"{inherited_entry_count} ({(inherited_entry_count*10000)//len(entries)/100}%) of which are inherited and "
                  f"{inherited_string_entry_count} ({(inherited_string_entry_count*10000)//len(entries)/100}%) of which are inherited strings"
                  )
            if show_inherited and (inherited_entry_count > 0 or inherited_string_entry_count > 0):
                print(f"{language.get_name()}: Inherited keys: {language.get_inherited_keys()}")
                print(f"{language.get_name()}: Inherited string keys: {language.get_inherited_string_keys()}")
        else:
            print(f"{language.get_name()}: Got no entries!")
        if dest_path is not None:
            dest_path.joinpath(language.get_name() + ".json").write_text(json.dumps(entries), "UTF-8")
    
    if differences_in_languages:
        print("Found some key differences in languages")
        return -1
    else:
        print(f"Parsed {len(first_no_parent_language.get_entries())} entries per language")
        return 0


def _parse_language_recursive(file_path: Path, dependents_stack: Union[List[str], None] = None) -> Language:
    assert file_path.name.endswith(_LANGUAGE_FILE_ENDING)
    if not file_path.is_file():
        raise ParseError(-1, f"Language file {file_path} does not exist or is not a file")
    print(f"Parsing {file_path.name}")
    language_name = file_path.name[:-len(_LANGUAGE_FILE_ENDING)]
    language = _parse(language_name, file_path.read_text("UTF-8"))
    if language.get_parent_name() is None:
        language.resolve(None)
        return language
    
    parent_name = language.get_parent_name()
    if dependents_stack is not None and parent_name in dependents_stack:
        raise ParseError(-1, f"{language.get_name()}: Loop in language dependencies: {dependents_stack}")
    
    parent_language = _parse_language_recursive(
        file_path.parent.joinpath(f"{parent_name}{_LANGUAGE_FILE_ENDING}"),
        ([] if dependents_stack is None else dependents_stack) + [language_name]
    )
    language.resolve(parent_language)
    return language


def _parse_single_file(source_path: Path, dest_path: Union[Path, None]) -> int:
    if not source_path.is_file():
        print("Source does not exist or is not a file")
        return -1
    if dest_path is not None and dest_path.exists() and not dest_path.is_file():
        print("Destination already exists but is not a file")
        return -1
    
    language = _parse_language_recursive(source_path)
    print(f"{language.get_name()}: Parsed {len(language.get_entries())} entries successfully")
    if dest_path is not None:
        dest_path.write_text(json.dumps(language.get_entries()), "UTF-8")
    return 0


def _run_main():
    arg_parser = argparse.ArgumentParser("language_parser")
    arg_parser.add_argument("--src", "-s",
                            help="File which contains the language definition",
                            type=str,
                            required=True)
    arg_parser.add_argument("--dest", "-d",
                            help="File into which the parsed JSON data is written",
                            type=str,
                            required=False)
    arg_parser.add_argument("--folder", "-f",
                            help=f"If specified src (and dest) must be a folder of language files. Only files ending "
                                 f"with {_LANGUAGE_FILE_ENDING} are parsed",
                            action="store_true",
                            required=False)
    arg_parser.add_argument("--show-inherited", "-i",
                            help="If specified a list of inherited keys is printed for each language (only if -f is specified)",
                            action="store_true",
                            required=False)
    arg_parser.usage = _HELP_MESSAGE
    args = arg_parser.parse_args()
    source_path = Path(args.src)
    dest_path = None if args.dest is None else Path(args.dest)
    show_inherited = bool(args.show_inherited)
    
    if args.folder:
        return _parse_all_files_in_dir(source_path, dest_path, show_inherited)
    else:
        if show_inherited:
            raise RuntimeError("Cannot show inherited keys for single file")
        return _parse_single_file(source_path, dest_path)


if __name__ == "__main__":
    _run_main()
