import { basePath } from "../../basepath";
import { ApiError } from "./ApiError";

import type {
    GetStatusResponse,
    GetHomepageResponse,
    GetCoursesResponse,
    int,
    id_string,
    GetCourseResponse,
    GetCourseLectureResponse,
    AuthenticationPasswordRequest,
    AuthenticationFsmpiRequest,
    AuthenticationFsmpiResponse,
    AuthenticationStartOAuthRequest,
    AuthenticationStartOAuthResponse,
    AuthenticationStatusRequest,
    AuthenticationStatusResponse,
    PutFeedbackNewRequest,
    GetFeedbackRequest,
    GetFeedbackResponse,
    GetConfigurationResponse,
    GetNewConfigurationResponse,
    CreateConfiguredObjectRequest,
    CreateConfiguredObjectResponse,
    UpdateConfiguredObjectRequest,
    ModifyManyObjectsRequest,
    ChapterSuggestionRequest,
    GetChangelogRequest,
    GetChangelogResponse,
    SearchResponse,
    GetSorterLogResponse,
    GetUsersResponse,
    GetMySettingsResponse,
    UpdateMySettingsRequest,
    GetJobsRequest,
    GetJobsResponse,
    JobActionRequest,
} from "./api_v1_types";
/* global RequestInit */

export const JOB_STATES = ["ready", "running", "finished", "failed", "deleted"];
export const JOB_TYPES = [
    "probe",
    "probe-raw",
    "remux",
    "thumbnail",
    "transcode",
    "publish_video",
    "simple_live_transcode",
    "complex_live_transcode",
    "live_forward",
];
export const OBJECT_TYPES = [
    "announcement",
    "featured",
    "course",
    "lecture",
    "media_source",
    "chapter",
    "permission",
];

export type Backend = BackendImpl;

export class BackendImpl {
    private csrfToken?: string;
    public isPrivileged = false;

    protected fetchImpl(url: string, init?: RequestInit): Promise<Response> {
        return fetch(url, init);
    }

    assetUrl(): string {
        return "https://video.fsmpi.rwth-aachen.de/files";
    }

    staticUrl(): string {
        return window.location.origin + basePath + "/static";
    }

    baseUrl(): string {
        if (process.env.NEXT_PUBLIC_API_URL_PREFIX === undefined)
            throw new Error(
                "Missing NEXT_PUBLIC_API_URL_PREFIX env variable. For development put 'NEXT_PUBLIC_API_URL_PREFIX=http://localhost:5000' into .env.local",
            );
        if (process.env.NEXT_PUBLIC_API_URL_PREFIX === "origin")
            return window.location.origin + basePath + "/api/v0";

        return process.env.NEXT_PUBLIC_API_URL_PREFIX + "/api/v0";
    }

    // Helper method to process response
    private processResponse<T>(response: Promise<Response>): Promise<T> {
        return response
            .then((res) => {
                if (res.headers.get("content-type") !== "application/json") {
                    throw new Error(
                        "Invalid content type: " +
                            res.headers.get("content-type") +
                            ", url=" +
                            res.url,
                    );
                }
                if (!res.ok) {
                    // Attempt to parse the response as json to check for api_error
                    return res.json().then((errorBody) => {
                        // Check if the error body has the structure of api_error
                        if (errorBody.error_code && errorBody.message) {
                            throw new ApiError(errorBody, res.status, res.url);
                        } else {
                            // If the response does not match the api_error structure, throw a generic error
                            throw new Error(`${res.statusText}, invalid error, url=${res.url}`);
                        }
                    });
                }
                return res.json();
            })
            .catch((error) => {
                console.error("API call failed:", error);
                throw error;
            });
    }

    private fetch(url: string, init?: RequestInit): Promise<Response> {
        if (!init) init = {};
        if (!init.headers) init.headers = {};
        if (this.csrfToken) {
            (init.headers as any)["X-Csrf-Token"] = this.csrfToken;
        } else {
            // get token from cookie
            const match = document.cookie.match(/csrf_token=([^;]+)/);
            if (match) {
                (init.headers as any)["X-Csrf-Token"] = match[1];
            }
        }

        if (this.isPrivileged) {
            init.cache = "no-cache";
        }
        init.credentials = "include";
        return this.fetchImpl(url, init);
    }

    setCsrfToken(csrfToken?: string) {
        this.csrfToken = csrfToken;
    }

    // GET /status
    getStatus(): Promise<GetStatusResponse> {
        return this.processResponse<GetStatusResponse>(this.fetch(`${this.baseUrl()}/status`));
    }

    // GET /homepage
    getHomepage(): Promise<GetHomepageResponse> {
        return this.processResponse<GetHomepageResponse>(this.fetch(`${this.baseUrl()}/homepage`));
    }

    // GET /courses
    getCourses(): Promise<GetCoursesResponse> {
        return this.processResponse<GetCoursesResponse>(this.fetch(`${this.baseUrl()}/courses`));
    }

    // GET /course/{course_id}?include_lectures={include_lectures}
    getCourse(course_id: int | id_string, include_lectures?: boolean): Promise<GetCourseResponse> {
        let url = `${this.baseUrl()}/course/${course_id}`;
        if (include_lectures !== undefined) {
            url += `?include_lectures=${include_lectures}`;
        }
        return this.processResponse<GetCourseResponse>(this.fetch(url));
    }

    // GET /lecture/{lecture_id}
    getCourseLecture(lecture_id: int): Promise<GetCourseLectureResponse> {
        const url = `${this.baseUrl()}/lecture/${lecture_id}`;
        return this.processResponse<GetCourseLectureResponse>(this.fetch(url));
    }

    // POST /authentication/password
    authenticatePassword(request: AuthenticationPasswordRequest): Promise<{}> {
        return this.processResponse<{}>(
            this.fetch(`${this.baseUrl()}/authentication/password`, {
                method: "POST",
                headers: {
                    "Content-Type": "application/json",
                },
                body: JSON.stringify(request),
            }),
        );
    }

    // POST /authentication/fsmpi
    authenticateFsmpi(request: AuthenticationFsmpiRequest): Promise<AuthenticationFsmpiResponse> {
        return this.processResponse<AuthenticationFsmpiResponse>(
            this.fetch(`${this.baseUrl()}/authentication/fsmpi`, {
                method: "POST",
                headers: {
                    "Content-Type": "application/json",
                },
                body: JSON.stringify(request),
            }),
        );
    }

    // POST /authentication/start_oauth
    startAuthentication(
        request: AuthenticationStartOAuthRequest,
    ): Promise<AuthenticationStartOAuthResponse> {
        return this.processResponse<AuthenticationStartOAuthResponse>(
            this.fetch(`${this.baseUrl()}/authentication/start_oauth`, {
                method: "POST",
                headers: {
                    "Content-Type": "application/json",
                },
                body: JSON.stringify(request),
            }),
        );
    }

    // POST /authentication/status
    getAuthenticationStatus(
        request?: AuthenticationStatusRequest,
    ): Promise<AuthenticationStatusResponse> {
        if (request === undefined) request = {};
        return this.processResponse<AuthenticationStatusResponse>(
            this.fetch(`${this.baseUrl()}/authentication/status`, {
                method: "POST",
                headers: {
                    "Content-Type": "application/json",
                },
                body: JSON.stringify(request),
            }),
        );
    }

    // POST /authentication/moderator_logout
    logout(): Promise<{}> {
        return this.processResponse<{}>(
            this.fetch(`${this.baseUrl()}/authentication/moderator_logout`, {
                method: "POST",
                headers: {
                    "Content-Type": "application/json",
                },
            }),
        );
    }

    // PUT /feedback/new
    putNewFeedback(req: PutFeedbackNewRequest): Promise<{}> {
        return this.processResponse<{}>(
            this.fetch(`${this.baseUrl()}/feedback/new`, {
                method: "PUT",
                headers: {
                    "Content-Type": "application/json",
                },
                body: JSON.stringify(req),
            }),
        );
    }

    // GET /feedback
    getFeedback(req: GetFeedbackRequest): Promise<GetFeedbackResponse> {
        let query_params = [];
        for (const [key, value] of Object.entries(req)) {
            if (value !== undefined) {
                query_params.push(`${key}=${value}`);
            }
        }
        return this.processResponse<GetFeedbackResponse>(
            this.fetch(`${this.baseUrl()}/feedback?${query_params.join("&")}`),
        );
    }

    // GET /object_management/{object_type}/{object_id}/configuration
    getOMConfiguration(object_type: string, object_id: int): Promise<GetConfigurationResponse> {
        return this.processResponse<GetConfigurationResponse>(
            this.fetch(
                `${this.baseUrl()}/object_management/${object_type}/${object_id}/configuration`,
            ),
        );
    }

    // GET /object_management/{object_type}/new/configuration
    getNewOMConfiguration(object_type: string): Promise<GetNewConfigurationResponse> {
        return this.processResponse<GetNewConfigurationResponse>(
            this.fetch(`${this.baseUrl()}/object_management/${object_type}/new/configuration`),
        );
    }

    // PUT /object_management/{object_type}/new
    createOMObject(
        object_type: string,
        request: CreateConfiguredObjectRequest,
    ): Promise<CreateConfiguredObjectResponse> {
        return this.processResponse<CreateConfiguredObjectResponse>(
            this.fetch(`${this.baseUrl()}/object_management/${object_type}/new`, {
                method: "PUT",
                headers: {
                    "Content-Type": "application/json",
                },
                body: JSON.stringify(request),
            }),
        );
    }

    // PATCH /object_management/{object_type}/{object_id}/configuration
    updateOMObject(
        object_type: string,
        object_id: int,
        request: UpdateConfiguredObjectRequest,
    ): Promise<{}> {
        return this.processResponse<{}>(
            this.fetch(
                `${this.baseUrl()}/object_management/${object_type}/${object_id}/configuration`,
                {
                    method: "PATCH",
                    headers: {
                        "Content-Type": "application/json",
                    },
                    body: JSON.stringify(request),
                },
            ),
        );
    }

    // POST /object_management/modify_many
    modifyManyOMObjects(request: ModifyManyObjectsRequest): Promise<{}> {
        return this.processResponse<{}>(
            this.fetch(`${this.baseUrl()}/object_management/modify_many`, {
                method: "POST",
                headers: {
                    "Content-Type": "application/json",
                },
                body: JSON.stringify(request),
            }),
        );
    }

    // DELETE /object_management/{object_type}/{object_id}
    deleteOMObject(object_type: string, object_id: int): Promise<{}> {
        return this.processResponse<{}>(
            this.fetch(`${this.baseUrl()}/object_management/${object_type}/${object_id}`, {
                method: "DELETE",
                headers: {
                    "Content-Type": "application/json",
                },
            }),
        );
    }

    // PUT /lecture/{lecture_id}/chapter_suggestion
    suggestChapter(lecture_id: int, request: ChapterSuggestionRequest): Promise<{}> {
        return this.processResponse<{}>(
            this.fetch(`${this.baseUrl()}/lecture/${lecture_id}/chapter_suggestion`, {
                method: "PUT",
                headers: {
                    "Content-Type": "application/json",
                },
                body: JSON.stringify(request),
            }),
        );
    }

    // GET /lecture_import/rwth_online/{rwth_course_id}, mimetype: text/calendar
    importRWTHOnlineCourse(rwth_course_id: int): Promise<string> {
        return this.fetch(`${this.baseUrl()}/lecture_import/rwth_online/${rwth_course_id}`, {
            headers: {
                "Content-Type": "text/calendar",
            },
        })
            .then((res) => {
                if (!res.ok) {
                    throw new Error(`${res.statusText}, url=${res.url}`);
                }
                return res.text();
            })
            .catch((error) => {
                console.error("API call failed:", error);
                throw error;
            });
    }

    // GET /object_management/changelog?object_type={object_type}&object_id={object_id}&field={field}&entries_per_page={entries_per_page}&page={page}
    getChangelog(req: GetChangelogRequest): Promise<GetChangelogResponse> {
        let query_params = [];
        for (const [key, value] of Object.entries(req)) {
            if (value !== undefined) {
                query_params.push(`${key}=${value}`);
            }
        }
        return this.processResponse<GetChangelogResponse>(
            this.fetch(`${this.baseUrl()}/object_management/changelog?${query_params.join("&")}`),
        );
    }

    // GET /search?q={query}
    search(query: string): Promise<SearchResponse> {
        return this.processResponse<SearchResponse>(
            this.fetch(`${this.baseUrl()}/search?q=${query}`),
        );
    }

    // POST /object_management/{object_type}/{object_id}/resurrect
    resurrectOMObject(object_type: string, object_id: int): Promise<{}> {
        return this.processResponse<{}>(
            this.fetch(
                `${this.baseUrl()}/object_management/${object_type}/${object_id}/resurrect`,
                {
                    method: "POST",
                    headers: {
                        "Content-Type": "application/json",
                    },
                },
            ),
        );
    }

    // GET /sorter/log
    getSorterLog(): Promise<GetSorterLogResponse> {
        return this.processResponse<GetSorterLogResponse>(
            this.fetch(`${this.baseUrl()}/sorter/log`),
        );
    }

    // DELETE /sorter/error_log/{log_id}
    deleteSorterErrorLog(log_id: int): Promise<{}> {
        return this.processResponse<{}>(
            this.fetch(`${this.baseUrl()}/sorter/error_log/${log_id}`, {
                method: "DELETE",
                headers: {
                    "Content-Type": "application/json",
                },
            }),
        );
    }

    // GET /users
    getUsers(): Promise<GetUsersResponse> {
        return this.processResponse<GetUsersResponse>(this.fetch(`${this.baseUrl()}/users`));
    }

    // GET /user/me/settings
    getMySettings(): Promise<GetMySettingsResponse> {
        return this.processResponse<GetMySettingsResponse>(
            this.fetch(`${this.baseUrl()}/user/me/settings`),
        );
    }

    // PATCH /user/me/settings
    updateMySettings(request: UpdateMySettingsRequest): Promise<{}> {
        return this.processResponse<{}>(
            this.fetch(`${this.baseUrl()}/user/me/settings`, {
                method: "PATCH",
                headers: {
                    "Content-Type": "application/json",
                },
                body: JSON.stringify(request),
            }),
        );
    }

    // GET /jobs?entries_per_page={entries_per_page}&page={page}&type={type}&worker={worker_id}&state={state}
    getJobs(req: GetJobsRequest): Promise<GetJobsResponse> {
        let query_params = [];
        for (const [key, value] of Object.entries(req)) {
            if (value !== undefined) {
                query_params.push(`${key}=${value}`);
            }
        }
        return this.processResponse<GetJobsResponse>(
            this.fetch(`${this.baseUrl()}/jobs?${query_params.join("&")}`),
        );
    }

    // DELETE /job/{id}
    deleteJob(id: int): Promise<{}> {
        return this.processResponse<{}>(
            this.fetch(`${this.baseUrl()}/job/${id}`, {
                method: "DELETE",
            }),
        );
    }

    // POST /job/{id}/action
    jobAction(id: int, action: JobActionRequest["action"]): Promise<{}> {
        return this.processResponse<{}>(
            this.fetch(`${this.baseUrl()}/job/${id}/action`, {
                method: "POST",
                headers: {
                    "Content-Type": "application/json",
                },
                body: JSON.stringify({ action }),
            }),
        );
    }

    // Special method for language files
    getLanguageFile(lang: string): Promise<Record<string, string>> {
        return this.fetch(`${this.staticUrl()}/lang/${lang}.json`)
            .then((res) => {
                if (!res.ok) {
                    throw new Error(`Cannot load language ${lang}. Status: ${res.status}`);
                }
                return res.json();
            })
            .catch((error) => {
                console.error(`Exception while loading language ${lang}:`, error);
                return {};
            });
    }
}

export class SmartCacheBackend extends BackendImpl {
    protected fetchImpl(url: string, init?: RequestInit) {
        let shouldCache = true;
        if (!SmartCacheBackend.isSupported()) shouldCache = false;
        if (init && init.cache === "no-cache") shouldCache = false;
        if (!url.startsWith(this.baseUrl())) shouldCache = false;
        if (!shouldCache) return super.fetchImpl(url, init);

        const cacheName = `api-v1-${this.isPrivileged ? "privileged" : "public"}`;
        return caches.open(cacheName).then((cache) => {
            return cache.match(url).then((response) => {
                if (response) {
                    console.log(`Cache hit (${cacheName}) on ${url}`);
                    return response;
                } else {
                    return fetch(url, init).then((response) => {
                        cache.put(url, response.clone());
                        return response;
                    });
                }
            });
        });
    }

    static isSupported() {
        return window.isSecureContext && "caches" in window;
    }
}

export class InvalidBackend extends BackendImpl {
    protected fetchImpl(url: string, init?: RequestInit): Promise<Response> {
        return Promise.reject(new Error("Invalid backend"));
    }
}
