// Base types
export type int = number;
export type long = number;
export type id_string = string;
export type semester_string = string;
export type duration = int;
export type datetime = string; // ISO 8601 format
export type timestamp = int;
export type url = string;
export type data_size = long;
export type view_permissions_type = string;

// Enums
export type announcement_type = "info" | "warning" | "important";
export type announcement_visibility = "only_main_page" | "all_pages" | "all_pages_and_embed";
export type authentication_method = "public" | "rwth" | "moodle" | "fsmpi" | "password";
export type authentication_methods = authentication_method[];
export type job_state = "ready" | "running" | "finished" | "failed" | "deleted";
export type job_type =
    | "probe"
    | "probe-raw"
    | "remux"
    | "thumbnail"
    | "transcode"
    | "publish_video"
    | "simple_live_transcode"
    | "complex_live_transcode"
    | "live_forward";
export type status_value = "available" | "readonly" | "unavailable";

// Complex Types
export interface announcement {
    id: int;
    type: announcement_type;
    visibility: announcement_visibility;
    text: string;
    is_currently_visible?: boolean;
    has_expired?: boolean;
}

export interface homepage {
    featured: featured[];
    upcoming_lectures: lecture[];
    new_lectures: lecture[];
    courses_context: { [key: int]: course };
}

export interface featured {
    type: "plain" | "image" | "lecture" | "course_list";
    title: string;
    text?: string;
    image_url?: url;
    lecture?: lecture;
    courses?: int[];
    id?: int;
    is_visible?: boolean;
    display_priority?: int;
}

export interface course {
    id: int;
    id_string: id_string;
    full_name: string;
    short_name: string;
    semester?: semester_string;
    organizer: string;
    topic: string;
    description: string;
    show_chapters_on_course: boolean;
    default_authentication_methods: authentication_methods;
    authentication_information?: string;
    lectures?: lecture[];
    is_listed?: boolean;
    is_visible?: boolean;
}

export interface lecture {
    id: int;
    course_id: int;
    title: string;
    speaker: string;
    location: string;
    time: datetime;
    duration: duration;
    description: string;
    thumbnail_url?: url;
    no_recording: boolean;
    livestream_planned: boolean;
    livestream_url?: url;
    allow_embed: boolean;
    chapters?: chapter[];
    media_sources?: media_source[];
    authentication_methods: authentication_methods;
    is_visible?: boolean;
    internal_comment?: string;
}

export interface chapter {
    start_time: timestamp;
    name: string;
    id?: int;
    is_visible?: boolean;
}

export interface media_source {
    quality: media_quality;
    size: data_size;
    comment: string;
    player_url: url;
    download_url?: url;
    id?: int;
    is_visible?: boolean;
    internal_comment?: string;
}

export interface media_quality {
    name: string;
    resolution: string;
    aspect_ratio: string;
    priority: int;
}

export interface feedback_entry {
    id: int;
    time_created: datetime;
    email: string;
    text: string;
}

export interface api_error {
    error_code: string;
    message: string;
    server: string;
}

export interface user_info {
    id: int;
    id_string: string;
    name: string;
}

export interface view_permissions {
    type: view_permissions_type;
    rwth_authentication?: boolean;
    fsmpi_authentication?: boolean;
    moodle_course_ids?: int[];
    passwords?: { [key: string]: string };
}

export interface field_description {
    id: id_string;
    type: id_string;
    default_value?: any;

    may_be_null?: boolean;
    min_value?: int | long;
    max_value?: int | long;
    min_length?: int;
    max_length?: int;
}

export interface field_value {
    value: any;

    description: field_description;
}

export interface changelog_entry {
    id: int;
    type: id_string;
    change_time: datetime;
    modifying_user_id: int;
    object_id: int;

    /// type: modification
    object_type?: id_string;
    old_value?: any;
    new_value?: any;
    field_description?: field_description;

    /// type: deletion_change
    // object_type?: id_string;
    is_now_deleted?: boolean;

    /// type: creation
    // object_type?: id_string;
    variant?: id_string;

    /// type: unknown
    unknown_type?: string;
    unknown_field?: string;
    /* old and new value are strings in this case */
}

export interface sorter_log_entry {
    time: datetime;
    file_path: string;
    media_source_id: int;
    lecture_id: int;
    lecture_title?: string;
    lecture_time?: datetime;
    course_id?: int;
    course_full_name?: string;
}

export interface sorter_error_log_entry {
    id: int;
    time: datetime;
    file_path: string;
    matched_course_id?: int;
    matched_lectures_ids?: int[];
}

export interface settings_entry {
    id: id_string;
    type: id_string;

    // subsection type
    entries?: settings_entry[];

    // boolean type
    value?: boolean;
}

export interface settings_section {
    name: id_string;
    entries: settings_entry[];
}

export interface settings {
    sections: settings_section[];
}

export interface job {
    id: int;
    type: id_string;
    priority: id_string;
    worker_id?: string;
    last_worker_ping?: datetime;
    state: id_string;
    is_canceled: boolean;
    creation_time: datetime;
    start_time?: datetime;
    finish_time?: datetime;
    data: string;
    status: string;
}

export interface worker {
    hostname: string;
    last_ping: datetime;
}

// Request objects:

/// GET /status
// Request Interface: Not needed as there are no request parameters

// Response Interface
export interface GetStatusResponse {
    status: status_value;
    is_debug: boolean;
    git_commit_hash?: string;
    server_name: string;
    announcements: announcement[];
}

/// GET /homepage
// Request Interface: Not needed as there are no request parameters

// Response Interface
export interface GetHomepageResponse {
    featured: featured[];
    upcoming_lectures: lecture[];
    new_lectures: lecture[];
    courses_context: { [key: string]: course };
}

/// GET /courses
// Request Interface: Not needed as there are no request parameters

// Response Interface
export interface GetCoursesResponse {
    courses: { [key: string]: course };
}

/// GET /courses/{course_id}?include_lectures={include_lectures}
// Request Interface
export interface GetCourseRequest {
    course_id: int | id_string;
    include_lectures?: boolean;
}

// Response Interface
export type GetCourseResponse = course;

/// GET /courses/{course_id}/lectures/{lecture_id}
// Request Interface
export interface GetCourseLectureRequest {
    course_id: int;
    lecture_id: int;
}

// Response Interface
export interface GetCourseLectureResponse {
    lecture: lecture;
}

/// POST /authentication/password
// Request Interface
export interface AuthenticationPasswordRequest {
    username: string;
    password: string;
    lecture_id: int;
}

/// POST /authentication/fsmpi
// Request Interface
export interface AuthenticationFsmpiRequest {
    username: string;
    password: string;
    lecture_id?: int;
}

// Response Interface
export interface AuthenticationFsmpiResponse {
    is_lecture_authenticated?: boolean;
    user?: user_info;
}

/// POST /authentication/start_oauth
// Request Interface
export interface AuthenticationStartOAuthRequest {
    type: "rwth" | "moodle";
}

// Response Interface
export interface AuthenticationStartOAuthResponse {
    verification_url: url;
}

/// GET /authentication/status
// Request Interface
export interface AuthenticationStatusRequest {
    lecture_id?: int;
}
// Response Interface
export interface AuthenticationStatusResponse {
    authenticated_methods: authentication_methods;
    in_progress_authentication?: "rwth" | "moodle";
    is_lecture_authenticated?: boolean;
    user_info?: user_info;
}

/// PUT /feedback/new
// Request Interface
export interface PutFeedbackNewRequest {
    email?: string;
    text: string;
}
// Response Interface: Not needed as there is no response

/// GET /feedback
// Request Interface
export interface GetFeedbackRequest {
    entries_per_page?: int;
    page?: int;
}
// Response Interface
export interface GetFeedbackResponse {
    page_count: int;
    page: feedback_entry[];
}

/// GET /object_management/{object_type}/{object_id}/configuration
// Request Interface
export interface GetConfigurationRequest {
    object_type:
        | "announcement"
        | "featured"
        | "course"
        | "lecture"
        | "permission"
        | "chapter"
        | "media_source";
    object_id: int;
}
// Response Interface
export interface GetConfigurationResponse {
    fields: field_value[];
    variant_fields?: { [key: string]: field_value[] };
}

/// GET /object_management/{object_type}/new/configuration
// Request Interface
export interface GetNewConfigurationRequest {
    object_type:
        | "announcement"
        | "featured"
        | "course"
        | "lecture"
        | "permission"
        | "chapter"
        | "media_source";
}
// Response Interface
export interface GetNewConfigurationResponse {
    fields: field_description[];
    variant_fields: { [key: string]: field_description[] };
}

/// PUT /object_management/{object_type}/new
// Request Interface
export interface CreateConfiguredObjectRequest {
    parent_type?: id_string;
    parent_id?: int;
    values: { [key: id_string]: any };
    variant?: id_string;
}
// Response Interface
export interface CreateConfiguredObjectResponse {
    id: int;
}

/// PATCH /object_management/{object_type}/{object_id}/configuration
// Request Interface
export interface UpdateConfiguredObjectRequest {
    updates: { [key: id_string]: any };
    expected_current_values: { [key: id_string]: any };
}
// Response Interface: Not needed as there is no response

/// POST /object_management/modify_many
// Request Interface
export interface ModifyManyObjectsRequest {
    objects: {
        type: id_string;
        id: int;
        updates: { [key: id_string]: any };
        expected_current_values: { [key: id_string]: any };
    }[];
}
// Response Interface: Not needed as there is no response

/// PUT /lecture/{lecture_id}/chapter_suggestion
// Request Interface
export type ChapterSuggestionRequest = chapter;
// Response Interface: Not needed as there is no response

/// GET /object_management/changelog?object_type={object_type}&object_id={object_id}&field={field}&entries_per_page={entries_per_page}&page={page}
// Request Interface
export interface GetChangelogRequest {
    entries_per_page?: int;
    page?: int;
    object_type?: id_string;
    object_id?: int;
    field?: id_string;
}
// Response Interface
export interface GetChangelogResponse {
    page_count: int;
    page: changelog_entry[];
}

/// GET /object_management/{object_type}/{object_id}/{field_id}/changelog
// Request Interface
export interface GetFieldChangelogRequest {
    object_type: id_string;
    object_id: int;
    field_id: id_string;
}
// Response Interface
export interface GetFieldChangelogResponse {
    changelog: changelog_entry[];
}

/// GET /search?q={query}
// Request Interface
export interface SearchRequest {
    query: string;
}
// Response Interface
export interface SearchResponse {
    courses: int[];
    lectures: lecture[];
    courses_context: { [key: int]: course };
}

/// POST /object_management/{object_type}/{object_id}/resurrect
// Request Interface
export interface ResurrectObjectRequest {
    object_type: id_string;
    object_id: int;
}
// Response Interface: Not needed as there is no response

/// GET /sorter/log
// Request Interface: Not needed as there are no request parameters
// Response Interface
export interface GetSorterLogResponse {
    log: sorter_log_entry[];
    error_log: sorter_error_log_entry[];
}

/// GET /users
// Request Interface: Not needed as there are no request parameters
// Response Interface
export interface GetUsersResponse {
    users: user_info[];
}

/// GET /user/me/settings
// Request Interface: Not needed as there are no request parameters
// Response Interface
export type GetMySettingsResponse = settings;

/// PATCH /user/me/settings
// Request Interface
export interface UpdateMySettingsRequest {
    updates: { [key: id_string]: any };
}

/// GET /jobs?entries_per_page={entries_per_page}&page={page}&type={type}&worker={worker_id}&state={state}
// Request Interface
export interface GetJobsRequest {
    entries_per_page?: int;
    page?: int;
    type?: job_type;
    worker_id?: string;
    state?: job_state;
}
// Response Interface
export interface GetJobsResponse {
    page_count: int;
    page: job[];
    workers: worker[];
}

/// DELETE /job/{id}
// Request Interface
export interface DeleteJobRequest {
    id: int;
}
// Response Interface: Not needed as there is no response

/// POST /job/{id}/action
// Request Interface
export interface JobActionRequest {
    id: int;
    action: "restart" | "cancel" | "copy";
}
// Response Interface: Not needed as there is no response
