import { useRouter } from "next/router";
import { useLocalStorageState } from "@/misc/Storage";
import { useEditMode } from "./EditModeProvider";
import { OMCreate, OMDelete, OMEdit, EmbeddedOMFieldComponent } from "./OMConfigComponent";
import { StylizedText } from "./StylizedText";
import { TooltipButton } from "@/misc/Util";

import type { announcement } from "@/api/api_v1_types";
import { useLanguage } from "./LanguageProvider";

function hashAnnouncement(announcement: announcement): string {
    const dataString =
        String(announcement.id ?? -1) +
        announcement.type +
        announcement.visibility +
        announcement.text;
    let hash: number = 0;
    for (let i = 0; i < dataString.length; i++) {
        const char: number = dataString.charCodeAt(i);
        // Nothing fancy, just really simple
        hash = (hash << 5) - hash;
        hash += char;
        hash &= 0x7fffffff;
    }
    return hash.toString(16);
}

export default function AnnouncementComponent({
    announcements,
    forceShowHidden,
}: {
    announcements: announcement[];
    forceShowHidden: boolean;
}) {
    const router = useRouter();
    const { editMode } = useEditMode();
    const { language } = useLanguage();
    const [hiddenAnnouncements, setHiddenAnnouncements] = useLocalStorageState<string[]>(
        "hidden_announcements",
        [],
    );

    const isHome = router.asPath === "/";

    const hideAnnouncement = (announcement: announcement) =>
        setHiddenAnnouncements((oldHidden: string[]) => {
            const hash = hashAnnouncement(announcement);

            const newHidden = oldHidden.slice();
            if (!newHidden.includes(hash)) {
                newHidden.push(hash);
            }
            // do some housekeeping
            const allHashes = new Set(
                announcements.map((announcement) => hashAnnouncement(announcement)),
            );
            for (let i = 0; i < newHidden.length; i++) {
                if (!allHashes.has(newHidden[i])) {
                    newHidden.splice(i, 1);
                    i--;
                }
            }
            return newHidden;
        });

    return (
        <div>
            {editMode && (
                <div className="">
                    <OMCreate object_type="announcement">
                        <button className="btn btn-secondary mb-3 w-100" type="button">
                            <i className="bi bi-plus" />
                            {language.get("ui.announcement.create_new")}
                        </button>
                    </OMCreate>
                </div>
            )}
            {announcements.map((announcement, ind) => {
                const hash = hashAnnouncement(announcement);
                let alertColor = "alert-info";
                if (announcement.type == "warning") alertColor = "alert-warning";
                else if (announcement.type == "important") alertColor = "alert-danger";

                if (announcement.visibility == "only_main_page" && !isHome) return null;

                if (!forceShowHidden && !editMode && hiddenAnnouncements.includes(hash))
                    return null;

                return (
                    <div
                        className={"hidden-print alert d-flex align-items-center " + alertColor}
                        key={hash + "#" + ind}
                        role="alert"
                    >
                        {announcement.id ? (
                            <EmbeddedOMFieldComponent
                                object_type="announcement"
                                object_id={announcement.id}
                                field_id="text"
                                field_type="string"
                                initialValue={announcement.text}
                                className="flex-grow-1"
                                allowMarkdown={true}
                            />
                        ) : (
                            <span className={"flex-grow-1 disable-last-paragraph-spacing"}>
                                <StylizedText markdown={true}>{announcement.text}</StylizedText>
                            </span>
                        )}

                        <div className="flex-grow-1" />
                        {editMode &&
                            (announcement.id ? (
                                <>
                                    <OMEdit
                                        object_type="announcement"
                                        object_id={announcement.id}
                                    />
                                    <OMDelete
                                        object_type="announcement"
                                        object_id={announcement.id}
                                    />
                                </>
                            ) : (
                                <TooltipButton>
                                    {language.getStyled(
                                        "ui.announcement.info_static_not_editable",
                                        true,
                                    )}
                                </TooltipButton>
                            ))}
                        {!forceShowHidden && !editMode && (
                            <button
                                type="button"
                                className="btn-close "
                                aria-label="Close"
                                onClick={() => hideAnnouncement(announcement)}
                            />
                        )}
                    </div>
                );
            })}
        </div>
    );
}
