import { useEffect, useRef, useState } from "react";
import Modal from "react-bootstrap/Modal";
import DropdownButton from "react-bootstrap/DropdownButton";
import DropdownItem from "react-bootstrap/DropdownItem";
import { sortSourceNames, sortSources, sourceToName } from "./Player";
import { showWarningToast } from "@/misc/ErrorHandlers";

import type { GetCourseResponse, media_source } from "@/api/api_v1_types";
import { useLanguage } from "./LanguageProvider";
import { StylizedText } from "./StylizedText";

interface DownloadStatus {
    downloaded: number;
    total: number;
    done: boolean;
    percent: number;
    error?: string;
}

async function downloadWithFileSystemAPI(
    replaceStatus: (id: number, fn: (old?: DownloadStatus) => DownloadStatus) => void,
    course: GetCourseResponse,
    chosenLectures: any,
    chosenQualities: any,
    requestStopDownload: any,
) {
    if (!(window as any).showDirectoryPicker) {
        throw "unsupported";
    }
    const saveDir = await (window as any).showDirectoryPicker();
    if (!saveDir) {
        return "cancelled";
    }

    await saveDir.requestPermission({ mode: "readwrite" });

    for (let lecture of course.lectures!) {
        if (chosenLectures.current.get(lecture.id) && lecture.media_sources) {
            try {
                if (requestStopDownload.current) {
                    replaceStatus(lecture.id, () => ({
                        downloaded: 0,
                        total: 0,
                        done: false,
                        percent: 0,
                        error: "Download cancelled",
                    }));
                    continue;
                }
                let chosenSource: media_source | undefined = undefined;
                for (let source of lecture.media_sources) {
                    if (sourceToName(source) === chosenQualities.current.get(lecture.id)) {
                        chosenSource = source;
                        break;
                    }
                }
                if (!chosenSource || !chosenSource.download_url) continue;

                let filename = chosenSource.download_url!.split("/").pop();
                if (!filename || filename.length === 0 || !filename.includes("."))
                    filename = `videoag_${course.id}_${lecture.id}_${chosenSource.quality.name}.mp4`;

                let fileHandle = await saveDir.getFileHandle(filename, { create: true });
                let writable = await fileHandle.createWritable();
                let response = await fetch(chosenSource.download_url!);
                if (response.ok === false)
                    throw new Error("Download failed, status " + response.status);
                let reader = response.body!.getReader();
                let writer = writable.getWriter();
                let lastUpdate = Date.now();
                let downloaded = 0;
                let total = chosenSource.size;
                replaceStatus(lecture.id, () => ({
                    downloaded,
                    total,
                    done: false,
                    percent: 0,
                }));
                while (true) {
                    const { done, value } = await reader.read();
                    if (done || requestStopDownload.current) {
                        break;
                    }
                    await writer.write(value);
                    downloaded += value!.byteLength;
                    let now = Date.now();
                    // after 0.2s, update progress
                    if (now - lastUpdate > 200) {
                        replaceStatus(lecture.id, () => ({
                            downloaded,
                            total,
                            done: false,
                            percent: (downloaded / total) * 100,
                        }));
                        lastUpdate = now;
                    }
                }
                await writer.close();
                if (requestStopDownload.current) {
                    // remove incomplete file
                    saveDir.removeEntry(filename);
                    replaceStatus(lecture.id, (old) => ({
                        ...old!,
                        error: "Download cancelled",
                    }));
                } else {
                    replaceStatus(lecture.id, () => ({
                        downloaded,
                        total,
                        done: true,
                        percent: 100,
                    }));
                }
            } catch (e: any) {
                replaceStatus(lecture.id, (old) => ({
                    ...old!,
                    error: e + "",
                }));
            }
        }
    }

    return "success";
}

async function downloadWithOPFSAPI(
    replaceStatus: (id: number, fn: (old?: DownloadStatus) => DownloadStatus) => void,
    course: GetCourseResponse,
    chosenLectures: any,
    chosenQualities: any,
    requestStopDownload: any,
) {
    if (!navigator.storage || !navigator.storage.getDirectory) {
        throw "unsupported";
    }
    let saveDir = await navigator.storage.getDirectory();
    if (!saveDir) {
        return "cancelled";
    }
    // clear the directory
    saveDir.removeEntry("videoag", { recursive: true });
    saveDir = await saveDir.getDirectoryHandle("videoag", { create: true });
    if (!saveDir) {
        return "cancelled";
    }

    for (let lecture of course.lectures!) {
        if (chosenLectures.current.get(lecture.id) && lecture.media_sources) {
            try {
                if (requestStopDownload.current) {
                    replaceStatus(lecture.id, () => ({
                        downloaded: 0,
                        total: 0,
                        done: false,
                        percent: 0,
                        error: "Download cancelled",
                    }));
                    continue;
                }
                let chosenSource: media_source | undefined = undefined;
                for (let source of lecture.media_sources) {
                    if (sourceToName(source) === chosenQualities.current.get(lecture.id)) {
                        chosenSource = source;
                        break;
                    }
                }
                if (!chosenSource || !chosenSource.download_url) continue;

                let filename = chosenSource.download_url!.split("/").pop();
                if (!filename || filename.length === 0 || !filename.includes("."))
                    filename = `videoag_${course.id}_${lecture.id}_${chosenSource.quality.name}.mp4`;

                let fileHandle = await saveDir.getFileHandle(filename, { create: true });
                let writable = await fileHandle.createWritable();
                let response = await fetch(chosenSource.download_url!);
                if (response.ok === false)
                    throw new Error("Download failed, status " + response.status);
                let reader = response.body!.getReader();
                let writer = writable.getWriter();
                let lastUpdate = Date.now();
                let downloaded = 0;
                let total = chosenSource.size;
                replaceStatus(lecture.id, () => ({
                    downloaded,
                    total,
                    done: false,
                    percent: 0,
                }));
                while (true) {
                    const { done, value } = await reader.read();
                    if (done || requestStopDownload.current) {
                        break;
                    }
                    await writer.write(value);
                    downloaded += value!.byteLength;
                    let now = Date.now();
                    // after 0.2s, update progress
                    if (now - lastUpdate > 200) {
                        replaceStatus(lecture.id, () => ({
                            downloaded,
                            total,
                            done: false,
                            percent: (downloaded / total) * 100,
                        }));
                        lastUpdate = now;
                    }
                }
                await writer.close();
                if (requestStopDownload.current) {
                    // remove incomplete file
                    saveDir.removeEntry(filename);
                    replaceStatus(lecture.id, (old) => ({
                        ...old!,
                        error: "Download cancelled",
                    }));
                } else {
                    // create a element to download the file
                    let a = document.createElement("a");
                    a.href = URL.createObjectURL(await fileHandle.getFile());
                    a.download = filename;
                    a.target = "_blank";
                    a.click();

                    replaceStatus(lecture.id, () => ({
                        downloaded,
                        total,
                        done: true,
                        percent: 100,
                    }));
                }
            } catch (e: any) {
                replaceStatus(lecture.id, (old) => ({
                    ...old!,
                    error: e + "",
                }));
            }
        }
    }
    return "success";
}

async function downloadWithDownloadAPI(
    replaceStatus: (id: number, fn: (old?: DownloadStatus) => DownloadStatus) => void,
    course: GetCourseResponse,
    chosenLectures: any,
    chosenQualities: any,
    requestStopDownload: any,
) {
    for (let lecture of course.lectures!) {
        if (chosenLectures.current.get(lecture.id) && lecture.media_sources) {
            try {
                if (requestStopDownload.current) {
                    replaceStatus(lecture.id, () => ({
                        downloaded: 0,
                        total: 0,
                        done: false,
                        percent: 0,
                        error: "Download cancelled",
                    }));
                    continue;
                }
                let chosenSource: media_source | undefined = undefined;
                for (let source of lecture.media_sources) {
                    if (sourceToName(source) === chosenQualities.current.get(lecture.id)) {
                        chosenSource = source;
                        break;
                    }
                }
                if (!chosenSource || !chosenSource.download_url) continue;

                let filename = chosenSource.download_url!.split("/").pop();
                if (!filename || filename.length === 0 || !filename.includes("."))
                    filename = `videoag_${course.id}_${lecture.id}_${chosenSource.quality.name}.mp4`;

                let a = document.createElement("a");
                a.href = chosenSource.download_url!;
                a.download = filename;
                a.target = "_blank";
                a.click();
            } catch (e: any) {
                replaceStatus(lecture.id, (old) => ({
                    ...old!,
                    error: e + "",
                }));
            }
        }
    }

    return "success";
}

export default function DownloadAllModal({ course }: { course: GetCourseResponse }) {
    const [showModal, setShowModal] = useState(false);
    const chosenQualities = useRef(new Map<number, string>());
    const chosenLectures = useRef(new Map<number, boolean>());
    const [forceReload, setForceReload] = useState(0);
    const [downloadStarted, setDownloadStarted] = useState(false);
    const requestStopDownload = useRef(false);
    const [downloadStatus, setDownloadStatus] = useState<Map<number, DownloadStatus>>();
    const { language } = useLanguage();

    const closeClicked = () => {
        if (!downloadStarted) setShowModal(false);
        else showWarningToast("A download is in progress");
    };

    useEffect(() => {
        if (!showModal) return;

        setDownloadStatus(undefined);
        requestStopDownload.current = false;

        chosenQualities.current = new Map<number, string>();
        chosenLectures.current = new Map<number, boolean>();
        for (let lecture of course.lectures!) {
            let sortedsources = sortSources(lecture.media_sources).filter((q) => q.download_url);
            chosenLectures.current.set(lecture.id, sortedsources.length > 0);
            if (sortedsources.length > 0)
                chosenQualities.current.set(lecture.id, sourceToName(sortedsources[0]));
        }
        setForceReload(forceReload + 1);
        // eslint-disable-next-line react-hooks/exhaustive-deps
    }, [showModal, course.lectures]);

    let lectures = [];
    let allQualities: string[] = [];
    for (let lecture of course.lectures!) {
        let sortedsources = sortSources(lecture.media_sources).filter((q) => q.download_url);

        for (let source of lecture.media_sources ?? []) {
            let qName = source.quality.name;
            if (qName.length === 0) {
                qName = "<empty>";
            }
            if (!allQualities.includes(qName)) {
                allQualities.push(qName);
            }
        }

        let selectedQuality: media_source | undefined = undefined;
        if (chosenQualities.current.get(lecture.id)) {
            selectedQuality = lecture.media_sources?.find(
                (source) => sourceToName(source) === chosenQualities.current.get(lecture.id),
            );
        }

        let displayedProgress;
        if (downloadStatus !== undefined) {
            if (chosenLectures.current.get(lecture.id) === true) {
                // Lecture needs to be downloaded
                if (downloadStatus.get(lecture.id) === undefined) {
                    // Not yet started
                    displayedProgress = (
                        <div className="spinner-border" role="status">
                            <span className="visually-hidden">Loading...</span>
                        </div>
                    );
                } else {
                    // started
                    if (downloadStatus.get(lecture.id)?.error) {
                        displayedProgress = (
                            <span className="text-danger">
                                {downloadStatus.get(lecture.id)?.error}
                            </span>
                        );
                    } else if (downloadStatus.get(lecture.id)?.done) {
                        displayedProgress = <span className="text-success">Done</span>;
                    } else {
                        displayedProgress = (
                            <>
                                <div className="progress" style={{ width: "100px" }}>
                                    <div
                                        className="progress-bar"
                                        role="progressbar"
                                        style={{
                                            width: downloadStatus.get(lecture.id)?.percent + "%",
                                        }}
                                        aria-valuenow={downloadStatus.get(lecture.id)?.percent}
                                        aria-valuemin={0}
                                        aria-valuemax={100}
                                    />
                                </div>
                                <span>
                                    {Math.round(
                                        (downloadStatus.get(lecture.id)?.downloaded ?? 0) /
                                            1024 /
                                            1024,
                                    )}{" "}
                                    MB /{" "}
                                    {Math.round(
                                        (downloadStatus.get(lecture.id)?.total ?? 0) / 1024 / 1024,
                                    )}{" "}
                                    MB
                                </span>
                            </>
                        );
                    }
                }
            } else {
                // Lecture should not be downloaded
                displayedProgress = <></>;
            }
        } else {
            displayedProgress = (
                <a href={selectedQuality?.download_url} download>
                    <button
                        type="button"
                        className="btn btn-primary"
                        disabled={chosenQualities.current.get(lecture.id) === undefined}
                    >
                        {language.get("ui.download_manager.download_now")}
                    </button>
                </a>
            );
        }

        lectures.push(
            <tr key={lecture.id}>
                <td className="col-1">
                    <input
                        type="checkbox"
                        className="form-check-input"
                        checked={chosenLectures.current.get(lecture.id) ?? false}
                        disabled={downloadStarted}
                        onChange={(e) => {
                            let nextVal =
                                e.target.checked &&
                                (lecture.media_sources ?? []).some((q) => q.download_url);
                            chosenLectures.current.set(lecture.id, nextVal);
                            setForceReload(forceReload + 1);
                        }}
                    />
                </td>
                <td
                    className={
                        "col-7 make-span-overflow-scroll " +
                        (chosenLectures.current.get(lecture.id) ?? false ? "" : "text-secondary")
                    }
                >
                    <StylizedText>{lecture.title}</StylizedText>
                </td>
                <td className="col-2">
                    <select
                        className="form-select"
                        disabled={downloadStarted || sortedsources.length === 0}
                        value={chosenQualities.current.get(lecture.id)}
                        onChange={(e) => {
                            chosenQualities.current.set(lecture.id, e.target.value);
                            setForceReload(forceReload + 1);
                        }}
                    >
                        {sortedsources.map((source) => {
                            let sourceName = sourceToName(source);

                            return (
                                <option key={sourceName} value={sourceName}>
                                    {sourceName}
                                </option>
                            );
                        })}
                    </select>
                    {selectedQuality && `~${Math.round(selectedQuality.size / 1024 / 1024)}MB`}
                </td>
                <td className="col-2 text-center">{displayedProgress}</td>
            </tr>,
        );
    }
    allQualities = sortSourceNames(allQualities);

    const selectAll = () => {
        if (downloadStarted) return;
        for (let lecture of course.lectures!) {
            if (
                lecture.media_sources &&
                lecture.media_sources.some((source) => source.download_url)
            ) {
                chosenLectures.current.set(lecture.id, true);
            }
        }
        setForceReload(forceReload + 1);
    };
    const selectNone = () => {
        if (downloadStarted) return;
        for (let lecture of course.lectures!) {
            chosenLectures.current.set(lecture.id, false);
        }
        setForceReload(forceReload + 1);
    };

    const startDownloads = async () => {
        let newDownloadStatus = new Map<number, DownloadStatus>();
        setDownloadStatus(newDownloadStatus);

        const replaceStatus = (id: number, fn: (old?: DownloadStatus) => DownloadStatus) => {
            setDownloadStatus((oldStatus) => {
                let newStatus = new Map(oldStatus);
                newStatus.set(id, fn(oldStatus?.get(id)));
                return newStatus;
            });
        };

        const success = await downloadWithFileSystemAPI(
            replaceStatus,
            course,
            chosenLectures,
            chosenQualities,
            requestStopDownload,
        )
            .catch((e) => {
                console.log(e);
                return downloadWithOPFSAPI(
                    replaceStatus,
                    course,
                    chosenLectures,
                    chosenQualities,
                    requestStopDownload,
                );
            })
            .catch((e) => {
                console.log(e);
                return downloadWithDownloadAPI(
                    replaceStatus,
                    course,
                    chosenLectures,
                    chosenQualities,
                    requestStopDownload,
                );
            })
            .catch((e) => {
                console.log(e);
                return "error";
            });
        console.log(success);

        setDownloadStarted(false);
    };

    const toggleDownload = () => {
        if (downloadStarted) {
            requestStopDownload.current = true;
        } else {
            setDownloadStarted(true);
            requestStopDownload.current = false;
            startDownloads();
            requestStopDownload.current = false;
        }
    };

    const setAllQualities = (q: string) => {
        for (let lecture of course.lectures!) {
            if (lecture.media_sources) {
                let source = lecture.media_sources.find((s) => s.quality.name === q);
                if (source) {
                    chosenQualities.current.set(lecture.id, sourceToName(source));
                }
            }
        }
        setForceReload(forceReload + 1);
    };

    return (
        <>
            <button type="button" className="btn btn-secondary" onClick={() => setShowModal(true)}>
                <i className="bi bi-download me-2" />
                {language.get("ui.course.download_all_videos")}
            </button>
            <Modal show={showModal} size="xl" onHide={closeClicked}>
                <Modal.Header closeButton>
                    <Modal.Title>{language.get("ui.download_manager.title")}</Modal.Title>
                </Modal.Header>
                <Modal.Body>
                    <div className="mb-2">
                        <button
                            type="button"
                            className="btn btn-primary"
                            onClick={toggleDownload}
                            disabled={downloadStarted && requestStopDownload.current}
                        >
                            <i
                                className={
                                    "bi me-1 " +
                                    (downloadStarted ? "bi-pause-fill" : "bi-play-fill")
                                }
                            />
                            {downloadStarted
                                ? language.get("ui.download_manager.stop_download")
                                : language.get("ui.download_manager.start_download")}
                        </button>
                    </div>
                    <div className="mb-2 d-flex">
                        <button
                            type="button"
                            className="btn btn-secondary me-2"
                            onClick={selectAll}
                            disabled={downloadStarted}
                        >
                            {language.get("ui.download_manager.select_all")}
                        </button>

                        <button
                            type="button"
                            className="btn btn-secondary me-2"
                            onClick={selectNone}
                            disabled={downloadStarted}
                        >
                            {language.get("ui.download_manager.select_none")}
                        </button>

                        <DropdownButton
                            variant="primary"
                            title={language.get("ui.download_manager.quality")}
                            className=""
                        >
                            {allQualities.map((q) => (
                                <DropdownItem
                                    key={q}
                                    type="button"
                                    onClick={() => setAllQualities(q)}
                                >
                                    {q}
                                </DropdownItem>
                            ))}
                        </DropdownButton>
                    </div>

                    <table
                        className="table table-bordered table-responsive-sm"
                        style={{ tableLayout: "fixed" }}
                    >
                        <tbody>{lectures}</tbody>
                    </table>
                </Modal.Body>
                <Modal.Footer>
                    <button
                        className="btn btn-secondary"
                        onClick={closeClicked}
                        disabled={downloadStarted}
                    >
                        {language.get("ui.download_manager.close")}
                    </button>
                </Modal.Footer>
            </Modal>
        </>
    );
}
