import { createContext, useContext, useEffect, useRef, useState } from "react";

import type React from "react";

const EditModeContext = createContext({
    editMode: false,
    setEditMode: (set: boolean | ((i: boolean) => boolean)) => {},
    lock: () => {},
    unlock: () => {},
});

export function EditModeProvider({ children }: { children: React.ReactNode }) {
    const [editMode, setEditMode] = useState(false);
    const lockCount = useRef(0);

    return (
        <EditModeContext.Provider
            value={{
                editMode,
                setEditMode: (set: boolean | ((i: boolean) => boolean)) => {
                    if (lockCount.current > 0) {
                        return;
                    }
                    if (typeof set === "function") {
                        setEditMode((prev) => set(prev));
                    } else {
                        setEditMode(set);
                    }
                },
                lock: () => {
                    lockCount.current++;
                },
                unlock: () => {
                    lockCount.current--;
                    if (lockCount.current < 0) {
                        throw new Error("Unlock called too many times");
                    }
                },
            }}
        >
            {children}
        </EditModeContext.Provider>
    );
}

export function useEditMode() {
    return useContext(EditModeContext);
}

export function LockEditMode({ children }: { children?: React.ReactNode }) {
    const { lock, unlock } = useEditMode();

    useEffect(() => {
        lock();
        return unlock;
    }, [lock, unlock]);

    return children;
}
