import { useBackendContext } from "@/components/BackendProvider";
import React, { useEffect, useState, useRef, ReactNode } from "react";
import { useSearchParams } from "next/navigation";
import { useRouter } from "next/router";
import { useDebounceWithArgument } from "@/misc/PromiseHelpers";
import { ChooserEditor } from "@/components/TypeEditor";
import { mapMap } from "@/misc/Util";

import type { int } from "@/api/api_v1_types";

export function useFilteredDataView<D>(
    queryData: (filters: Map<string, any>) => Promise<D & { page_count: int }>,
): {
    filters: Map<string, any>;
    setFilter: (
        key: string,
        new_value: any | undefined,
        isValid: boolean,
        updateNow: boolean,
    ) => void;
    reloadData: () => void;
    isLoading: boolean;
    data: (D & { page_count: int }) | undefined;
    error: any;
} {
    const api = useBackendContext();
    const [isLoading, setIsLoading] = useState(true);
    const [data, setData] = useState<D & { page_count: int }>();
    const [error, setError] = useState<any>();
    // Holds the current values (updated immediately while typing)
    const [filters, setFilters] = useState<Map<string, [any, boolean]>>(
        new Map([
            ["page", [0, true]],
            ["entries_per_page", [50, true]],
        ]),
    );
    //Holds the filters to apply (change => reload data)
    const [activeFilters, setActiveFilters] = useState<Map<string, any>>(
        new Map([
            ["page", 0],
            ["entries_per_page", 50],
        ]),
    );
    const searchParams = useSearchParams();
    const knownParamsRef = useRef<string>();
    const router = useRouter();
    const page_id = parseInt(activeFilters.get("page") ?? "0");

    const [debounceUpdateDeferred, debounceUpdateNow] = useDebounceWithArgument<
        Map<string, [any, boolean]>,
        void
    >((newFilters) => {
        const newParams = new URLSearchParams();
        const newActiveFilters = new Map();
        let areAllValid = true;
        newFilters.forEach((tup, key) => {
            const [value, isValid] = tup;
            if (!isValid) areAllValid = false;
            newParams.set(key, JSON.stringify(value));
            newActiveFilters.set(key, value);
        });
        if (!areAllValid) return;

        setActiveFilters(newActiveFilters);
        const newParamsString = newParams.toString();
        router.push({ search: newParamsString });
        knownParamsRef.current = newParamsString;
    }, 500);

    const setFilter = (
        key: string,
        newValue: any | undefined,
        isValid: boolean,
        updateNow: boolean,
    ) =>
        setFilters((prev) => {
            const newFilters = new Map(prev);
            if (newValue !== undefined) {
                newFilters.set(key, [newValue, isValid]);
            } else {
                newFilters.delete(key);
            }
            if (updateNow) {
                debounceUpdateNow(newFilters);
            } else {
                debounceUpdateDeferred(newFilters);
            }
            return newFilters;
        });

    const reloadData = () => {
        if (page_id < 0) {
            setFilter("page", 0, true, true);
            return;
        }
        setIsLoading(true);
        queryData(activeFilters)
            .then((data) => {
                setData(data);
                setError(undefined);
                if (page_id >= data.page_count && data.page_count > 0) {
                    setFilter("page", 0, true, true);
                }
            })
            .catch((e) => {
                setData(undefined);
                setError(e);
            })
            .then(() => setIsLoading(false));
    };

    // eslint-disable-next-line react-hooks/exhaustive-deps
    useEffect(reloadData, [api, activeFilters]);

    useEffect(() => {
        if (searchParams.toString() === knownParamsRef.current) return;
        const newFilters = new Map();
        let foundFilter = false;
        searchParams.forEach((value, key) => {
            try {
                const parsedValue = JSON.parse(value);
                if (parsedValue !== undefined) {
                    newFilters.set(key, [parsedValue, true]); //TODO An API error might be thrown if values are invalid
                    foundFilter = true;
                }
            } catch (ignored) {
                // eslint-disable-next-line no-empty
            }
        });
        if (foundFilter) {
            //searchParams seem to be empty on first render
            setFilters(newFilters);
            debounceUpdateNow(newFilters);
        }
    }, [searchParams]); // eslint-disable-line react-hooks/exhaustive-deps

    return {
        filters: mapMap(filters, (val) => val[0]),
        setFilter: setFilter,
        reloadData: reloadData,
        isLoading: isLoading,
        data: data,
        error: error,
    };
}

type AdditionalProps = {
    value?: any;
    updateValue: (newValue: any, isValid: boolean, affectNow: boolean) => void;
};

export function FilterInput({
    component,
    label,
    filterId,
    setFilter,
    filters,
}: {
    component: (props: AdditionalProps) => ReactNode;
    label: string;
    filterId: string;
    setFilter: (
        filterId: string,
        newValue: any | undefined,
        isValid: boolean,
        updateNow: boolean,
    ) => void;
    filters: Map<string, any>;
}) {
    const addArgs: AdditionalProps = {
        value: filters.get(filterId),
        updateValue: (newValue: any, isValid: boolean, affectNow: boolean) =>
            setFilter(filterId, newValue, isValid, affectNow),
    };

    return (
        <div className="mb-2 me-3 d-flex">
            <label className="m-2">{label}</label>
            {component(addArgs)}
        </div>
    );
}

export function PagingNavigation({
    setFilter,
    filters,
    pageCount,
}: {
    setFilter: (
        filterId: string,
        newValue: any | undefined,
        isValid: boolean,
        updateNow: boolean,
    ) => void;
    filters: Map<string, any>;
    pageCount: int | undefined;
}) {
    const curPage = parseInt(filters.get("page") ?? "0");

    let displayPages = [];
    for (let i = Math.max(0, curPage - 5); i < Math.min(pageCount ?? 0, curPage + 5); i++) {
        if (i === curPage) {
            displayPages.push(
                <li className="page-item active" key={i}>
                    <button
                        type="button"
                        className="page-link"
                        onClick={() => setFilter("page", i, true, true)}
                    >
                        {i}
                    </button>
                </li>,
            );
        } else {
            displayPages.push(
                <li className="page-item" key={i}>
                    <button
                        type="button"
                        className="page-link"
                        onClick={() => setFilter("page", i, true, true)}
                    >
                        {i}
                    </button>
                </li>,
            );
        }
    }

    return (
        <>
            <FilterInput
                component={(props) => (
                    <ChooserEditor possibleValues={[10, 50, 100, 250, 500]} {...props} />
                )}
                label="Einträge pro Seite"
                filterId="entries_per_page"
                setFilter={setFilter}
                filters={filters}
            />
            <ul className="pagination mb-0">
                <li className="page-item">
                    <button
                        className="page-link"
                        type="button"
                        onClick={() => setFilter("page", 0, true, true)}
                        disabled={curPage === 0}
                    >
                        <span aria-hidden="true">&laquo;&laquo;</span>
                        <span className="visually-hidden">First</span>
                    </button>
                </li>
                <li className="page-item">
                    <button
                        className="page-link"
                        type="button"
                        onClick={() => setFilter("page", curPage - 1, true, true)}
                        disabled={curPage === 0}
                    >
                        <span aria-hidden="true">&laquo;</span>
                        <span className="visually-hidden">Previous</span>
                    </button>
                </li>
                {displayPages}
                <li className="page-item">
                    <button
                        className="page-link"
                        type="button"
                        onClick={() => setFilter("page", curPage + 1, true, true)}
                        disabled={pageCount === undefined || curPage + 1 >= pageCount}
                    >
                        <span aria-hidden="true">&raquo;</span>
                        <span className="visually-hidden">Next</span>
                    </button>
                </li>
                <li className="page-item">
                    <button
                        className="page-link"
                        type="button"
                        onClick={() => setFilter("page", (pageCount ?? 1) - 1, true, true)}
                        disabled={pageCount === undefined || curPage === pageCount - 1}
                    >
                        <span aria-hidden="true">&raquo;&raquo;</span>
                        <span className="visually-hidden">Last</span>
                    </button>
                </li>
            </ul>
        </>
    );
}
