import { useEffect, useState, useRef } from "react";
import { useBackendContext } from "./BackendProvider";
import Modal from "react-bootstrap/Modal";
import { useReloadBoundary } from "./ReloadBoundary";
import { useDebounceUpdateData } from "@/misc/PromiseHelpers";
import { LockEditMode, useEditMode } from "./EditModeProvider";
import { datetimeToString, semesterToHuman } from "@/misc/Formatting";
import { useLanguage } from "./LanguageProvider";
import { basePath } from "../../basepath";
import { showError, showErrorToast } from "@/misc/ErrorHandlers";
import { deepEquals, TooltipButton } from "@/misc/Util";
import { createOMFieldEditor } from "@/components/TypeEditor";

import type React from "react";
import type {
    GetConfigurationResponse,
    GetNewConfigurationResponse,
    field_description,
    field_value,
    int,
} from "@/api/api_v1_types";
import { StylizedText } from "./StylizedText";
import StopNavigation from "./StopNavigation";

const markdownSupportingFields = [
    "announcement.text",
    "featured.title",
    "featured.text",
    "course.organizer",
    "course.description",
    "course.authentication_information",
    "lecture.speaker",
    "lecture.description",
];

export function ListOMFieldComponent({
    object_type,
    object_id,
    field_desc,
    initialValue,
    currentValue,
    updateCurrentValue,
    inCreation = false,
}: {
    object_type: string;
    object_id?: int;
    field_desc: field_description;
    initialValue?: any;
    currentValue?: any;
    updateCurrentValue: (newValue: any, isValid: boolean) => void;
    inCreation?: boolean;
}) {
    const { language } = useLanguage();

    useEffect(() => {
        if (inCreation && currentValue === undefined && field_desc.default_value !== undefined)
            updateCurrentValue(field_desc.default_value, true);
    }, [inCreation, field_desc, currentValue, updateCurrentValue]);

    let hasChanged = !inCreation && !deepEquals(initialValue, currentValue);
    const editor = createOMFieldEditor(field_desc, {
        value: currentValue,
        updateValue: (val, isValid, affectNow) => updateCurrentValue(val, isValid),
        disabled: false,
        autoFocus: false,
    });

    return (
        <tr>
            <td>
                <div className="d-flex">
                    <div
                        className={"vr me-1 " + (hasChanged ? "opacity-25" : "opacity-0")}
                        style={{ color: "orange", width: "4px" }}
                    />
                    <div className="me-2">
                        {`${language.get(`object.${object_type}.${field_desc.id}`)}: `}
                        {!inCreation && (
                            <OMHistory
                                object_id={object_id!}
                                object_type={object_type}
                                field_id={field_desc.id}
                                className="btn-sm py-0"
                            />
                        )}
                        {language.has(`object.${object_type}.${field_desc.id}.tooltip`) && (
                            <TooltipButton>
                                {language.getStyled(
                                    `object.${object_type}.${field_desc.id}.tooltip`,
                                    true,
                                )}
                            </TooltipButton>
                        )}
                        {markdownSupportingFields.includes(`${object_type}.${field_desc.id}`) && (
                            <TooltipButton iconClassName="bi-markdown">
                                {language.getStyled("ui.object.field_supports_markdown", true)}
                            </TooltipButton>
                        )}
                    </div>
                </div>
            </td>
            <td className="align-middle">
                <div className="w-100 d-inline-grid">{editor}</div>
            </td>
        </tr>
    );
}

export function EmbeddedOMFieldComponent({
    object_type,
    object_id,
    field_id,
    field_type,
    initialValue,
    allowMarkdown = false,
    inline = false,
    className = "",
}: {
    object_type: string;
    object_id: int;
    field_id: string;
    field_type: "boolean" | "string" | "datetime" | "semester_string";
    initialValue: any;
    allowMarkdown?: boolean;
    inline?: boolean;
    className?: string;
}) {
    const api = useBackendContext();
    const reloadFunc = useReloadBoundary();
    const { language } = useLanguage();

    const [value, setValue] = useState(initialValue);
    const [isValid, setIsValid] = useState(true);
    const [isEditing, setIsEditing] = useState(false);
    const [field, setField] = useState<field_value>();
    // We need an additional references since a state is not updated immediately and sometimes the API call would fail
    // due to an unexpected current value. (Only using a ref does not work, since the rendering needs this to indicate
    // if changes have been made)
    const serverSideFieldValueRef = useRef<field_value>(); // This is what we expect the value to be on the server
    const updatedSinceLastReloadRef = useRef<boolean>(false);
    const beforeEditingValue = useRef();
    const { editMode } = useEditMode();

    // These types can be edited without loading the configuration or starting editing (e.g. isEditing is ignored)
    // The description of these types may have no additional fields (only id, type, default_value)
    const isInstantEdit = ["boolean"].includes(field_type);

    useEffect(() => {
        if (isInstantEdit) {
            setField({
                description: {
                    id: field_id,
                    type: field_type,
                    //Default value does not matter
                },
                value: initialValue,
            });
        }

        serverSideFieldValueRef.current = initialValue;
        setValue(initialValue);
        // eslint-disable-next-line react-hooks/exhaustive-deps
    }, [isInstantEdit, field_id, field_type, initialValue]);

    useEffect(() => {
        if (!isEditing) return;

        // TODO: disable editing while config is still loading, this is especially annoying with high latency
        api.getOMConfiguration(object_type, object_id)
            .then((config) => {
                const field = config.fields?.find((field) => field.description.id === field_id);
                if (field === undefined) {
                    showErrorToast("Cannot find this field's configuration");
                    setIsEditing(false);
                    return;
                }
                setField(field);
                serverSideFieldValueRef.current = field.value;
                setValue(field.value);
            })
            .catch((error) => {
                showError(error, "Unable to load configuration");
                setIsEditing(false);
            });
    }, [api, isEditing, object_type, object_id, field_id]);

    const [deferredSendValueUpdate, nowSendValueUpdate] = useDebounceUpdateData<any, void>(
        async (newValue: any) => {
            if (field === undefined) return;
            if (deepEquals(newValue, serverSideFieldValueRef.current)) return;

            return api
                .updateOMObject(object_type, object_id, {
                    updates: { [field_id]: newValue },
                    expected_current_values: { [field_id]: serverSideFieldValueRef.current },
                })
                .then(() => {
                    serverSideFieldValueRef.current = newValue;
                    setField(field === undefined ? undefined : { ...field, value: newValue });
                    if (isInstantEdit) {
                        reloadFunc();
                    } else {
                        updatedSinceLastReloadRef.current = true;
                    }
                })
                .catch((error) => {
                    showError(error, "Unable to update field");
                });
        },
        isInstantEdit ? 10 : 1000 * 15,
    );

    const disableEditing = () => {
        if (!isValid) return;
        if (!isEditing) return;
        nowSendValueUpdate(value).then(() => {
            setIsEditing(false);
            if (updatedSinceLastReloadRef.current) {
                reloadFunc();
                updatedSinceLastReloadRef.current = false;
            }
        });
    };

    const resetValue = () => {
        if (!isEditing) return;
        setValue(beforeEditingValue.current);
        setIsValid(true);
        deferredSendValueUpdate(beforeEditingValue.current);
    };

    useEffect(() => {
        if (editMode === false && isEditing === true) disableEditing();
        // eslint-disable-next-line react-hooks/exhaustive-deps
    }, [editMode, isEditing]);

    if (!isInstantEdit && (!isEditing || field === undefined)) {
        let formatted: any = "Value of unknown type: " + value;
        switch (field_type) {
            case "string":
                formatted = <StylizedText markdown={allowMarkdown}>{value}</StylizedText>;
                break;
            case "datetime":
                formatted = datetimeToString(value);
                break;
            case "semester_string":
                formatted = semesterToHuman(value, true);
                break;
        }
        return (
            <span
                className={
                    className +
                    (inline ? " p-children-inline" : "") +
                    " disable-last-paragraph-spacing"
                }
                onDoubleClick={() => {
                    if (!editMode) return;
                    beforeEditingValue.current = value;
                    setIsEditing(true);
                }}
            >
                {formatted}
            </span>
        );
    }

    let description: field_description = {
        id: field_id,
        type: field_type,
    };
    if (!isInstantEdit) {
        description = field!.description;
    }
    let hasChanged = !deepEquals(field?.value ?? initialValue, value);
    const editor = createOMFieldEditor(description, {
        value: value,
        updateValue: (newValue: any, isValid: boolean, affectNow: boolean) => {
            setValue(newValue);
            setIsValid(isValid);
            if (isValid) {
                if (affectNow) {
                    nowSendValueUpdate(newValue);
                } else {
                    deferredSendValueUpdate(newValue);
                }
            }
        },
        disabled: false,
        autoFocus: !isInstantEdit,
    });

    let title = language.getOrDefault(`object.${object_type}.${field_id}`, "");
    const tooltipKey = `object.${object_type}.${field_id}.tooltip`;
    if (language.has(tooltipKey)) {
        title = title + ":\n" + language.get(tooltipKey); //TODO This does not work with markdown
    }

    return (
        <span
            className={
                "h-auto d-inline-flex align-middle " + className + ` ${isEditing ? "w-100" : ""}`
            }
            title={title}
            onBlur={disableEditing}
            onSubmit={disableEditing}
        >
            {isEditing && hasChanged && (
                <StopNavigation warningText={language.get("ui.object.unsaved_changes")} />
            )}
            <div
                className={"vr mx-1 " + (hasChanged ? "opacity-25" : "opacity-0")}
                style={{ color: "orange", width: "4px", verticalAlign: "baseline" }}
            />
            {editor}
            {!isInstantEdit && (
                <button
                    type="button"
                    className={"btn btn-outline-danger m-1"}
                    onClick={resetValue}
                    onMouseDown={(e) => e.preventDefault()}
                    title="Reset"
                >
                    <i className="bi bi-arrow-counterclockwise" />
                </button>
            )}
        </span>
    );
}

export function OMEdit({
    object_type,
    object_id,
    ...props
}: {
    object_type: string;
    object_id: int;
    [key: string]: any;
}) {
    const api = useBackendContext();
    const reloadFunc = useReloadBoundary();
    const { language } = useLanguage();
    const [showModal, setShowModal] = useState(false);
    const [isLoading, setIsLoading] = useState(false);
    const [config, setConfig] = useState<GetConfigurationResponse>();
    const [currentValues, setCurrentValues] = useState<{ [id: string]: [any, boolean] }>({});

    useEffect(() => {
        if (showModal === false) return;
        setIsLoading(true);
        api.getOMConfiguration(object_type, object_id)
            .then((config) => {
                const values: { [id: string]: [any, boolean] } = {};
                for (let field of config.fields) {
                    values[field.description.id] = [field.value, true];
                }
                setConfig(config);
                setCurrentValues(values);
            })
            .catch((err) => {
                showError(err, "Unable to load configuration");
                setShowModal(false);
            })
            .then(() => {
                setIsLoading(false);
            });
    }, [api, object_type, object_id, showModal]);

    const updates: { [id: string]: any } = {};
    const expectedValues: { [id: string]: any } = {};
    let isValid = true;
    for (let field of config?.fields ?? []) {
        const currentValue = currentValues[field.description.id]; //Never undefined, since tuple is set after config loading
        if (!currentValue[1]) isValid = false;
        if (deepEquals(currentValue[0], field.value)) continue;
        updates[field.description.id] = currentValue[0];
        expectedValues[field.description.id] = field.value;
    }
    const hasAnyChanges = Object.keys(updates).length > 0;

    const saveChanges = () => {
        if (!hasAnyChanges || !isValid) return;

        api.updateOMObject(object_type, object_id, {
            updates: updates,
            expected_current_values: expectedValues,
        })
            .then(() => {
                setShowModal(false);
                setCurrentValues({});
                setConfig(undefined);
                reloadFunc();
            })
            .catch((err) => {
                showError(err, "Unable to update object");
            });
    };

    const closeClicked = () => {
        if (!hasAnyChanges || confirm("Do you want to discard your changes?")) {
            setShowModal(false);
            setCurrentValues({});
            setConfig(undefined);
        }
    };

    const { className, ...otherProps } = props;
    return (
        <>
            <button
                type="button"
                className={"btn btn-outline-primary m-1 " + (className ?? "")}
                onClick={() => setShowModal(true)}
                title="Edit"
                {...otherProps}
            >
                <i className="bi bi-pencil" />
            </button>

            <Modal show={showModal} size="xl" onHide={closeClicked}>
                <LockEditMode />
                <Modal.Header closeButton>
                    <Modal.Title>Konfiguration</Modal.Title>
                </Modal.Header>
                <Modal.Body>
                    {isLoading && (
                        <>
                            <div className="spinner-border me-2" role="status" />
                            Lade Konfiguration
                        </>
                    )}
                    {config?.fields !== undefined && (
                        <>
                            <div className="alert alert-warning">
                                {/* TODO: remove after release */}
                                Achtung: Die alte Webseite kann kein Markdown rendern, nur HTML.
                                Wenn hier also Markdown verwendet wird, wird dies auf der alten
                                Seite nicht richtig gerendert. Während der Transition werden aber
                                auf der neuen Seite die folgenden HTML-Tags gerendert: a, b, br, p
                                und strong (Welche auf der alten Seite auch korrekt angezeigt
                                werden)
                            </div>
                            <small className="text-muted">
                                Tipp: Viele Felder können im Edit-Mode auch direkt durch einen
                                Doppelklick bearbeitet werden
                            </small>
                            <table className="table">
                                <thead>
                                    <tr>
                                        <th></th>
                                        <th className="col-8"></th>
                                    </tr>
                                </thead>
                                <tbody>
                                    {config!.fields?.map((field) => (
                                        <ListOMFieldComponent
                                            key={field.description.id}
                                            object_type={object_type}
                                            object_id={object_id}
                                            field_desc={field.description}
                                            initialValue={field.value}
                                            currentValue={currentValues[field.description.id][0]}
                                            updateCurrentValue={(newValue: any, isValid: boolean) =>
                                                setCurrentValues((prev) => ({
                                                    ...prev,
                                                    [field.description.id]: [newValue, isValid],
                                                }))
                                            }
                                        />
                                    ))}
                                </tbody>
                            </table>
                        </>
                    )}
                </Modal.Body>
                <Modal.Footer>
                    <button className="btn btn-secondary" onClick={closeClicked}>
                        {language.get("ui.generic.cancel")}
                    </button>
                    <button
                        className="btn btn-primary"
                        onClick={saveChanges}
                        disabled={!hasAnyChanges || !isValid}
                    >
                        {language.get("ui.generic.save_changes")}
                    </button>
                </Modal.Footer>
            </Modal>
        </>
    );
}

export function OMCreate({
    object_type,
    parent_type,
    parent_id,
    children,
}: {
    object_type: string;
    parent_type?: string;
    parent_id?: int;
    children: React.ReactNode;
}) {
    const api = useBackendContext();
    const reloadFunc = useReloadBoundary();
    const { language } = useLanguage();
    const [showModal, setShowModal] = useState(false);
    const [isLoading, setIsLoading] = useState(false);
    const [config, setConfig] = useState<GetNewConfigurationResponse>();
    const [currentValues, setCurrentValues] = useState<{ [id: string]: [any, boolean] }>({});
    const [chosenVariant, setChosenVariant] = useState<string>();

    const variants = config?.variant_fields ? Object.keys(config.variant_fields) : [];

    const createValues: { [id: string]: any } = {};
    let isValid = true;
    let hasUnsavedData = false;
    let effectiveFields = config?.fields ?? [];
    if (chosenVariant !== undefined && variants.includes(chosenVariant)) {
        effectiveFields = effectiveFields.concat(config!.variant_fields[chosenVariant]);
    } else if (variants.length > 0) {
        isValid = false;
    }
    for (let field_desc of effectiveFields) {
        const [currentValue, isValueValid] = currentValues[field_desc.id]; // Never undefined, since tuple is set after config loading
        const isBasicValue =
            currentValue === "" || currentValue === undefined || currentValue === null;
        if (!isValueValid) {
            if (!isBasicValue) hasUnsavedData = true;
            isValid = false;
            break;
        }
        if (deepEquals(currentValue, field_desc.default_value)) continue;
        createValues[field_desc.id] = currentValue;
        if (!isBasicValue) hasUnsavedData = true;
    }

    const createObject = () => {
        if (!isValid) return;

        api.createOMObject(object_type, {
            parent_type: parent_type,
            parent_id: parent_id,
            values: createValues,
            variant: chosenVariant,
        })
            .then(() => {
                setShowModal(false);
                reloadFunc();
            })
            .catch((err) => {
                showError(err, "Unable to create object");
            });
    };

    useEffect(() => {
        if (showModal === false) {
            setCurrentValues({});
            setChosenVariant(undefined);
            setConfig(undefined);
            return;
        }
        if (config !== undefined) return;
        setIsLoading(true);
        api.getNewOMConfiguration(object_type)
            .then(
                (config) => {
                    const values: { [id: string]: [any, boolean] } = {};
                    for (let field_desc of config.fields) {
                        values[field_desc.id] = [
                            field_desc.default_value,
                            field_desc.default_value !== undefined,
                        ];
                    }
                    if (
                        config.variant_fields !== undefined &&
                        Object.keys(config.variant_fields).length > 0
                    ) {
                        for (let variant of Object.keys(config.variant_fields)) {
                            const var_fields = config.variant_fields[variant];
                            for (let field_desc of var_fields) {
                                values[field_desc.id] = [
                                    field_desc.default_value,
                                    field_desc.default_value !== undefined,
                                ];
                            }
                        }
                        setChosenVariant(Object.keys(config.variant_fields)[0]);
                    }
                    setConfig(config);
                    setCurrentValues(values);
                },
                (error) => {
                    setShowModal(false);
                    showError(error, "Unable to load creation configuration");
                },
            )
            .then(() => setIsLoading(false));
    }, [api, config, showModal, object_type]);

    const onClose = () => {
        if (!hasUnsavedData || confirm("Do you want to discard your changes?")) {
            setShowModal(false);
        }
    };
    return (
        <>
            <div onClick={() => setShowModal(true)}>{children}</div>

            <Modal show={showModal} size="xl" onHide={onClose}>
                <LockEditMode />
                <Modal.Header closeButton>
                    <Modal.Title>Neues Objekt</Modal.Title>
                </Modal.Header>
                <Modal.Body>
                    {isLoading && (
                        <>
                            <div className="spinner-border me-2" role="status" />
                            Lade Konfiguration
                        </>
                    )}
                    {config !== undefined && config.fields.length > 0 && (
                        <>
                            <h5>Felder</h5>
                            <table className="table w-100">
                                <thead>
                                    <tr>
                                        <th></th>
                                        <th className="col-8"></th>
                                    </tr>
                                </thead>
                                <tbody>
                                    {config.fields.map((field_desc) => (
                                        <ListOMFieldComponent
                                            key={field_desc.id}
                                            object_type={object_type}
                                            field_desc={field_desc}
                                            currentValue={currentValues[field_desc.id][0]}
                                            updateCurrentValue={(newValue: any, isValid: boolean) =>
                                                setCurrentValues((prev) => ({
                                                    ...prev,
                                                    [field_desc.id]: [newValue, isValid],
                                                }))
                                            }
                                            inCreation={true}
                                        />
                                    ))}
                                </tbody>
                            </table>
                        </>
                    )}
                    {variants.length > 0 && (
                        <div>
                            <div>Variante</div>
                            <select
                                className="form-select"
                                onChange={(e) => {
                                    setChosenVariant(e.target.value);
                                }}
                            >
                                {variants.map((variant) => (
                                    <option key={variant} value={variant}>
                                        {variant}
                                    </option>
                                ))}
                            </select>
                        </div>
                    )}
                    {chosenVariant &&
                        config?.variant_fields![chosenVariant].map((field_desc) => (
                            <ListOMFieldComponent
                                key={field_desc.id}
                                object_type={object_type}
                                field_desc={field_desc}
                                currentValue={currentValues[field_desc.id][0]}
                                updateCurrentValue={(newValue: any, isValid: boolean) =>
                                    setCurrentValues((prev) => ({
                                        ...prev,
                                        [field_desc.id]: [newValue, isValid],
                                    }))
                                }
                                inCreation={true}
                            />
                        ))}
                </Modal.Body>
                <Modal.Footer>
                    <button className="btn btn-secondary" onClick={onClose}>
                        {language.get("ui.generic.cancel")}
                    </button>
                    <button className="btn btn-primary" onClick={createObject} disabled={!isValid}>
                        {language.get("ui.generic.create")}
                    </button>
                </Modal.Footer>
            </Modal>
        </>
    );
}

export function OMDelete({
    object_type,
    object_id,
    className,
}: {
    object_type: string;
    object_id: int;
    className?: string;
}) {
    const api = useBackendContext();
    const reloadFunc = useReloadBoundary();

    const deleteObject = () => {
        if (confirm("Really delete this?")) {
            api.deleteOMObject(object_type, object_id)
                .then(reloadFunc)
                .catch((err) => {
                    showError(err, "Unable to delete object");
                });
        }
    };

    return (
        <button
            type="button"
            className={"btn btn-outline-danger m-1 " + (className ?? "")}
            onClick={deleteObject}
            title="Delete"
        >
            <i className="bi bi-trash-fill" />
        </button>
    );
}

export function OMHistory({
    object_type,
    object_id,
    field_id,
    className,
}: {
    object_type: string;
    object_id: int;
    field_id?: string;
    className?: string;
}) {
    return (
        <button
            type="button"
            className={"btn " + (className ?? "")}
            onClick={() => {
                // Open new tab at /internal/changelog
                if (field_id) {
                    window.open(
                        `${basePath}/internal/changelog?object_type="${object_type}"&object_id=${object_id}&field="${field_id}"`,
                        "_blank",
                    );
                } else {
                    window.open(
                        `${basePath}/internal/changelog?object_type="${object_type}"&object_id=${object_id}`,
                        "_blank",
                    );
                }
            }}
        >
            <i className="bi bi-clock-history" title="History" />
        </button>
    );
}
