import { MouseEvent, useEffect, useRef, useState } from "react";
import "video.js/dist/video-js.min.css";
import "@silvermine/videojs-quality-selector/dist/css/quality-selector.css";
import Link from "next/link";
import { useBackendContext } from "./BackendProvider";
import { filesizeToHuman, timestampToString } from "@/misc/Formatting";
import OverlayTrigger from "react-bootstrap/OverlayTrigger";
import Popover from "react-bootstrap/Popover";

import { OMDelete, OMEdit, OMHistory, EmbeddedOMFieldComponent } from "./OMConfigComponent";
import { useEditMode } from "./EditModeProvider";
import { useUserContext } from "./UserDataProvider";
import { useReloadBoundary } from "./ReloadBoundary";
import { basePath } from "../../basepath";
import Title from "./TitleComponent";
import { showError } from "@/misc/ErrorHandlers";

import type React from "react";
import VideoJsPlayerType from "video.js/dist/types/player";
import type {
    course,
    GetCourseResponse,
    authentication_method,
    authentication_methods,
    chapter,
    media_source,
    lecture,
} from "@/api/api_v1_types";
import Dropdown from "react-bootstrap/Dropdown";
import { ResourceType } from "@/misc/PromiseHelpers";
import { PlayerData } from "@/pages/dynamic_player";
import { useLanguage } from "./LanguageProvider";
import { StylizedText } from "./StylizedText";
import { ApiError } from "@/api/ApiError";
import Head from "next/head";
import { VideoCard } from "./VideoCard";
import { urlForLecture } from "@/misc/Util";
import VideoJsMarkers from "@/misc/videojs-markers";

function initPlayer(
    player: VideoJsPlayerType,
    media_sources: media_source[],
    chapters?: chapter[],
) {
    if (!player.getChild("ControlBar")?.getChild("QualitySelector")) {
        player
            .getChild("ControlBar")!
            .addChild("QualitySelector", {}, player.getChild("ControlBar")!.children().length - 1);
    }
    let sources = media_sources.map((source) => {
        return {
            src: source.player_url,
            type: "video/mp4",
            label: source.quality.name,
            selected: false,
            __quality: source.quality,
        };
    });
    if (sources.length > 0) {
        // find highest priority source
        let bestSource = sources.reduce((prev, curr) => {
            if (curr.__quality.priority > prev.__quality.priority) return curr;

            return prev;
        });
        bestSource.selected = true;
    }
    // sort sources by quality, highest first
    sources.sort((a, b) => {
        let aq = parseInt(a.label.split("p")[0]);
        let bq = parseInt(b.label.split("p")[0]);
        return bq - aq;
    });

    player.src(sources);

    let chapterMarkers =
        chapters
            ?.filter((a) => a.is_visible === undefined || a.is_visible === true)
            .sort((a, b) => a.start_time - b.start_time) ?? [];
    let markers = [];
    for (let i = 0; i < chapterMarkers!.length; i++) {
        let c = chapterMarkers![i];

        markers.push({
            time: c.start_time,
            text: c.name,
            duration:
                i < chapterMarkers?.length - 1
                    ? chapterMarkers[i + 1].start_time - c.start_time
                    : -1,
        });
    }

    (player as any).markers.reset(markers);
}

function VideoPlayer({ lecture, className }: { lecture: lecture; className?: string }) {
    const videoRef = useRef<HTMLVideoElement>(null);

    useEffect(() => {
        import("videojs-hotkeys")
            .then(() => import("video.js"))
            .then(async (videojs) => {
                (await import("@silvermine/videojs-quality-selector")).default(videojs.default);

                const lang =
                    (navigator &&
                        ((navigator.languages && navigator.languages[0]) || navigator.language)) ||
                    "de";
                if (videoRef.current === null) {
                    console.error("VideoRef is null");
                    return;
                }

                const videojsMarkers = VideoJsMarkers(videojs.default);

                const player =
                    videojs.default.getPlayer(videoRef.current!) ??
                    videojs.default(videoRef.current!, {
                        language: lang,
                        userActions: { hotkeys: false },
                        playbackRates: [0.5, 0.75, 1, 1.25, 1.5, 1.75, 2, 2.25, 2.5, 3, 3.5, 4],
                        controlBar: {
                            pictureInPictureToggle: false,
                        },
                        plugins: {
                            hotkeys: {
                                seekStep: 15,
                                enableVolumeScroll: false,
                                alwaysCaptureHotkeys: true,
                            },
                        },
                        enableSmoothSeeking: true,
                        inactivityTimeout: 1000,
                    });

                if (
                    videojs.default.getPlugin("markers") === undefined ||
                    !Object.hasOwn(player, "markers")
                ) {
                    videojsMarkers(player);
                    (player as any).markers({
                        markerStyle: {
                            width: "8px",
                            height: "100%",
                            "margin-left": "-3px",
                            "border-radius": "0",
                            "background-color": "white",
                            opacity: "0.7",
                        },
                        markerTip: {
                            display: true,
                            text: function (marker: any) {
                                return marker.text;
                            },
                        },
                    });
                }

                initPlayer(player, lecture.media_sources!, lecture.chapters);
            });
    }, [lecture.media_sources, lecture.chapters]);

    return (
        <div className="w-100" style={{ aspectRatio: "16 / 9" }}>
            <video
                ref={videoRef}
                className={
                    "video-js w-100 vjs-big-play-centered vjs-fill vjs-rwth " + (className ?? "")
                }
                id="videoplayer"
                controls
                data-wasnotplayed="1"
                width={1080}
                height={720}
            />
        </div>
    );
}

function KapitelPopover({
    timeStr,
    closePop,
    lectureId,
}: {
    timeStr: string;
    closePop: () => void;
    lectureId: number;
}) {
    const userContext = useUserContext();
    const api = useBackendContext();
    const reloadFunc = useReloadBoundary();
    const [isSubmitting, setIsSubmitting] = useState(false);

    const handleSubmit = (e: React.FormEvent<HTMLFormElement>) => {
        e.preventDefault();
        if (isSubmitting) return;
        let form = e.currentTarget;
        let timeHuman = form.time.value; // Format HH:MM:SS
        let time = 0;
        let timeSplit = timeHuman.split(":");
        for (let i = 0; i < timeSplit.length; i++) {
            time +=
                Math.max(0, Math.min(60, parseInt(timeSplit[i]))) *
                Math.pow(60, timeSplit.length - i - 1);
        }

        let text = form.text.value;
        setIsSubmitting(true);
        if (userContext.canEditStuff()) {
            api.createOMObject("chapter", {
                parent_type: "lecture",
                parent_id: lectureId,
                values: {
                    start_time: time,
                    name: text,
                    is_visible: true,
                },
            })
                .then(() => {
                    closePop();
                    reloadFunc();
                })
                .catch((e) => {
                    showError(e, "Unable to create chapter");
                })
                .finally(() => setIsSubmitting(false));
        } else {
            api.suggestChapter(lectureId, {
                start_time: time,
                name: text,
            })
                .then(() => {
                    closePop();
                    reloadFunc();
                })
                .catch((e) => {
                    showError(e, "Unable to suggest chapter");
                })
                .finally(() => setIsSubmitting(false));
        }

        return false;
    };

    return (
        <form className="needs-validation" onSubmit={handleSubmit}>
            <input
                className="form-control"
                style={{ marginTop: "6px", marginBottom: "6px" }}
                placeholder="00:00"
                name="time"
                type="text"
                defaultValue={timeStr}
                pattern="(|[0-9]{1,2}:(|[0-9]{1,2}:))[0-9]{1,2}"
            />
            <input
                className="form-control"
                placeholder="Titel"
                name="text"
                type="text"
                style={{ marginBottom: "15px" }}
                required
            />
            <input
                type="submit"
                className="btn btn-primary"
                style={{ marginBottom: "6px" }}
                value={userContext.canEditStuff() ? "Hinzufügen" : "Vorschlagen"}
                disabled={isSubmitting}
            />
            {isSubmitting && <div className="spinner-border ms-2" role="status" />}
        </form>
    );
}

function AuthorizeHelper({
    course,
    lecture,
    authed_methods,
}: {
    course: GetCourseResponse;
    lecture: lecture;
    authed_methods: authentication_methods;
}) {
    const api = useBackendContext();
    const reloadFunc = useReloadBoundary();
    const [userPwErrorState, setUserPwErrorState] = useState<any>();
    const userContext = useUserContext();
    const hasUserInfo = userContext.hasUserInfo();
    const [isLoggingIn, setIsLoggingIn] = useState(false);
    const { language } = useLanguage();

    useEffect(() => {
        if (hasUserInfo === true) {
            // reload page when the user logs in
            reloadFunc();
        }
        // eslint-disable-next-line react-hooks/exhaustive-deps
    }, [hasUserInfo]);

    const startOauth = (method: authentication_method) => {
        if (method !== "moodle" && method !== "rwth") return;
        api.startAuthentication({ type: method })
            .then((res) => {
                const pop = window.open(res.verification_url, "_blank");
                if (pop) {
                    pop.focus();
                    let popInterval = setInterval(() => {
                        if (pop.closed) {
                            clearInterval(popInterval);
                            reloadFunc();
                        }
                    }, 100);
                }
            })
            .catch((e) => {
                showError(e, "Cannot start OAuth");
            });
    };

    const handleUserPass = (e: React.FormEvent<HTMLFormElement>) => {
        e.preventDefault();
        let form = e.currentTarget;
        let user = form.username.value;
        let pass = form.password.value;
        setIsLoggingIn(true);
        api.authenticatePassword({ username: user, password: pass, lecture_id: lecture.id })
            .then(reloadFunc)
            .catch(setUserPwErrorState)
            .finally(() => setIsLoggingIn(false));
        return false;
    };

    let errorElement = <></>;
    if (userPwErrorState) {
        if (
            userPwErrorState instanceof ApiError &&
            userPwErrorState.error_code === "authentication_failed"
        ) {
            errorElement = (
                <p className="alert alert-warning">
                    {language.get("ui.video_player.login_user_password.user_or_pass_incorrect")}
                </p>
            );
        } else errorElement = <p className="alert alert-warning">{userPwErrorState + ""}</p>;
    } else if (authed_methods.includes("password")) {
        errorElement = (
            <p className="alert alert-warning">
                {language.get("ui.video_player.login_user_password.current_password_incorrect")}
            </p>
        );
    }

    return (
        <div
            className="col-12 rounded-2 h-100"
            style={{ padding: "10px", backgroundColor: "rgba(0, 0, 0, 0.3)" }}
        >
            <h3 className="text-center my-2">{language.get("ui.video_player.login_required")}</h3>
            {course.authentication_information !== undefined && (
                <p className="text-center">
                    <StylizedText markdown>{course.authentication_information}</StylizedText>
                </p>
            )}
            <div style={{ paddingBottom: "20px" }} className="container-fluid">
                <div className="row">
                    {lecture.authentication_methods.includes("password") && (
                        <div className="col-sm-4">
                            <h4 className="text-center">
                                <span className="bi bi-lock" aria-hidden="true" />{" "}
                                {language.get("ui.video_player.login_user_password.description")}
                            </h4>
                            {errorElement}
                            <form onSubmit={handleUserPass}>
                                <div className="mb-3">
                                    <label htmlFor="exampleInputEmail1" className="form-label">
                                        {language.get("object.permission.username")}
                                    </label>
                                    <input
                                        type="text"
                                        className="form-control"
                                        id="username"
                                        name="username"
                                        placeholder=""
                                    />
                                </div>
                                <div className="mb-3">
                                    <label htmlFor="exampleInputPassword1" className="form-label">
                                        {language.get("object.permission.password")}
                                    </label>
                                    <input
                                        type="password"
                                        className="form-control"
                                        id="password"
                                        name="password"
                                        placeholder=""
                                    />
                                </div>
                                <button
                                    type="submit"
                                    className="btn btn-secondary"
                                    disabled={isLoggingIn}
                                >
                                    {language.get("ui.video_player.login")}
                                </button>
                                {isLoggingIn && (
                                    <div
                                        className="ms-2 spinner-border text-primary"
                                        role="status"
                                    />
                                )}
                            </form>
                        </div>
                    )}
                    {lecture.authentication_methods.includes("rwth") && (
                        <div className="col-sm-4">
                            <h4 className="text-center">
                                <span className="bi bi-person-fill" aria-hidden="true" /> RWTH
                            </h4>
                            <p>{language.get("ui.video_player.login_rwth.description")}</p>
                            <button
                                type="button"
                                onClick={() => startOauth("rwth")}
                                className="btn btn-secondary"
                            >
                                {language.get("ui.video_player.login")}
                            </button>
                        </div>
                    )}
                    {lecture.authentication_methods.includes("moodle") && (
                        <div className="col-sm-4">
                            <h4 className="text-center">
                                <span className="bi bi-person-fill" aria-hidden="true" /> Moodle
                            </h4>
                            <p>{language.get("ui.video_player.login_moodle.description")}</p>

                            {authed_methods.includes("moodle") ? (
                                <>
                                    <p className="alert alert-info">
                                        {language.get("ui.video_player.login_moodle.not_in_course")}
                                    </p>
                                    <button
                                        type="button"
                                        onClick={() => startOauth("moodle")}
                                        className="btn btn-secondary"
                                    >
                                        {language.get(
                                            "ui.video_player.login_moodle.refresh_course",
                                        )}
                                    </button>
                                </>
                            ) : (
                                <button
                                    type="button"
                                    onClick={() => startOauth("moodle")}
                                    className="btn btn-secondary"
                                >
                                    {language.get("ui.video_player.login")}
                                </button>
                            )}
                        </div>
                    )}
                </div>
                {!lecture.authentication_methods.includes("rwth") &&
                    !lecture.authentication_methods.includes("moodle") &&
                    !lecture.authentication_methods.includes("password") && (
                        <p className="alert alert-info" style={{ marginTop: "2em" }}>
                            {language.get("ui.video_player.student_council_only")}
                        </p>
                    )}
            </div>
        </div>
    );
}

function Chapters({ chapters, seekTo }: { chapters?: chapter[]; seekTo: (time: number) => void }) {
    const { editMode } = useEditMode();

    if (!chapters || chapters.length == 0) return <></>;

    return (
        <div className="col-12 table-responsive" style={{ paddingTop: "10px" }}>
            <h4>
                <strong>Kapitel:</strong>
            </h4>
            <table className="table">
                <tbody>
                    {chapters
                        .sort((a, b) => a.start_time - b.start_time)
                        .map((c) => {
                            const bgColor = c.is_visible === false ? "bg-danger-subtle" : "";
                            return (
                                <tr
                                    key={`${c.start_time}+${c.name}`}
                                    className={"w-100 " + bgColor}
                                >
                                    <td style={{ width: "130px" }} className={bgColor}>
                                        <a
                                            className="chapterlink"
                                            onClick={() => {
                                                seekTo(c.start_time);
                                            }}
                                            href="#"
                                        >
                                            {timestampToString(c.start_time)}
                                        </a>
                                    </td>
                                    <td className={bgColor}>{c.name}</td>
                                    {editMode && (
                                        <td className={"d-flex " + bgColor}>
                                            <div className="flex-fill" />
                                            <EmbeddedOMFieldComponent
                                                object_type="chapter"
                                                object_id={c.id!}
                                                field_id="is_visible"
                                                field_type="boolean"
                                                initialValue={c.is_visible}
                                                className="me-2 align-self-center"
                                            />
                                            <OMDelete object_type="chapter" object_id={c.id!} />
                                        </td>
                                    )}
                                </tr>
                            );
                        })}
                </tbody>
            </table>
        </div>
    );
}

function VideoSuggestions({ course, lecture }: { course: course; lecture: lecture }) {
    const { language } = useLanguage();
    // find previous and next lecture
    const prevLecture = course.lectures?.findLast((l) => new Date(l.time) < new Date(lecture.time));
    const nextLecture = course.lectures?.find((l) => new Date(l.time) > new Date(lecture.time));

    return (
        <div className="d-flex w-100 flex-column flex-sm-row">
            {prevLecture && (
                <div className="card">
                    <Link
                        className="card-header text-center bg-light-subtle"
                        href={urlForLecture(course.id_string, prevLecture.id)}
                    >
                        <i className="bi bi-arrow-left-circle me-2" />
                        {language.get("ui.video_player.previous_lecture")}
                    </Link>
                    <div className="card-body">
                        <VideoCard course={course} lecture={prevLecture} size="small" />
                    </div>
                </div>
            )}

            <div className="flex-fill p-2" />
            {nextLecture && (
                <div className="card">
                    <Link
                        className="card-header text-center bg-light-subtle"
                        href={urlForLecture(course.id_string, nextLecture.id)}
                    >
                        {language.get("ui.video_player.next_lecture")}
                        <i className="bi bi-arrow-right-circle ms-2" />
                    </Link>
                    <div className="card-body">
                        <VideoCard course={course} lecture={nextLecture} size="small" />
                    </div>
                </div>
            )}
        </div>
    );
}

export function sourceToName(source: media_source) {
    if (source.comment.length == 0)
        return `${source.quality.name} (${filesizeToHuman(source.size)})`;

    return `${source.quality.name} (${source.comment})`;
}

export function sortSources(sources?: media_source[]) {
    return Array.from(sources ?? []).sort((a, b) => {
        let aq = parseInt(a.quality.name.split("p")[0]);
        let bq = parseInt(b.quality.name.split("p")[0]);
        if (aq === bq) {
            return (a.id ?? 0) - (b.id ?? 0);
        }
        return bq - aq;
    });
}

export function sortSourceNames(sourceNames?: string[]) {
    return Array.from(sourceNames ?? []).sort((a, b) => {
        let aq = parseInt(a.split("p")[0]);
        let bq = parseInt(b.split("p")[0]);
        if (aq === bq) {
            return a.localeCompare(b);
        }
        return bq - aq;
    });
}

export function EditSources({ media_sources }: { media_sources: media_source[] }) {
    const { editMode } = useEditMode();

    let sortedsources = sortSources(media_sources);

    return (
        <ul>
            {sortedsources.map((source) => (
                <li
                    key={source.id}
                    className={
                        "d-flex border align-items-center px-2 py-1 rounded " +
                        (source.is_visible === false ? "bg-danger-subtle" : "")
                    }
                >
                    <span className="bi bi-download" />
                    <a href={source.download_url ?? source.player_url} className="mx-1">
                        {sourceToName(source)}
                    </a>
                    {editMode && (
                        <>
                            <div className="flex-fill" />
                            <EmbeddedOMFieldComponent
                                object_type="media_source"
                                object_id={source.id!}
                                field_id="is_visible"
                                field_type="boolean"
                                initialValue={source.is_visible}
                                className="me-2 align-self-center py-0"
                            />
                            <OMDelete
                                object_type="media_source"
                                object_id={source.id!}
                                className="py-0"
                            />
                        </>
                    )}
                </li>
            ))}
        </ul>
    );
}

export function DownloadSources({
    media_sources,
    className,
    direction,
    ...props
}: {
    media_sources: media_source[];
    className?: string;
    direction: "up" | "down";
    [key: string]: any;
}) {
    const { editMode } = useEditMode();
    const { language } = useLanguage();

    let sortedsources = sortSources(media_sources);

    return (
        <Dropdown drop={direction}>
            <Dropdown.Toggle
                variant="primary"
                disabled={media_sources === undefined || media_sources.length === 0}
                {...props}
            >
                {language.get("ui.generic.download")}
            </Dropdown.Toggle>

            <Dropdown.Menu>
                {sortedsources.map((v, ind) => {
                    if (v.download_url === undefined) return null;
                    let bgColor = "";
                    if (v.is_visible === false) bgColor = "bg-danger-subtle";

                    return (
                        <div
                            key={ind + "#" + v.quality.name}
                            className={"d-flex align-items-center " + bgColor}
                        >
                            <Dropdown.Item href={v.download_url} download>
                                {sourceToName(v)}
                            </Dropdown.Item>
                            {editMode && (
                                <>
                                    <EmbeddedOMFieldComponent
                                        object_type="media_source"
                                        object_id={v.id!}
                                        field_id="is_visible"
                                        field_type="boolean"
                                        initialValue={v.is_visible}
                                        className="me-2"
                                    />
                                    <OMDelete object_type="media_source" object_id={v.id!} />
                                </>
                            )}
                        </div>
                    );
                })}
            </Dropdown.Menu>
        </Dropdown>
    );
}

export function EmbeddedPlayer({ playerData }: { playerData: ResourceType<PlayerData> }) {
    const { course, perms, lectureId } = playerData.read()!;

    const lecture = course.lectures!.find((l) => l.id === lectureId)!;

    useEffect(() => {
        import("video.js");
    }, []);

    if (lecture.allow_embed === false) {
        return (
            <div className="alert alert-danger" role="alert">
                Dieses Video kann nicht eingebettet werden.
            </div>
        );
    }

    let pageContent = (
        <AuthorizeHelper
            lecture={lecture}
            course={course}
            authed_methods={perms.authenticated_methods}
        />
    );
    if (perms.is_lecture_authenticated) {
        pageContent = <VideoPlayer lecture={lecture} className="h-100" />;
    }

    return <>{pageContent} </>;
}

export default function Player({ playerData }: { playerData: ResourceType<PlayerData> }) {
    const { course, lectureId, perms, loaderHadUserInfo } = playerData.read()!;
    const api = useBackendContext();
    const userContext = useUserContext();
    const hasUserInfo = userContext.hasUserInfo();
    const { language } = useLanguage();

    const lecture = course.lectures!.find((l) => l.id === lectureId)!;

    useEffect(() => {
        import("video.js");
    }, []);

    const [popContent, setPopContent] = useState<React.ReactNode | null>(null);
    const [showPop, setShowPop] = useState(false);
    const { editMode } = useEditMode();
    const reloadFunc = useReloadBoundary();

    useEffect(() => {
        if (hasUserInfo === loaderHadUserInfo) return;
        reloadFunc();
        // eslint-disable-next-line react-hooks/exhaustive-deps
    }, [loaderHadUserInfo, hasUserInfo]);

    const vorschlagenClick = () => {
        if (showPop) {
            setShowPop(false);
            return;
        }
        import("video.js").then((videojs) => {
            let timestamp = videojs.default("videoplayer").currentTime()!;
            let tstr = timestampToString(timestamp);

            setPopContent(
                <KapitelPopover
                    timeStr={tstr}
                    key={tstr}
                    closePop={() => setShowPop(false)}
                    lectureId={lectureId}
                />,
            );
            setShowPop(true);
        });
    };

    let seekTo = (time: number) => {
        import("video.js").then((videojs) => {
            let player = videojs.default("videoplayer");
            player.currentTime(time);
        });
    };

    const kapitelPopover = (
        <Popover>
            <Popover.Header as="h3">
                {userContext.canEditStuff() ? "Neues Kapitel" : "Kapitelmarker vorschlagen"}
            </Popover.Header>
            <Popover.Body>{popContent}</Popover.Body>
        </Popover>
    );

    let einbettenText = `<div style="position:relative;padding-top:56.25%;"><iframe style="position:absolute;top:0;left:0;width:100%;height:100%;overflow: hidden;border: none;" src="${window.location.origin}${basePath}/${course.id_string}/${lecture.id}/embed" allowfullscreen="true" scrolling="no"></iframe></div>`;
    const einbettenCopy = () => {
        var copyText = document.getElementById("einbettenInput") as HTMLInputElement;

        copyText.select();
        copyText.setSelectionRange(0, copyText.value.length); // For mobile devices

        navigator.clipboard.writeText(copyText.value);
    };
    const einbettenPopover = (
        <Popover>
            <Popover.Header as="h3">Einbetten</Popover.Header>
            <Popover.Body>
                <input
                    id="einbettenInput"
                    type="text"
                    className="w-100"
                    onClick={(e: MouseEvent<HTMLInputElement>) =>
                        (e.target as HTMLInputElement).select()
                    }
                    value={einbettenText}
                    readOnly
                />
                <button type="button" className="btn btn-primary mt-2" onClick={einbettenCopy}>
                    Kopieren
                </button>
            </Popover.Body>
        </Popover>
    );

    let pageContent = (
        <AuthorizeHelper
            lecture={lecture}
            course={course}
            authed_methods={perms.authenticated_methods}
        />
    );
    if (perms.is_lecture_authenticated) {
        pageContent = (
            <>
                <div className="col-12 p-0">
                    <VideoPlayer lecture={lecture} />
                </div>
                <div className="col-12 pt-4">
                    <OverlayTrigger
                        trigger="click"
                        placement="top"
                        overlay={kapitelPopover}
                        show={showPop}
                    >
                        <button
                            className="btn btn-primary"
                            id="hintnewchapter"
                            onClick={vorschlagenClick}
                        >
                            {userContext.canEditStuff()
                                ? language.get("ui.video_player.add_chapter")
                                : language.get("ui.video_player.suggest_chapter")}
                        </button>
                    </OverlayTrigger>

                    <ul className="list-inline float-none float-sm-end mb-0 mt-2 mt-sm-0">
                        {lecture.allow_embed && (
                            <li className="list-inline-item">
                                <OverlayTrigger
                                    trigger="click"
                                    placement="top"
                                    overlay={einbettenPopover}
                                >
                                    <a className="btn btn-primary">
                                        <span>{language.get("ui.video_player.embed")}</span>
                                    </a>
                                </OverlayTrigger>
                            </li>
                        )}
                        {lecture.media_sources &&
                            lecture.media_sources.some((ele) => ele.download_url !== undefined) && (
                                <li className="list-inline-item">
                                    <DownloadSources
                                        className="list-inline-item"
                                        direction="up"
                                        media_sources={lecture.media_sources ?? []}
                                    />
                                </li>
                            )}
                    </ul>
                </div>
                {lecture.description.length > 0 && (
                    <div className="mt-2">
                        <h5>Beschreibung</h5>

                        {
                            <EmbeddedOMFieldComponent
                                object_type="lecture"
                                object_id={lecture.id}
                                field_id="description"
                                field_type="string"
                                initialValue={lecture.description}
                                allowMarkdown={true}
                            />
                        }
                    </div>
                )}
                <Chapters chapters={lecture.chapters} seekTo={seekTo} />
            </>
        );
    }

    return (
        <>
            <Head>
                <meta property="og:title" content={`${course.short_name} - ${lecture.title}`} />
                <meta property="og:description" content={course.description} />
                <meta
                    property="og:image"
                    content={
                        lecture.thumbnail_url ?? `${api.assetUrl()}/thumbnail/l_${lecture.id}.jpg`
                    }
                />
            </Head>
            <Title title={`${course.short_name} - ${lecture.title}`} />
            <div className="card">
                <div className="card-header d-flex">
                    <div className="flex-fill align-self-center">
                        <strong>
                            <Link href={`/${course.id_string}#lecture-${lecture.id}`}>
                                {course.full_name}
                            </Link>
                        </strong>
                        :{" "}
                        <EmbeddedOMFieldComponent
                            object_type="lecture"
                            object_id={lecture.id!}
                            field_id="title"
                            field_type="string"
                            initialValue={lecture.title}
                        />{" "}
                        (
                        <EmbeddedOMFieldComponent
                            object_type="lecture"
                            object_id={lecture.id!}
                            field_id="time"
                            field_type="datetime"
                            initialValue={lecture.time}
                        />
                        )
                    </div>

                    <div>
                        {editMode && (
                            <>
                                <OMHistory object_type="lecture" object_id={lecture.id} />
                                <OMEdit object_type="lecture" object_id={lecture.id} />
                                <OMDelete object_type="lecture" object_id={lecture.id} />
                            </>
                        )}
                    </div>
                </div>
                <div className="card-body">
                    <div className="row p-0 mt-1">
                        <div className="col-12 pb-4">
                            <Link
                                href={`/${course.id_string}#lecture-${lecture.id}`}
                                className="btn btn-primary"
                            >
                                <span className="bi bi-chevron-left" />{" "}
                                {language.get("ui.video_player.back_to_course")}
                            </Link>
                        </div>
                    </div>
                    <div className="row mb-2">{pageContent}</div>

                    <VideoSuggestions course={course} lecture={lecture} />
                </div>
            </div>
        </>
    );
}
