import React, { useEffect, useRef, useState } from "react";
import Dropdown from "react-bootstrap/Dropdown";
import { showError } from "@/misc/ErrorHandlers";
import { deepEquals } from "@/misc/Util";
import { useLanguage } from "@/components/LanguageProvider";
import { useBackendContext } from "./BackendProvider";

import type {
    GetUsersResponse,
    field_description,
    view_permissions,
    user_info,
    int,
} from "@/api/api_v1_types";

type EditorArgs = {
    value?: any;
    //affectNow specifies whether value should be applied (e.g. reload data, apply filter) immediately (e.g. user hit enter)
    //or a bit later (debounced) (e.g. user typed in text field)
    updateValue: (new_value: any, isValid: boolean, affectNow: boolean) => void;
    disabled?: boolean;
    autoFocus?: boolean;
    hideErrors?: boolean;
};

export function createOMFieldEditor(description: field_description, args: EditorArgs) {
    const { value } = args;
    switch (description.type) {
        case "string":
        case "url": //TODO
            return (
                <StringEditor
                    {...args}
                    minLength={description.min_length}
                    maxLength={description.max_length}
                />
            );

        case "course_id_string": {
            const forbiddenCourseIdStrings = [
                "courses",
                "faq",
                "imprint",
                "licenses",
                "search",
                "site",
                "internal",
            ];
            // TODO check unique
            return (
                <StringEditor
                    {...args}
                    regex={
                        new RegExp(`^(?!(?:${forbiddenCourseIdStrings.join("|")})$)[a-zA-Z0-9_-]+$`)
                    }
                    regexInvalidMessage={
                        <>
                            Ungültiger Kurs ID-String:
                            <ul>
                                <li>Erlaubte Zeichen: a-z, A-Z, 0-9, _, -</li>
                                <li>Länge muss zwischen 1 und 100 Zeichen sein</li>
                                <li>
                                    Verbotene Strings: &apos;{forbiddenCourseIdStrings.join("', '")}
                                    &apos;
                                </li>
                            </ul>
                        </>
                    }
                />
            );
        }

        case "semester_string":
            return (
                <StringEditor
                    {...args}
                    regex={/^([0-9]{4}(ws|ss)|none)$/}
                    regexInvalidMessage="Ungültiges Semester. Semester muss entweder 'none' oder im Format '2024ws' bzw. '2024ss' sein"
                />
            );

        case "boolean":
            return <BooleanEditor {...args} />;

        case "int":
        case "lecture_id": //Nullable //TODO
        case "user_id": //Nullable
            return (
                <IntEditor
                    {...args}
                    minValue={description.min_value}
                    maxValue={description.max_value}
                />
            );

        case "datetime":
            return <DatetimeEditor {...args} mayBeNull={description.may_be_null} />;

        case "legacy_announcement_type":
            return (
                <ChooserEditor
                    {...args}
                    langKeyPrefix="type.legacy_announcement_type"
                    possibleValues={[
                        "info_main_page",
                        "info_all_pages",
                        "warning_all_pages",
                        "important_all_pages_embed",
                    ]}
                />
            );

        case "announcement_type":
            return (
                <ChooserEditor
                    {...args}
                    langKeyPrefix="type.announcement_type"
                    possibleValues={["info", "warning", "important"]}
                />
            );

        case "announcement_visibility":
            return (
                <ChooserEditor
                    {...args}
                    langKeyPrefix="type.announcement_visibility"
                    possibleValues={["only_main_page", "all_pages", "all_pages_and_embed"]}
                />
            );

        case "course_filter_criteria":
            return (
                <ChooserEditor
                    {...args}
                    langKeyPrefix="type.course_filter_criteria"
                    possibleValues={["title", "semester", "organizer", "topic"]}
                />
            );

        case "user_id_list":
            return <UserIdListEditor {...args} />;

        case "view_permissions":
            return <ViewPermissionsEditor {...args} />;

        default:
            return (
                <span>
                    Unknown field type {description.type}. {JSON.stringify(value)}
                </span>
            );
    }
}

export function StringEditor({
    value,
    updateValue,
    disabled,
    autoFocus,
    hideErrors,
    placeholder,
    minLength,
    maxLength,
    allowUndefined,
    regex,
    regexInvalidMessage,
}: EditorArgs & {
    placeholder?: string;
    minLength?: int;
    maxLength?: int;
    allowUndefined?: boolean;
    regex?: RegExp;
    regexInvalidMessage?: string | React.ReactNode;
}) {
    useEffect(() => {
        if (allowUndefined !== true && value === undefined) updateValue("", true, true);
        if (allowUndefined === true && value === "") updateValue(undefined, true, true);
        // eslint-disable-next-line react-hooks/exhaustive-deps
    }, [value, allowUndefined]);

    const isValidCheck = (val: string | undefined) => {
        if (val === undefined) return allowUndefined === true;
        if (minLength !== undefined && val.length < minLength) return false;
        if (maxLength !== undefined && val.length > maxLength) return false;
        if (regex !== undefined && !regex.test(val)) return false;
        return true;
    };
    const update = (val: any, affectNow: boolean) => {
        if (val === "" && allowUndefined === true) val = undefined;
        updateValue(val, isValidCheck(val), affectNow);
    };
    const isValid = isValidCheck(value);

    var invalidMessage = null;
    if (!isValid && hideErrors !== true) {
        if (
            value === undefined ||
            (minLength !== undefined && value.length < minLength) ||
            (maxLength !== undefined && value.length > maxLength)
        ) {
            invalidMessage = (
                <>
                    Länge muss{" "}
                    {maxLength !== undefined
                        ? `zwischen ${minLength ?? 0} und ${maxLength}`
                        : `größer gleich ${minLength ?? 0}`}{" "}
                    Zeichen sein
                </>
            );
        } else {
            invalidMessage = regexInvalidMessage;
        }
    }
    const className = "form-control w-100 " + (isValid || hideErrors === true ? "" : "is-invalid");
    if (maxLength && maxLength < 300)
        return (
            <div className="d-inline-grid w-100">
                <input
                    className={className}
                    type="text"
                    placeholder={placeholder}
                    value={value ?? ""}
                    onChange={(e) => update(e.target.value, false)}
                    onKeyDown={(e) => {
                        if (e.key === "Enter") update((e.target as HTMLInputElement).value, true);
                    }}
                    disabled={disabled}
                    autoFocus={autoFocus}
                />
                {invalidMessage !== null && (
                    <span className="d-inline invalid-feedback">{invalidMessage}</span>
                )}
            </div>
        );
    else
        return (
            <div className="d-inline-grid w-100">
                <textarea
                    className={className}
                    style={{ minHeight: "6em" }}
                    placeholder={placeholder}
                    value={value ?? ""}
                    onChange={(e) => update(e.target.value, false)}
                    disabled={disabled}
                    autoFocus={autoFocus}
                />
                {invalidMessage !== null && (
                    <span className="d-inline invalid-feedback">{invalidMessage}</span>
                )}
            </div>
        );
}

export function BooleanEditor({ value, updateValue, disabled, autoFocus, hideErrors }: EditorArgs) {
    useEffect(() => {
        if (value === undefined) updateValue(false, true, true);
        // eslint-disable-next-line react-hooks/exhaustive-deps
    }, [value]);
    return (
        <input
            type="checkbox"
            className="form-check-input m-0"
            checked={value ?? false}
            onChange={(e) => updateValue(e.target.checked, true, true)}
            disabled={disabled}
            autoFocus={autoFocus}
        />
    );
}

export function IntEditor({
    value,
    updateValue,
    disabled,
    autoFocus,
    hideErrors,
    minValue,
    maxValue,
    allowUndefined,
}: EditorArgs & {
    minValue?: int;
    maxValue?: int;
    allowUndefined?: boolean;
}) {
    const isValidCheck = (val: undefined | string) => {
        if (val === undefined) return allowUndefined === true;
        if (!/^[+-]?[0-9]+$/.test(val)) return false;
        const valInt = parseInt(val);
        if (minValue !== undefined && valInt < minValue) return false;
        if (maxValue !== undefined && valInt > maxValue) return false;
        return true;
    };
    const update = (val: any, affectNow: boolean) => {
        if (val === "" && allowUndefined === true) {
            val = undefined;
        }
        const isValid = isValidCheck(val);
        updateValue(
            val === undefined ? undefined : isValid ? parseInt(val) : val,
            isValid,
            affectNow,
        );
    };
    const isValid = isValidCheck(value);

    let invalidMessage = undefined;
    if (!isValid && hideErrors !== true) {
        let subMessage = "";
        if (minValue !== undefined && maxValue !== undefined) {
            subMessage = `zwischen ${minValue} und ${maxValue}`;
        } else if (minValue !== undefined) {
            subMessage = `größer gleich ${minValue}`;
        } else if (maxValue !== undefined) {
            subMessage = `kleiner gleich ${maxValue}`;
        }
        invalidMessage = (
            <span className="d-inline invalid-feedback">
                Muss eine ganze Zahl {subMessage} sein
            </span>
        );
    }
    return (
        <div className="d-inline-grid">
            <input
                className={
                    "form-control w-auto " + (isValid || hideErrors === true ? "" : "is-invalid")
                }
                type="text"
                inputMode="numeric"
                value={value ?? ""}
                onChange={(e) => update(e.target.value, false)}
                onKeyDown={(e) => {
                    if (e.key === "Enter") update((e.target as HTMLInputElement).value, true);
                }}
                disabled={disabled}
                autoFocus={autoFocus}
            />
            {invalidMessage}
        </div>
    );
}

export function DatetimeEditor({
    value,
    updateValue,
    disabled,
    autoFocus,
    hideErrors,
    mayBeNull,
}: EditorArgs & {
    mayBeNull?: boolean;
}) {
    // We need to store the default value locally to enable smooth editing. If the input is partially invalid, we just
    // get nothing and setting that back as the value will cause the whole input to reset
    const knownValue = useRef<string | null>();
    const [defaultValue, setDefaultValue] = useState<string | null>();
    const [forceInputUpdate, setForceInputUpdate] = useState<int>(0);

    useEffect(() => {
        if (value === knownValue.current) return;
        //Value was changed not by us, update internal state
        knownValue.current = value;
        setDefaultValue(value);
        // Needed if the previous defaultValue equals the new value, but is not equal to the current value in the input
        setForceInputUpdate((i) => i + 1);
    }, [knownValue, value, setForceInputUpdate]);

    const isValidCheck = (val: string | null | undefined) => {
        if (val === null || val === undefined || isNaN(new Date(val).valueOf())) {
            return mayBeNull === true;
        }
        return true;
    };

    const updateRawValue = (newValue: string | null | undefined) => {
        let isValid = true;
        if (newValue === null || newValue === undefined || isNaN(new Date(newValue).valueOf())) {
            newValue = null;
            if (mayBeNull !== true) {
                isValid = false;
            }
        } else {
            const date = new Date(newValue);
            const intToStr = (v: int, length: int) => {
                let str = v.toString();
                while (str.length < length) {
                    str = "0" + str;
                }
                return str;
            };
            const dateStr = `${intToStr(date.getFullYear(), 4)}-${intToStr(date.getMonth() + 1, 2)}-${intToStr(date.getDate(), 2)}`;
            const timeStr = `${intToStr(date.getHours(), 2)}:${intToStr(date.getMinutes(), 2)}:${intToStr(date.getSeconds(), 2)}`;
            newValue = `${dateStr}T${timeStr}`;
        }
        knownValue.current = newValue;
        updateValue(newValue, isValid, false);
    };

    const isValid = isValidCheck(value);

    let message = undefined;
    if (!isValid) {
        if (hideErrors !== true) {
            message = <span className="d-inline invalid-feedback">Ungültiges Datum</span>;
        }
    } else if (value === null) {
        message = <span>Ungültiges Datum. Wird als nicht gesetzt interpretiert</span>;
    }
    return (
        <div className="d-inline-grid">
            <span>
                <input
                    className={
                        "form-control d-inline w-auto " +
                        (isValid || hideErrors === true ? "" : "is-invalid")
                    }
                    type="datetime-local"
                    defaultValue={defaultValue === null ? undefined : defaultValue}
                    key={forceInputUpdate}
                    onChange={(e) => updateRawValue(e.target.value)}
                    disabled={disabled}
                    autoFocus={autoFocus}
                />
                {mayBeNull === true && (
                    <button
                        type="button"
                        className="ms-2 btn btn-outline-danger"
                        onClick={() => {
                            knownValue.current = null;
                            setDefaultValue(null);
                            updateValue(null, true, true);
                        }}
                        disabled={value === null}
                    >
                        <i className="bi bi-x" />
                    </button>
                )}
            </span>
            {message}
        </div>
    );
}

export function ChooserEditor({
    value,
    updateValue,
    disabled,
    autoFocus,
    hideErrors,
    langKeyPrefix,
    possibleValues,
    unchosenLangKey, // If not present, a value must be chosen
}: EditorArgs & {
    langKeyPrefix?: string;
    possibleValues: string[] | number[];
    unchosenLangKey?: string;
}) {
    const { language } = useLanguage();
    return (
        <select
            className="form-select d-inline w-auto"
            value={value !== undefined ? JSON.stringify(value) : "unchosen"}
            onChange={(e) =>
                updateValue(
                    e.target.value === "unchosen" ? undefined : JSON.parse(e.target.value),
                    true,
                    true,
                )
            }
            disabled={disabled}
            autoFocus={autoFocus}
        >
            <option disabled={unchosenLangKey === undefined} value="unchosen">
                {unchosenLangKey === undefined ? "Choose ..." : language.get(unchosenLangKey)}
            </option>
            {possibleValues.map((value) => (
                <option key={value} value={JSON.stringify(value)}>
                    {langKeyPrefix !== undefined
                        ? language.get(`${langKeyPrefix}.${value}`)
                        : value}
                </option>
            ))}
        </select>
    );
}

export function UserIdListEditor({
    value,
    updateValue,
    disabled,
    autoFocus,
    hideErrors,
}: EditorArgs) {
    const api = useBackendContext();
    const knownValue = useRef<int[]>();
    const [users, setUsers] = useState<GetUsersResponse>();
    const [search, setSearch] = useState<string>("");

    useEffect(() => {
        if (knownValue.current === value) return;
        // Value was not changed by us, reset internal state
        knownValue.current = value;
        setUsers(undefined);
        setSearch("");
        api.getUsers()
            .then(setUsers)
            .catch((err) => {
                showError(err, "Unable to load users");
            });
    }, [api, knownValue, value]);

    useEffect(() => {
        if (value !== undefined) return;
        const newValue: int[] = [];
        knownValue.current = newValue;
        updateValue(newValue, true, true);
    });

    const selectedArray = value ?? [];

    let userList = selectedArray.length > 0 ? selectedArray.join(", ") : "None";
    if (users !== undefined && selectedArray.length > 0) {
        userList = selectedArray
            .map((id: int) => {
                const user = users.users.find((u: user_info) => u.id === id);
                if (user === undefined) return `#${id}`;
                return `${user.name} (${user.id_string}, #${user.id})`;
            })
            .join(", ");
    }
    if (userList.length > 100) userList = userList.slice(0, 100) + "...";

    const searchTerm = search.toLowerCase();

    const userElements = (users === undefined ? [] : users.users).flatMap((user: user_info) => {
        if (
            !user.name.toLowerCase().includes(searchTerm) &&
            !user.id_string.toLowerCase().includes(searchTerm)
        )
            return [];
        const toggle = () => {
            const index = selectedArray.indexOf(user.id);
            let newArray = selectedArray.slice();
            if (index === -1) {
                newArray.splice(newArray.length, 0, user.id); //Add user.id at end
            } else {
                newArray.splice(index, 1); //Remove user.id
            }
            knownValue.current = newArray;
            updateValue(newArray, true, true);
        };
        return [
            <tr key={user.id} className={selectedArray.includes(user.id) ? "table-primary" : ""}>
                <td>
                    <input
                        type="checkbox"
                        checked={selectedArray.includes(user.id)}
                        onChange={toggle}
                    />
                </td>
                <td onClick={toggle} className="pe-auto" style={{ cursor: "pointer" }}>
                    <span>{`${user.name} (${user.id_string}, #${user.id})`}</span>
                </td>
            </tr>,
        ];
    });

    return (
        <Dropdown>
            <Dropdown.Toggle variant="primary" style={{ textWrap: "wrap" }} disabled={disabled}>
                {userList}
            </Dropdown.Toggle>
            <Dropdown.Menu className="p-1">
                <div className="input-group mb-2 p-1">
                    <span className="input-group-text">
                        <i className="bi bi-search" />
                    </span>
                    <input
                        type="text"
                        className="form-control"
                        placeholder="Username"
                        value={search}
                        autoFocus
                        onChange={(e) => setSearch(e.target.value)}
                    />
                    <button
                        type="button"
                        className="btn btn-outline-secondary"
                        onClick={() => setSearch("")}
                        disabled={search === ""}
                    >
                        <i className="bi bi-x-circle" />
                    </button>
                </div>
                {users === undefined ? (
                    <div className="spinner-border" role="status" />
                ) : (
                    <>
                        {userElements.length === 0 && (
                            <div className="text-center">No users found</div>
                        )}
                        <div className="overflow-auto" style={{ maxHeight: "500px" }}>
                            <table className="table table-sm table-hover w-100">
                                <thead>
                                    <tr>
                                        <th></th>
                                        <th className="w-100"></th>
                                    </tr>
                                </thead>
                                <tbody>{userElements}</tbody>
                            </table>
                        </div>
                    </>
                )}
            </Dropdown.Menu>
        </Dropdown>
    );
}

function createListEditor<V>(
    disabled: boolean,
    list: V[],
    defaultNewValue: V,
    onNewValueList: (newList: V[]) => void,
    elementFactory: (value: V, onNewValue: (newVal: V) => void) => React.ReactNode,
) {
    return (
        <>
            {list.map((value: V, index: int) => (
                <div key={index} className="d-block">
                    {elementFactory(value, (newVal: V) =>
                        onNewValueList(
                            list.map((subValue: V, subIndex: int) =>
                                index === subIndex ? newVal : subValue,
                            ),
                        ),
                    )}
                    <button
                        type="button"
                        className={"btn btn-outline-danger m-1"}
                        onClick={(e) => {
                            const newList = list.slice();
                            newList.splice(index, 1);
                            onNewValueList(newList);
                        }}
                        title="Delete"
                    >
                        <i className="bi bi-trash-fill" />
                    </button>
                </div>
            ))}
            <button
                className="btn btn-secondary"
                type="button"
                title="Add entry"
                onClick={(e) => {
                    onNewValueList(list.concat([defaultNewValue]));
                }}
                disabled={disabled}
            >
                +
            </button>
        </>
    );
}

export function ViewPermissionsEditor({
    value,
    updateValue,
    disabled,
    autoFocus,
    hideErrors,
}: EditorArgs) {
    const permTypes = ["inherit", "public", "private", "authentication"];

    // We need to store the current value locally since there is no order for the passwords and it would be really
    // annoying if the order changes while typing
    // This also gives the benefit of storing values for types which are currently not selected
    type view_permissions_with_invalid = Omit<view_permissions, "passwords"> & {
        passwords?: null | { [key: string]: string };
    };
    const { language } = useLanguage();
    const knownValue = useRef<view_permissions_with_invalid>();
    const [allDataValue, setAllDataValue] = useState<view_permissions_with_invalid>({
        type: "inherit",
    });
    const [passwordsArray, setPasswordsArray] = useState<[string, string][]>([]);

    useEffect(() => {
        if (deepEquals(value, knownValue.current)) return;
        // Value was not changed by us, update internal state
        const effectiveValue = value ?? { type: "inherit" };
        const passwordsMap = effectiveValue.passwords ?? {};
        setPasswordsArray(
            Object.keys(passwordsMap).map((username: string) => [username, passwordsMap[username]]),
        );
        setAllDataValue(effectiveValue);
        knownValue.current = effectiveValue;
        if (effectiveValue !== value) updateValue(effectiveValue, true, true);
    }, [value, updateValue]);

    const updateData = (key: string, newData: any) =>
        setAllDataValue((prev: view_permissions_with_invalid | undefined) => {
            let isValid = true;

            prev = prev ?? { type: "inherit" };
            let newAllData: view_permissions_with_invalid = { ...prev, [key]: newData };
            if (key === "passwords") {
                setPasswordsArray(newData);
                let newMap: { [key: string]: string } | null = {};
                for (let ele of newData) {
                    if (newMap[ele[0]] !== undefined) {
                        newMap = null;
                        break;
                    }
                    newMap[ele[0]] = ele[1];
                }
                newAllData = { ...prev, passwords: newMap };
            }

            let newValue: view_permissions_with_invalid = { type: newAllData.type };
            if (newValue.type === "authentication") {
                newValue.rwth_authentication = newAllData.rwth_authentication ?? false;
                newValue.fsmpi_authentication = newAllData.fsmpi_authentication ?? false;
                newValue.moodle_course_ids = newAllData.moodle_course_ids ?? [];
                newValue.passwords = newAllData.passwords !== undefined ? newAllData.passwords : {};
                if (newValue.passwords === null) {
                    isValid = false;
                } else if (
                    !newValue.rwth_authentication &&
                    !newValue.fsmpi_authentication &&
                    newValue.moodle_course_ids.length === 0 &&
                    Object.keys(newValue.passwords).length === 0
                ) {
                    isValid = false;
                }
            }
            knownValue.current = newValue;
            updateValue(newValue, isValid, true);
            return newAllData;
        });

    let invalidMessage = undefined;
    if (
        hideErrors !== true &&
        allDataValue?.type === "authentication" &&
        allDataValue.passwords !== null &&
        !(
            allDataValue.rwth_authentication === true ||
            allDataValue.fsmpi_authentication === true ||
            (allDataValue.moodle_course_ids ?? []).length > 0 ||
            Object.keys(allDataValue.passwords ?? {}).length > 0
        )
    ) {
        invalidMessage = (
            <div className="d-block invalid-feedback">
                Für Typ <i>Authentifizierung</i> muss mindestens eine Authentifizierungsmethode
                vorhanden sein
            </div>
        );
    }

    let passwordsInvalidMessage;
    if (hideErrors !== true && allDataValue?.type === "authentication") {
        const passwordsMap: { [key: string]: null } = {};
        for (let ele of passwordsArray) {
            if (passwordsMap[ele[0]] !== undefined) {
                passwordsInvalidMessage = (
                    <div className="d-block invalid-feedback">
                        Doppelter Nutzername &apos;{ele[0]}&apos;
                    </div>
                );
                break;
            }
            passwordsMap[ele[0]] = null;
        }
    }

    return (
        <div className="d-inline-grid">
            <select
                className="w-auto form-select"
                value={allDataValue?.type ?? "inherit"}
                onChange={(e) => updateData("type", e.target.value)}
                disabled={disabled}
                autoFocus={autoFocus}
            >
                {permTypes.map((type) => (
                    <option key={type} value={type}>
                        {language.get(`ui.view_permissions.type.${type}`)}
                    </option>
                ))}
            </select>
            {allDataValue?.type === "authentication" && (
                <table className="w-100 table">
                    <thead>
                        <tr>
                            <th scope="col" style={{ width: "30%" }} />
                            <th scope="col" style={{ width: "70%" }} />
                        </tr>
                    </thead>
                    <tbody>
                        <tr>
                            <td>RWTH authentication</td>
                            <td>
                                <input
                                    type="checkbox"
                                    className="form-check-input"
                                    name="rwth_authentication"
                                    defaultChecked={allDataValue?.rwth_authentication ?? false}
                                    disabled={disabled}
                                    onChange={(e) =>
                                        updateData("rwth_authentication", e.target.checked)
                                    }
                                />
                            </td>
                        </tr>
                        <tr>
                            <td>FSMPI authentication</td>
                            <td>
                                <input
                                    type="checkbox"
                                    className="form-check-input"
                                    name="fsmpi_authentication"
                                    defaultChecked={allDataValue?.fsmpi_authentication ?? false}
                                    disabled={disabled}
                                    onChange={(e) =>
                                        updateData("fsmpi_authentication", e.target.checked)
                                    }
                                />
                            </td>
                        </tr>
                        <tr>
                            <td>Moodle course ids</td>
                            <td>
                                {createListEditor<int>(
                                    disabled ?? false,
                                    allDataValue?.moodle_course_ids ?? [],
                                    0,
                                    (newList: int[]) => updateData("moodle_course_ids", newList),
                                    (value: int, onNewValue: (newVal: int) => void) => (
                                        <input
                                            className="mb-1"
                                            type="number"
                                            value={value}
                                            onChange={(e) => {
                                                onNewValue(parseInt(e.target.value));
                                            }}
                                            disabled={disabled ?? false}
                                        />
                                    ),
                                )}
                            </td>
                        </tr>
                        <tr>
                            <td>Passwords</td>
                            <td>
                                <div className="d-inline-grid">
                                    <span>
                                        {createListEditor<[string, string]>(
                                            disabled ?? false,
                                            passwordsArray,
                                            ["", ""],
                                            (newList: [string, string][]) =>
                                                updateData("passwords", newList),
                                            (
                                                value: [string, string],
                                                onNewValue: (newVal: [string, string]) => void,
                                            ) => (
                                                <>
                                                    <input
                                                        className="me-1"
                                                        type="text"
                                                        placeholder="Username"
                                                        value={value[0]}
                                                        onChange={(e) =>
                                                            onNewValue([e.target.value, value[1]])
                                                        }
                                                        disabled={disabled ?? false}
                                                    />
                                                    <input
                                                        type="text"
                                                        placeholder="Password"
                                                        value={value[1]}
                                                        onChange={(e) =>
                                                            onNewValue([value[0], e.target.value])
                                                        }
                                                        disabled={disabled ?? false}
                                                    />
                                                </>
                                            ),
                                        )}
                                    </span>
                                    {passwordsInvalidMessage}
                                </div>
                            </td>
                        </tr>
                    </tbody>
                </table>
            )}
            {invalidMessage}
        </div>
    );
}
