import { DateTime } from "luxon";

export interface ICalEvent {
    summary?: string;
    startDate: DateTime;
    endDate?: DateTime;
    duration?: number;
    location?: string;
    description?: string;
    id_in_videoag_system?: number;
    id_in_foreign_system?: number;
    locally_unique_id: string;
}

export function parseICal(fileContent: string): ICalEvent[] {
    // Array to store parsed events
    const events: ICalEvent[] = [];

    // Split the iCal content into lines
    const lines = fileContent.split("\n");

    // Variables to store current event details and timezone
    let currentEvent: Partial<ICalEvent> = {};
    let inEvent = false;
    let calendarTimezone: string | null = null;

    // Function to adjust date with timezone (placeholder for actual adjustment logic)
    const adjustDateForTimezone = (dateString: string, timezone: string): DateTime => {
        // Convert the date string to ISO 8601 format
        const dt = DateTime.fromISO(dateString).setZone(timezone);
        if (!dt.isValid) {
            console.error("Invalid DateTime after adjustment:", dt.invalidExplanation);
        }
        return dt;
    };

    // Iterate through each line
    for (const line of lines) {
        if (line.startsWith("X-WR-TIMEZONE:")) {
            calendarTimezone = line.split(":")[1].trim();
        } else if (line.startsWith("BEGIN:VEVENT")) {
            inEvent = true;
            currentEvent = {};
        } else if (line.startsWith("END:VEVENT")) {
            inEvent = false;
            if (currentEvent.startDate && currentEvent.endDate) {
                currentEvent.duration = currentEvent.endDate
                    .diff(currentEvent.startDate)
                    .as("minutes");
            }
            if (currentEvent.startDate) {
                events.push(currentEvent as ICalEvent);
            } else {
                console.error("Event without start date");
            }
        } else if (inEvent) {
            // Parse individual lines within the event
            const [key, value] = line.trim().split(":");
            if (key && value) {
                switch (key) {
                    case "SUMMARY":
                        currentEvent.summary = value;
                        break;
                    case "DTSTART":
                    case "DTEND":
                        if (calendarTimezone) {
                            currentEvent[key === "DTSTART" ? "startDate" : "endDate"] =
                                adjustDateForTimezone(value, calendarTimezone);
                        } else {
                            currentEvent[key === "DTSTART" ? "startDate" : "endDate"] =
                                DateTime.fromISO(value);
                        }
                        break;
                    case "LOCATION":
                        currentEvent.location = value;
                        break;
                    case "DESCRIPTION":
                        currentEvent.description = value;
                        break;
                }
            }
        }
    }

    return events;
}
