import { toast, Bounce } from "react-toastify";
import { ApiError } from "@/api/ApiError";
import Four04 from "@/pages/404";
import { useReloadBoundary } from "@/components/ReloadBoundary";

import type React from "react";

export function showError(error: Error, prefix_message = "Ein Fehler ist aufgetreten") {
    if (error instanceof ApiError) {
        if (error.error_code === "site_is_disabled") {
            showErrorToast(<div>Unable to execute action. The site is currently disabled.</div>);
        } else if (error.error_code === "site_is_readonly") {
            showErrorToast(<div>Unable to execute action. The site is currently read-only.</div>);
        } else {
            showErrorToast(
                <div>
                    {prefix_message}:
                    <br />
                    {error.api_message}
                    <br />
                    <small className="text-muted">
                        {error.error_code} - {error.url} - {error.server_name}
                    </small>
                </div>,
            );
        }
    } else {
        showErrorToast(`${prefix_message}: ${error.message}`);
    }
}

export function showErrorToast(message: React.ReactNode) {
    toast.error(message, {
        position: "top-center",
        autoClose: false,
        closeOnClick: false,
        draggable: false,
        progress: undefined,
        theme: "light", // Colors overridden in css
        transition: Bounce,
    });
}

export function showWarningToast(message: React.ReactNode) {
    toast.warning(message, {
        position: "top-center",
        autoClose: false,
        closeOnClick: false,
        draggable: false,
        progress: undefined,
        theme: "light", // Colors overridden in css
        transition: Bounce,
    });
}

const expectedErrorCodesMessages = new Map([
    ["unauthorized", ["401 Nicht autorisiert", "Du musst eingeloggt sein um diese Seite zu sehen"]],
    ["access_forbidden", ["403 Zugriff verweigert", "Du hast keinen Zugriff auf diese Seite"]],
    [
        "rate_limited",
        [
            "429 Rate limited",
            "Du sendest zu viele Anfragen und wurdest kurzzeitig blockiert. Versuche es später erneut",
        ],
    ],
]);

export function ErrorPage({
    error,
    objectName,
    objectPlural = false,
    showButtons = true,
    expectedErrorCodes = ["rate_limited"],
}: {
    error: any;
    objectName: string;
    objectPlural?: boolean;
    showButtons?: boolean;
    expectedErrorCodes?: string[];
}) {
    const reloadFunc = useReloadBoundary();

    if (error instanceof ApiError && expectedErrorCodes.includes(error.error_code)) {
        let title = error.error_code;
        let message = error.api_message;
        if (expectedErrorCodesMessages.has(error.error_code)) {
            [title, message] = expectedErrorCodesMessages.get(error.error_code)!;
        } else if (error.error_code === "unknown_object") {
            title = "404";
            message = `${objectName} ${objectPlural ? "wurden" : "wurde"} nicht gefunden`;
        }
        return <Four04 title={title} text={message} showButtons={showButtons} />;
    }

    const mainMessage = `${objectName} ${objectPlural ? "konnten" : "konnte"} nicht geladen werden`;

    if (error instanceof ApiError) {
        return (
            <div className="alert alert-danger" role="alert">
                <h4 className="alert-heading">Ein unerwarteter Fehler hat dich besucht</h4>
                <p>
                    {mainMessage}: {error.api_message}
                </p>
                <p>
                    Falls das Problem länger bestehen sollte, schreib uns bitte eine Mail an{" "}
                    <a href="mailto:video@fsmpi.rwth-aachen.de">video@fsmpi.rwth-aachen.de</a> in
                    der du folgenden Informationen nennst:
                    <ul>
                        <li>Uhrzeit: {new Date().toISOString()}</li>
                        <li>URL: {error.url}</li>
                        <li>Server: {error.server_name}</li>
                        <li>
                            Fehler: {error.error_code} - {error.api_message}
                        </li>
                    </ul>
                    Wir werden uns dann schnellst möglich darum kümmern.
                </p>
                <button type="button" className="btn btn-success mb-2" onClick={reloadFunc}>
                    <span className="bi bi-arrow-counterclockwise" />
                    Neuladen
                </button>
            </div>
        );
    }
    return (
        <div className="alert alert-danger" role="alert">
            <h4 className="alert-heading">Unerwarteter Fehler!</h4>
            <p>{mainMessage}</p>
            <div>
                Falls das Problem länger bestehen sollte, schreib uns bitte eine Mail an{" "}
                <a href="mailto:video@fsmpi.rwth-aachen.de">video@fsmpi.rwth-aachen.de</a> in der du
                folgenden Informationen nennst:
                <ul>
                    <li>Uhrzeit: {new Date().toISOString()}</li>
                    <li>
                        {window?.location?.href
                            ? `URL: ${window.location.href}`
                            : "Seite auf welcher der Fehler aufgetreten ist"}
                    </li>
                    <li>Fehler: {error.message}</li>
                    <li>Was hast du getan bevor der Fehler aufgetreten ist?</li>
                    <li>Wie kann man den Fehler reproduzieren?</li>
                </ul>
                Wir werden uns dann schnellst möglich darum kümmern.
            </div>
            <button type="button" className="btn btn-success mt-2" onClick={reloadFunc}>
                <span className="bi bi-arrow-counterclockwise me-1" />
                Neuladen
            </button>
        </div>
    );
}
