import { useRef } from "react";
/* global NodeJS */

export function fetchDataWrapper<T>(prom: Promise<T>) {
    let status = "pending";
    let result: T;
    let suspender = new Promise<void>((resolve, reject) => {
        prom.then((r) => {
            status = "success";
            result = r;
            resolve();
        }).catch((e) => {
            status = "error";
            result = e;
            reject();
        });
    });
    return {
        read() {
            if (status === "pending") {
                throw suspender;
            } else if (status === "error") {
                throw result;
            } else if (status === "success") {
                return result;
            }
        },
    };
}

export function fetchDataDelayWrapper<T>(prom: Promise<T>, artificialDelay: number) {
    let status = "pending";
    let result: T;
    let suspender = new Promise<void>((resolve, reject) => {
        setTimeout(() => {
            prom.then((r) => {
                status = "success";
                result = r;
                resolve();
            }).catch((e) => {
                status = "error";
                result = e;
                reject();
            });
        }, artificialDelay);
    });
    return {
        read() {
            if (status === "pending") {
                throw suspender;
            } else if (status === "error") {
                throw result;
            } else if (status === "success") {
                return result;
            }
        },
    };
}

export type ResourceType<T> = { read(): T | undefined };

// Returns [deferred, now]
export function useDebounce<T extends () => any>(func: T, delay: number): [() => void, T] {
    let timeoutId = useRef<NodeJS.Timeout>();
    if (delay === 0) {
        return [func, func];
    }

    return [
        () => {
            clearTimeout(timeoutId.current);
            timeoutId.current = setTimeout(() => {
                func();
            }, delay);
        },
        (() => {
            clearTimeout(timeoutId.current);
            return func();
        }) as T,
    ];
}

export function useDebounceWithArgument<V, R>(
    func: (param: V) => R,
    delay: number,
): [(param: V) => void, (param: V) => R] {
    let timeoutId = useRef<NodeJS.Timeout>();
    if (delay === 0) {
        return [func, func];
    }

    return [
        (param: V) => {
            clearTimeout(timeoutId.current);
            timeoutId.current = setTimeout(() => {
                func(param);
            }, delay);
        },
        (param: V) => {
            clearTimeout(timeoutId.current);
            return func(param);
        },
    ];
}

export function useDebounceUpdateData<D, R>(
    updater: (newData: D) => Promise<R>,
    delay: number,
): [(newData: D) => void, (newData: D) => Promise<R>] {
    const timeoutIdRef = useRef<NodeJS.Timeout>();
    const processingPromiseRef = useRef<Promise<R> | undefined>(undefined);
    const unprocessedDataRef = useRef<D | undefined>(undefined);

    function updateData(): Promise<R> | undefined {
        if (processingPromiseRef.current !== undefined) {
            return processingPromiseRef.current;
        }
        if (unprocessedDataRef.current === undefined) {
            return undefined;
        }
        const data = unprocessedDataRef.current;
        const onUpdateFinished = (arg: any) => {
            processingPromiseRef.current = undefined;
            if (unprocessedDataRef.current !== undefined) {
                if (unprocessedDataRef.current === data) {
                    unprocessedDataRef.current = undefined;
                } else {
                    updateData();
                }
            }
            return arg;
        };
        unprocessedDataRef.current = undefined;
        const promise = updater(data).then(onUpdateFinished, onUpdateFinished);
        processingPromiseRef.current = promise;
        return promise;
    }
    return [
        (newData: D) => {
            unprocessedDataRef.current = newData;
            clearTimeout(timeoutIdRef.current);
            timeoutIdRef.current = setTimeout(updateData, delay);
        },
        (newData: D) => {
            clearTimeout(timeoutIdRef.current);
            unprocessedDataRef.current = newData;
            return updateData()!;
        },
    ];
}
