import { useEffect, useState } from "react";

export function storageGetOrDefault(key: string, defaultValue: any) {
    if (typeof window === "undefined") {
        throw new Error(
            "storageGetOrDefault is not available on the server side. Run it inside a useEffect hook.",
        );
    }

    if (typeof localStorage === "undefined") return defaultValue;
    const value = localStorage.getItem(key);
    if (value === null || value === undefined) {
        return defaultValue;
    }
    try {
        return JSON.parse(value);
    } catch (e) {
        console.error("Error while parsing storage value", e);
        return defaultValue;
    }
}

export function storageSet(key: string, value: any) {
    if (typeof window === "undefined") {
        throw new Error(
            "storageGetOrDefault is not available on the server side. Run it inside a useEffect hook.",
        );
    }
    if (typeof localStorage === "undefined") return;
    localStorage.setItem(key, JSON.stringify(value));
}

export function useLocalStorageState<T>(
    key: string,
    initialValue: T,
): [T, (newValue: T | ((oldValue: T) => T)) => void] {
    const [value, setValue] = useState<T>(initialValue);

    useEffect(() => {
        const val = storageGetOrDefault(key, undefined);
        if (val !== undefined) {
            setValue(val);
        }
    }, [key, setValue]);

    return [
        value,
        (newValue: T | ((oldValue: T) => T)) => {
            if (typeof newValue === "function") {
                setValue((oldValue: T) => {
                    const actualNewValue = (newValue as (oldValue: T) => T)(oldValue);
                    storageSet(key, actualNewValue);
                    return actualNewValue;
                });
            } else {
                storageSet(key, newValue);
                setValue(newValue);
            }
        },
    ];
}
