import type React from "react";
import { OverlayTrigger, Tooltip } from "react-bootstrap";

export function deepEquals(value: any, other: any) {
    if (value === null || other === null) {
        return value === null && other === null;
    }
    if (Array.isArray(value)) {
        if (!Array.isArray(other)) {
            return false;
        }
        if (value.length != other.length) {
            return false;
        }
        for (var i = value.length - 1; i >= 0; i--) {
            if (!deepEquals(value[i], other[i])) {
                return false;
            }
        }
        return true;
    }
    if (["string", "boolean", "number", "undefined"].includes(typeof value)) {
        return value === other;
    }
    if (typeof value !== "object" || typeof other !== "object") {
        return false;
    }
    const otherKeys = Object.keys(other);
    if (Object.keys(value).length !== otherKeys.length) {
        return false;
    }
    for (const [key, subValue] of Object.entries(value)) {
        if (!otherKeys.includes(key)) {
            return false;
        }
        if (!deepEquals(subValue, other[key])) {
            return false;
        }
    }
    return true;
}

export function mapMap<K, V, R>(map: Map<K, V>, func: (val: V) => R): Map<K, R> {
    const newMap = new Map<K, R>();
    map.forEach((value, key) => newMap.set(key, func(value)));
    return newMap;
}

export function TooltipButton({
    children,
    iconClassName,
}: {
    children: React.ReactNode;
    iconClassName?: string;
}) {
    return (
        <OverlayTrigger
            trigger="click"
            overlay={
                <Tooltip>
                    <span className="disable-last-paragraph-spacing">{children}</span>
                </Tooltip>
            }
        >
            <button className="btn btn-sm py-0" type="button">
                <i className={"bi " + (iconClassName ?? "bi-info-circle")} />
            </button>
        </OverlayTrigger>
    );
}

export function AuthenticationMethodIcons({
    authentication_methods,
}: {
    authentication_methods: string[];
}) {
    if (authentication_methods.length == 0) {
        return (
            <span title="Nicht verfügbar" className="d-inline-block m-1 align-middle">
                <span className="bi bi-ban" aria-hidden="true" />
            </span>
        );
    }
    return (
        <>
            {authentication_methods.map((method) => {
                let icon = undefined;
                let description = "";
                switch (method) {
                    case "public": {
                        icon = <span className="bi bi-globe" aria-hidden="true" />;
                        description = "Öffentlich verfügbar";
                        break;
                    }
                    case "rwth": {
                        icon = <span className="icon-rwth" aria-hidden="true" />;
                        description = "Für RWTH-Angehörige und aus dem RWTH-Netz verfügbar";
                        break;
                    }
                    case "moodle": {
                        icon = <span className="icon-moodle" aria-hidden="true" />;
                        description = "Für Teilnehmer des Moodle-Lernraums verfügbar";
                        break;
                    }
                    case "fsmpi": {
                        icon = <span className="icon-fsmpi" aria-hidden="true" />;
                        description = "Für FSMPI-Angehörige verfügbar";
                        break;
                    }
                    case "password": {
                        icon = <span className="bi bi-lock-fill" aria-hidden="true" />;
                        description = "Mit Passwort verfügbar";
                        break;
                    }
                    default: {
                        // Should never happen
                        icon = <span className="bi bi-question-circle" aria-hidden="true" />;
                        description = "Unknown";
                        break;
                    }
                }
                return (
                    <span
                        key={method}
                        title={description}
                        className="d-inline-block m-1 align-middle"
                    >
                        {icon}
                    </span>
                );
            })}
        </>
    );
}

export function urlForLecture(course_id: string, lecture_id: string | number) {
    return `${course_id}/${lecture_id}`;
}
