import { useBackendContext } from "@/components/BackendProvider";
import { useEditMode } from "@/components/EditModeProvider";
import { FallbackErrorBoundary } from "@/components/FallbackErrorBoundary";
import { OMCreate, OMEdit, EmbeddedOMFieldComponent } from "@/components/OMConfigComponent";
import { ReloadBoundary } from "@/components/ReloadBoundary";
import { useUserContext } from "@/components/UserDataProvider";
import { semesterToHuman } from "@/misc/Formatting";
import { ResourceType, fetchDataWrapper } from "@/misc/PromiseHelpers";
import { ErrorPage } from "@/misc/ErrorHandlers";
import Link from "next/link";
import { Suspense, useEffect, useState } from "react";
import DropdownButton from "react-bootstrap/DropdownButton";
import DropdownItem from "react-bootstrap/DropdownItem";

import type { GetCoursesResponse, course } from "@/api/api_v1_types";
import { useLanguage } from "@/components/LanguageProvider";

type GroupByTypes = "semester" | "full_name" | "organizer" | "topic";

export function CourseItem({
    course,
    showSemester = true,
    ...props
}: {
    course: course;
    showSemester?: boolean;
    [key: string]: any;
}) {
    const { editMode } = useEditMode();
    let color = "";
    if (course.is_visible === false) {
        color = "not-visible";
    } else if (course.is_listed === false) {
        color = "not-listed";
    }

    return (
        <li
            className={
                "list-group-item list-group-item-condensed highlighting-list-item d-flex " + color
            }
            id={`course-${course.id}`}
            {...props}
        >
            {editMode && <OMEdit object_type="course" object_id={course.id} className="py-0" />}
            <div className="row flex-fill ps-2 align-self-center">
                {showSemester ? (
                    <>
                        <span className="col-2 col-md-1">
                            {semesterToHuman(course.semester, false)}
                        </span>
                        <span className="col-10 col-md-6">
                            <Link href={`/${course.id_string}`}>{course.full_name}</Link>
                        </span>
                    </>
                ) : (
                    <>
                        <span className="col-12 col-md-7">
                            <Link href={`/${course.id_string}`}>{course.full_name}</Link>
                        </span>
                    </>
                )}

                <span className="col-8 col-md-3">
                    <EmbeddedOMFieldComponent
                        object_type="course"
                        object_id={course.id}
                        field_id="organizer"
                        field_type="string"
                        initialValue={course.organizer}
                        allowMarkdown={true}
                    />
                </span>
                <span className="col-4 col-md-2">
                    <EmbeddedOMFieldComponent
                        object_type="course"
                        object_id={course.id}
                        field_id="topic"
                        field_type="string"
                        initialValue={course.topic}
                    />
                </span>
            </div>
        </li>
    );
}

function CourseGroupCard({
    g,
    groupedby,
    onlyShowPublicCourses,
}: {
    g: { groupTitle: string; list: Array<course> };
    groupedby: string;
    onlyShowPublicCourses: boolean;
}) {
    return (
        <div className="card mb-3">
            <div className="card-body">
                <h4 className="card-title">{g.groupTitle}</h4>
                <div className="card-text">
                    <ul className="courses-list m-0 p-0">
                        {g.list.flatMap((c) => {
                            if (
                                onlyShowPublicCourses &&
                                (!c.default_authentication_methods?.includes("public") ||
                                    c.is_visible === false)
                            ) {
                                // I chose not to check c.is_listed here because
                                // it may be useful for videoag admins to check which courses are falsely set to visible and public
                                return [];
                            }
                            return [
                                <CourseItem
                                    key={c.id}
                                    course={c}
                                    showSemester={groupedby !== "semester"}
                                />,
                            ];
                        })}
                    </ul>
                </div>
            </div>
        </div>
    );
}

function CourseList({
    courseData,
    groupBy,
    onlyShowPublicCourses,
}: {
    courseData: ResourceType<GetCoursesResponse>;
    groupBy: GroupByTypes;
    onlyShowPublicCourses: boolean;
}) {
    const data = courseData.read()!;

    const groups = Object.values(data.courses).reduce(
        (acc, course) => {
            const groupKey = course[groupBy] ?? "none";
            if (acc[groupKey] === undefined) {
                let groupTitle = groupKey;
                if (groupBy === "semester") {
                    groupTitle = semesterToHuman(course.semester, true);
                }
                acc[groupKey] = {
                    groupTitle,
                    list: [],
                };
            }
            acc[groupKey].list.push(course);
            return acc;
        },
        {} as {
            [key: string]: { groupTitle: string; list: Array<course> };
        },
    );

    return Object.entries(groups)
        .sort(([a, _1], [b, _2]) => {
            const cmp = a.toLowerCase().localeCompare(b.toLowerCase());
            // descending order for semesters
            return groupBy === "semester" ? -cmp : cmp;
        })
        .map(([key, groupObject]) => {
            return (
                <CourseGroupCard
                    key={key}
                    g={groupObject}
                    groupedby={groupBy}
                    onlyShowPublicCourses={onlyShowPublicCourses}
                />
            );
        });
}

export default function Courses() {
    const api = useBackendContext();
    const userContext = useUserContext();
    const hasUserInfo = userContext.hasUserInfo();
    const { editMode } = useEditMode();
    const [courseData, setCourseData] = useState<ResourceType<GetCoursesResponse>>();
    const { language } = useLanguage();
    const [onlyShowPublic, setOnlyShowPublic] = useState(false);

    const reloadData = () => {
        setCourseData(fetchDataWrapper(api.getCourses()));
    };

    useEffect(reloadData, [api, hasUserInfo]);

    const [groupedBy, setGroupedBy] = useState<GroupByTypes>("semester");

    if (courseData === undefined) return <></>;

    const groupNames: { name: string; gtype: GroupByTypes }[] = [
        { name: language.get("object.course.semester"), gtype: "semester" },
        { name: language.get("object.course.full_name"), gtype: "full_name" },
        { name: language.get("object.course.organizer"), gtype: "organizer" },
        { name: language.get("object.course.topic"), gtype: "topic" },
    ];

    return (
        <ReloadBoundary reloadFunc={reloadData}>
            <FallbackErrorBoundary
                fallback={(e: any) => (
                    <ErrorPage error={e} objectName="Kurse" objectPlural={true} />
                )}
            >
                <div className="col-12 p-0">
                    <ul className="list-inline d-flex align-items-center">
                        <DropdownButton
                            variant="primary"
                            title={language.get("ui.course_list.group_by")}
                            className="pe-0 me-1"
                        >
                            {groupNames.map((g) => (
                                <DropdownItem
                                    key={g.gtype}
                                    className={groupedBy === g.gtype ? "bg-primary" : ""}
                                    type="button"
                                    onClick={() => setGroupedBy(g.gtype)}
                                >
                                    {g.name}
                                </DropdownItem>
                            ))}
                        </DropdownButton>

                        <div className="ms-2 border border-light-subtle rounded">
                            <div className="form-check form-switch m-2">
                                <input
                                    className="form-check-input"
                                    type="checkbox"
                                    role="switch"
                                    id="only_public_switch"
                                    onChange={() => {
                                        setOnlyShowPublic((old) => !old);
                                    }}
                                    checked={onlyShowPublic}
                                />
                                <label className="form-check-label " htmlFor="only_public_switch">
                                    {language.get("ui.course_list.show_only_public_courses")}
                                </label>
                            </div>
                        </div>

                        <li className="flex-fill" />

                        {editMode && (
                            <li>
                                <OMCreate object_type="course">
                                    <button type="button" className="btn btn-secondary">
                                        <i className="bi bi-plus" />
                                        Neue Veranstaltung
                                    </button>
                                </OMCreate>
                            </li>
                        )}
                    </ul>
                </div>

                <Suspense fallback={<div className="spinner-border" />}>
                    <CourseList
                        courseData={courseData}
                        groupBy={groupedBy}
                        onlyShowPublicCourses={onlyShowPublic}
                    />
                </Suspense>
            </FallbackErrorBoundary>
        </ReloadBoundary>
    );
}
