import { Suspense, useEffect, useState } from "react";
import Four04 from "./404";

import { useBackendContext } from "@/components/BackendProvider";
import { ReloadBoundary } from "@/components/ReloadBoundary";
import { useUserContext } from "@/components/UserDataProvider";
import { usePathname } from "next/navigation";
import { Backend } from "@/api/Backend";
import { AuthenticationStatusResponse, GetCourseResponse } from "@/api/api_v1_types";
import { ResourceType, fetchDataWrapper } from "@/misc/PromiseHelpers";
import Player, { EmbeddedPlayer } from "@/components/Player";
import { ErrorPage } from "@/misc/ErrorHandlers";
import { FallbackErrorBoundary } from "@/components/FallbackErrorBoundary";

export interface PlayerData {
    course: GetCourseResponse;
    perms: AuthenticationStatusResponse;
    loaderHadUserInfo: boolean;
    lectureId: number;
}

async function dataLoader(
    api: Backend,
    course_id_string: string,
    lecture_id: string,
    hasUserInfo: boolean,
): Promise<PlayerData> {
    let course: GetCourseResponse = await api.getCourse(course_id_string, true);
    let lecture = course.lectures!.find((l) => l.id === parseInt(lecture_id));
    if (lecture === undefined) {
        throw new Error("Lecture not found");
    }

    let perms: AuthenticationStatusResponse = {
        authenticated_methods: ["public"],
        is_lecture_authenticated: true,
    };
    if (
        lecture.authentication_methods.includes("public") === false &&
        lecture.authentication_methods.length > 0
    ) {
        perms = await api.getAuthenticationStatus({ lecture_id: lecture.id });
    }
    return { course, perms, loaderHadUserInfo: hasUserInfo, lectureId: parseInt(lecture_id) };
}

// This redirector handles both the regular player and the embedded player
export function ActualRedirector() {
    const api = useBackendContext();
    const userContext = useUserContext();
    const hasUserInfo = userContext.hasUserInfo();
    const pathname = usePathname();
    const matchedParamsPlayer = pathname.match(/^\/([a-z0-9_-]+)\/([a-z0-9_-]+)$/);
    const matchedParamsEmbed = pathname.match(/^\/([a-z0-9_-]+)\/([a-z0-9_-]+)\/embed$/);
    const [playerData, setPlayerData] = useState<ResourceType<PlayerData>>();

    const reloadData = () => {
        if (!matchedParamsPlayer && !matchedParamsEmbed) return;
        const matched = (matchedParamsPlayer ?? matchedParamsEmbed)!;
        const courseId = matched[1];
        const lectureId = matched[2];
        setPlayerData(
            fetchDataWrapper(dataLoader(api, courseId, lectureId, userContext.hasUserInfo())),
        );
    };

    // eslint-disable-next-line react-hooks/exhaustive-deps
    useEffect(reloadData, [api, hasUserInfo, pathname]);

    if (!matchedParamsPlayer && !matchedParamsEmbed)
        return <Four04 title="Unbekannter Pfad" text="dynamic_player" />;
    if (matchedParamsPlayer && matchedParamsEmbed)
        throw new Error("Both matchedParamsPlayer and matchedParamsEmbed are true");
    if (!playerData) return <></>;

    return (
        <ReloadBoundary reloadFunc={reloadData}>
            <FallbackErrorBoundary
                fallback={(e: any) => (
                    <ErrorPage
                        error={e}
                        objectName="Vorlesung"
                        expectedErrorCodes={[
                            "unauthorized",
                            "access_forbidden",
                            "unknown_object",
                            "rate_limited",
                        ]}
                    />
                )}
            >
                <Suspense fallback={<div className="spinner-border" />}>
                    {matchedParamsEmbed && <EmbeddedPlayer playerData={playerData} />}
                    {matchedParamsPlayer && <Player playerData={playerData} />}
                </Suspense>
            </FallbackErrorBoundary>
        </ReloadBoundary>
    );
}

export default function DynamicPlayer() {
    // The following is a hack to force nextjs to not pre-render this page at build time
    // this needs to be done because it is not known which lecture is going to be watched (this is dynamic!)
    const [clientsideForcer, setClientsideForcer] = useState(false);
    useEffect(() => {
        setClientsideForcer(true);
    }, []);

    if (!clientsideForcer) return <></>;

    return <ActualRedirector />;
}
