import type React from "react";
import { useState } from "react";
import Link from "next/link";
import { useBackendContext } from "@/components/BackendProvider";
import { useLanguage } from "@/components/LanguageProvider";
import { showError } from "@/misc/ErrorHandlers";
import { StringEditor, BooleanEditor } from "@/components/TypeEditor";

export default function Feedback() {
    const api = useBackendContext();
    const { language } = useLanguage();
    const [textData, setTextData] = useState<[string, boolean]>(["", false]);
    const [emailData, setEmailData] = useState<[string | undefined, boolean]>([undefined, true]);
    const [dataConsent, setDataConsent] = useState<boolean>(false);
    const [hideErrors, setHideErrors] = useState<boolean>(true);

    const [isSending, setIsSending] = useState<boolean>(false);
    const [showSuccessfulMessageSent, setShowSuccessfulMessageSent] = useState<boolean>(false);

    const sendMessage = () => {
        setHideErrors(false);
        if (!dataConsent || !textData[1] || !emailData[1]) return;
        setIsSending(true);

        api.putNewFeedback({
            email: emailData[0],
            text: textData[0],
        })
            .then(() => {
                setIsSending(false);
                setShowSuccessfulMessageSent(true);
            })
            .catch((err) => {
                setIsSending(false);
                showError(err, language.get("ui.feedback.error.unable_to_send"));
            });
    };

    if (showSuccessfulMessageSent) {
        return (
            <>
                <h2>{language.get("ui.feedback.title")}</h2>
                <div className="alert alert-success">
                    {language.get("ui.feedback.sent_successfully")}
                </div>
                <Link href="/" className="btn btn-primary">
                    <span className="bi bi-chevron-right" />
                    Zur Startseite
                </Link>
            </>
        );
    }

    return (
        <>
            <h2>{language.get("ui.feedback.title")}</h2>
            <div>{language.getStyled("ui.feedback.info_text", true)}</div>
            <table className="table table-borderless w-100">
                <thead>
                    <tr>
                        <th className="m-1" />
                        <th className="w-100" />
                    </tr>
                </thead>
                <tbody>
                    <tr>
                        <td>{language.get("ui.feedback.message")}</td>
                        <td>
                            <StringEditor
                                value={textData[0]}
                                updateValue={(newValue, isValid, affectNow) =>
                                    setTextData([newValue, isValid])
                                }
                                autoFocus={true}
                                hideErrors={hideErrors}
                                minLength={1}
                                maxLength={16384}
                            />
                        </td>
                    </tr>
                    <tr>
                        <td>{language.get("ui.feedback.email")}</td>
                        <td>
                            <StringEditor
                                value={emailData[0]}
                                updateValue={(newValue, isValid, affectNow) =>
                                    setEmailData([newValue, isValid])
                                }
                                hideErrors={hideErrors}
                                regex={/^\S+@\S+\.\S+$/}
                                regexInvalidMessage={language.get(
                                    "ui.feedback.email.regex_invalid_message",
                                )}
                                allowUndefined={true}
                                maxLength={128}
                            />
                        </td>
                    </tr>
                    <tr>
                        <td></td>
                        <td>
                            <div className="float-start">
                                <BooleanEditor
                                    value={dataConsent}
                                    updateValue={(newValue, isValid, affectNow) =>
                                        setDataConsent(newValue)
                                    }
                                    hideErrors={hideErrors}
                                />
                            </div>
                            <div className="ms-4">
                                {language.getStyled("ui.feedback.consent", true)}
                            </div>
                            {!hideErrors && !dataConsent && (
                                <div className="ms-4 d-inline invalid-feedback">
                                    {language.get("ui.feedback.consent.required")}
                                </div>
                            )}
                        </td>
                    </tr>
                    <tr>
                        <td></td>
                        <td>
                            <button
                                className="btn btn-primary"
                                onClick={sendMessage}
                                disabled={
                                    isSending ||
                                    (!hideErrors && (!dataConsent || !textData[1] || !emailData[1]))
                                }
                            >
                                {language.get("ui.feedback.send")}
                            </button>
                            {isSending && (
                                <div className="spinner-border ms-2 align-middle" role="status" />
                            )}
                        </td>
                    </tr>
                </tbody>
            </table>
        </>
    );
}
