import { useBackendContext } from "@/components/BackendProvider";
import { OBJECT_TYPES } from "@/api/Backend";
import ModeratorBarrier from "@/components/ModeratorBarrier";
import { ReloadBoundary, useReloadBoundary } from "@/components/ReloadBoundary";
import { ChooserEditor, StringEditor, IntEditor } from "@/components/TypeEditor";
import { useFilteredDataView, PagingNavigation, FilterInput } from "@/components/FilteredData";
import { DateTime } from "luxon";
import { showError, ErrorPage } from "@/misc/ErrorHandlers";

import type { GetChangelogResponse, int } from "@/api/api_v1_types";

function ChangelogList({
    changelog,
    setFilter,
}: {
    changelog: GetChangelogResponse;
    setFilter: (
        key: string,
        new_value: any | undefined,
        isValid: boolean,
        updateNow: boolean,
    ) => void;
}) {
    const reloadFunc = useReloadBoundary();
    const api = useBackendContext();

    return (
        <>
            <div className="table-responsive">
                <table className="table table-hover" style={{ tableLayout: "fixed" }}>
                    <thead>
                        <tr>
                            <th scope="col" style={{ width: "12%" }}>
                                Zeit
                            </th>
                            <th scope="col" style={{ width: "5%" }}>
                                User
                            </th>
                            <th scope="col" style={{ width: "10%" }}>
                                Aktion
                            </th>
                            <th scope="col" style={{ width: "20%" }}>
                                Pfad
                            </th>
                            <th scope="col" style={{ width: "22%" }}>
                                alter Wert
                            </th>
                            <th scope="col" style={{ width: "22%" }}>
                                neuer Wert
                            </th>
                            <th scope="col" style={{ width: "9%" }}></th>
                        </tr>
                    </thead>
                    <tbody>
                        {(changelog?.page ?? []).map((i) => {
                            const filterLevel = (lvl: number) => {
                                return () => {
                                    setFilter(
                                        "object_type",
                                        lvl >= 1 ? i.object_type : undefined,
                                        true,
                                        false,
                                    );
                                    setFilter(
                                        "object_id",
                                        lvl >= 2 ? i.object_id : undefined,
                                        true,
                                        false,
                                    );
                                    setFilter(
                                        "field",
                                        lvl >= 3 ? i.field_description!.id : undefined,
                                        true,
                                        true,
                                    );
                                };
                            };
                            let pfad = <></>;
                            switch (i.type) {
                                case "modification":
                                    pfad = (
                                        <>
                                            <button
                                                type="button"
                                                className="btn btn-link p-0 align-baseline"
                                                onClick={filterLevel(1)}
                                            >
                                                {`${i.object_type}`}
                                            </button>
                                            .
                                            <button
                                                type="button"
                                                className="btn btn-link p-0 align-baseline"
                                                onClick={filterLevel(2)}
                                            >
                                                {`${i.object_id}`}
                                            </button>
                                            .
                                            <button
                                                type="button"
                                                className="btn btn-link p-0 align-baseline"
                                                onClick={filterLevel(3)}
                                            >{`${i.field_description!.id}`}</button>
                                            {` (${i.field_description?.type})`}
                                        </>
                                    );
                                    break;
                                case "creation":
                                    pfad = (
                                        <>
                                            <button
                                                type="button"
                                                className="btn btn-link p-0 align-baseline"
                                                onClick={filterLevel(1)}
                                            >
                                                {`${i.object_type}`}
                                            </button>
                                            .
                                            <button
                                                type="button"
                                                className="btn btn-link p-0 align-baseline"
                                                onClick={filterLevel(2)}
                                            >
                                                {`${i.object_id}`}
                                            </button>
                                            {typeof i.variant === "string" &&
                                                ` (variant:${i.variant})`}
                                        </>
                                    );
                                    break;
                                case "deletion_change":
                                    pfad = (
                                        <>
                                            <button
                                                type="button"
                                                className="btn btn-link p-0 align-baseline"
                                                onClick={filterLevel(1)}
                                            >
                                                {`${i.object_type}`}
                                            </button>
                                            .
                                            <button
                                                type="button"
                                                className="btn btn-link p-0 align-baseline"
                                                onClick={filterLevel(2)}
                                            >
                                                {`${i.object_id}`}
                                            </button>
                                        </>
                                    );
                                    break;
                                case "unknown":
                                    pfad = (
                                        <>{`${i.unknown_type}.${i.unknown_field}, id=${i.object_id}`}</>
                                    );
                                    break;
                                default:
                                    pfad = <>{i.type}</>;
                            }
                            const valToStr = (val: any) => {
                                if (val === null) {
                                    return <span className="fst-italic text-muted">null</span>;
                                }
                                if (val === undefined) {
                                    return <span className="fst-italic text-muted">undefined</span>;
                                }
                                return JSON.stringify(val);
                            };
                            const formattedDate = DateTime.fromISO(i.change_time).toFormat(
                                "yyyy-MM-dd HH:mm:ss",
                            );
                            const undo = () => {
                                if (i.type !== "modification") {
                                    alert("TODO: Not implemented");
                                    return;
                                }
                                if (!confirm("Wirklich zurücksetzen?")) {
                                    return;
                                }
                                let updates: { [key: string]: any } = {};
                                updates[i.field_description!.id] = i.old_value;

                                api.updateOMObject(i.object_type!, i.object_id!, {
                                    updates: updates,
                                    expected_current_values: {}, //TODO have something here?
                                })
                                    .then(reloadFunc)
                                    .catch((e) => {
                                        showError(e, "Exception while reverting change");
                                    });
                            };
                            return (
                                <tr key={i.id}>
                                    <td>{formattedDate}</td>
                                    <td>{i.modifying_user_id}</td>
                                    <td>{i.type}</td>
                                    <td>{pfad}</td>
                                    <td>{valToStr(i.old_value)}</td>
                                    <td>{valToStr(i.new_value)}</td>
                                    <td>
                                        {i.type === "modification" && (
                                            <button className="btn btn-warning" onClick={undo}>
                                                revert to old
                                            </button>
                                        )}
                                    </td>
                                </tr>
                            );
                        })}
                    </tbody>
                </table>
            </div>
        </>
    );
}

function ChangelogImpl() {
    const api = useBackendContext();
    const { filters, setFilter, reloadData, isLoading, data, error } = useFilteredDataView(
        (filters) =>
            api.getChangelog({
                entries_per_page: filters.get("entries_per_page") as int | undefined,
                page: filters.get("page") as int | undefined,
                object_type: filters.get("object_type") as string | undefined,
                object_id: filters.get("object_id") as int | undefined,
                field: filters.get("field") as string | undefined,
            }),
    );
    return (
        <div className="card">
            <div className="card-header d-flex">Changelog</div>
            <div className="card-body">
                <p>
                    Hier werden alle Änderungen an Kursen/Veranstaltungen/Videos etc. geloggt und
                    können Rückgängig gemacht werden.
                </p>
                <ReloadBoundary reloadFunc={reloadData}>
                    <div className="d-flex">
                        <FilterInput
                            component={(props) => (
                                <ChooserEditor
                                    langKeyPrefix="object"
                                    unchosenLangKey="ui.generic.filter.any"
                                    possibleValues={OBJECT_TYPES}
                                    {...props}
                                />
                            )}
                            label="Object Type"
                            filterId="object_type"
                            setFilter={setFilter}
                            filters={filters}
                        />
                        <FilterInput
                            component={(props) => <IntEditor allowUndefined={true} {...props} />}
                            label="Object Id"
                            filterId="object_id"
                            setFilter={setFilter}
                            filters={filters}
                        />
                        {/*TODO suggestions? handle errors? handle no object type*/}
                        <FilterInput
                            component={(props) => (
                                <StringEditor maxLength={100} allowUndefined={true} {...props} />
                            )}
                            label="Field"
                            filterId="field"
                            setFilter={setFilter}
                            filters={filters}
                        />
                    </div>
                    <div className="d-flex">
                        <PagingNavigation
                            setFilter={setFilter}
                            filters={filters}
                            pageCount={data?.page_count}
                        />
                        {isLoading && <div className="spinner-border ms-3 mt-1" />}
                    </div>
                    {error ? (
                        <ErrorPage error={error} objectName="Changelog" showButtons={false} />
                    ) : (
                        data && <ChangelogList changelog={data} setFilter={setFilter} />
                    )}
                </ReloadBoundary>
            </div>
        </div>
    );
}

export default function Changelog() {
    return (
        <ModeratorBarrier>
            <ChangelogImpl />
        </ModeratorBarrier>
    );
}
