import { useBackendContext } from "@/components/BackendProvider";
import ModeratorBarrier from "@/components/ModeratorBarrier";
import { ReloadBoundary, useReloadBoundary } from "@/components/ReloadBoundary";
import { ICalEvent, parseICal } from "@/misc/Calendar";
import { DateTime } from "luxon";
import Link from "next/link";
import { useSearchParams } from "next/navigation";
import { MutableRefObject, useEffect, useRef, useState } from "react";
import { showError, showErrorToast } from "@/misc/ErrorHandlers";

import type { GetNewConfigurationResponse, course } from "@/api/api_v1_types";

function TerminBody({
    course,
    imported_events,
}: {
    course: course;
    imported_events: MutableRefObject<ICalEvent[]>;
}) {
    const api = useBackendContext();
    const reloadFunc = useReloadBoundary();
    const [lockedIds, setLockedIds] = useState<string[] | undefined>([]); // undefined has the special meaning that everything is locked

    const existingEvents: ICalEvent[] = (course.lectures ?? []).map((l) => {
        const obj: ICalEvent = {
            location: l.location,
            duration: l.duration,
            startDate: DateTime.fromISO(l.time),
            id_in_videoag_system: l.id,
            locally_unique_id: "videoag_" + l.id,
        };
        return obj;
    });

    const eventsSimilar = (a: ICalEvent, b: ICalEvent) => {
        return (
            a.location === b.location &&
            a.duration === b.duration &&
            Math.abs(a.startDate.diff(b.startDate).as("seconds")) < 1
        );
    };

    const removeDuplicates = (event: ICalEvent, index: number, self: ICalEvent[]) => {
        return index === self.findIndex((e) => eventsSimilar(e, event));
    };
    const notContainedIn = (other: ICalEvent[]) => {
        return (event: ICalEvent) => {
            return !other.some((oth) => {
                return eventsSimilar(event, oth);
            });
        };
    };
    const containedIn = (other: ICalEvent[]) => {
        const nCi = notContainedIn(other);
        return (event: ICalEvent) => !nCi(event);
    };

    // find events that are new (place, duration and time unique)
    const new_events = imported_events.current
        .filter(notContainedIn(existingEvents))
        .filter(removeDuplicates);
    // find existing events that are not in the imported events
    const not_imported_events = existingEvents
        .filter(notContainedIn(imported_events.current))
        .filter(removeDuplicates);
    // events that are in both
    const full_matches = imported_events.current
        .filter(containedIn(existingEvents))
        .filter(removeDuplicates).length;

    const importEvent = (event: ICalEvent) => {
        setLockedIds((old) => {
            if (old === undefined) {
                return old;
            } else {
                return old.concat(event.locally_unique_id);
            }
        });
        api.getNewOMConfiguration("lecture")
            .then((res: GetNewConfigurationResponse) => {
                const defaultValues: any = {};
                res.fields?.forEach((field) => {
                    if (field.default_value !== undefined) {
                        defaultValues[field.id] = field.default_value;
                    } else if (
                        field.type === "string" &&
                        (field.min_length === undefined || field.min_length <= 0)
                    ) {
                        defaultValues[field.id] = "";
                    }
                });

                const convertedTime = event.startDate?.toFormat("yyyy-MM-dd'T'HH:mm:ss");
                return api.createOMObject("lecture", {
                    parent_type: "course",
                    parent_id: course.id,
                    values: {
                        ...defaultValues,
                        title: event.summary,
                        location: event.location,
                        time: convertedTime,
                        duration: event.duration,
                    },
                });
            })
            .catch((e) => {
                showError(e, "Unable to create lecture");
            })
            .then(reloadFunc)
            .finally(() => {
                setLockedIds((old) => {
                    if (old === undefined) {
                        return old;
                    } else {
                        return old.filter((id) => id !== event.locally_unique_id);
                    }
                });
            });
    };

    const importAllEvents = () => {
        setLockedIds(undefined);
        api.getNewOMConfiguration("lecture")
            .then((res: GetNewConfigurationResponse) => {
                const defaultValues: any = {};
                res.fields?.forEach((field) => {
                    if (field.default_value !== undefined) {
                        defaultValues[field.id] = field.default_value;
                    } else if (
                        field.type === "string" &&
                        (field.min_length === undefined || field.min_length <= 0)
                    ) {
                        defaultValues[field.id] = "";
                    }
                });

                return Promise.allSettled(
                    new_events.map((event) => {
                        const convertedTime = event.startDate?.toFormat("yyyy-MM-dd'T'HH:mm:ss");
                        return api.createOMObject("lecture", {
                            parent_type: "course",
                            parent_id: course.id,
                            values: {
                                ...defaultValues,
                                title: event.summary,
                                location: event.location,
                                time: convertedTime,
                                duration: event.duration,
                            },
                        });
                    }),
                );
            })
            .then((results) => {
                let failed = 0;
                results.forEach((result, idx) => {
                    if (result.status === "rejected") {
                        failed++;
                    }
                });
                if (failed > 0) {
                    showErrorToast(
                        `${failed} lectures could not be imported, see console for details`,
                    );
                }
            })
            .then(reloadFunc)
            .finally(() => {
                setLockedIds([]);
            });
    };

    const removeEvent = (event: ICalEvent) => {
        if (!confirm("Really delete?")) return;
        if (event.id_in_videoag_system === undefined) {
            showErrorToast("Event not found");
            return;
        }
        setLockedIds((old) => {
            if (old === undefined) {
                return old;
            } else {
                return old.concat(event.locally_unique_id);
            }
        });
        api.deleteOMObject("lecture", event.id_in_videoag_system)
            .catch((e) => {
                showError(e, "Unable to delete lecture");
            })
            .then(reloadFunc)
            .finally(() => {
                setLockedIds((old) => {
                    if (old === undefined) {
                        return old;
                    } else {
                        return old.filter((id) => id !== event.locally_unique_id);
                    }
                });
            });
    };

    const removeAllEvents = () => {
        if (!confirm("Really delete all?")) return;
        setLockedIds(undefined);
        const promises = existingEvents.flatMap((event) => {
            if (event.id_in_videoag_system === undefined) {
                return [];
            }

            return [api.deleteOMObject("lecture", event.id_in_videoag_system)];
        });
        Promise.allSettled(promises)
            .then((results) => {
                let failed = 0;
                results.forEach((result, idx) => {
                    if (result.status === "rejected") {
                        failed++;
                    }
                });
                if (failed > 0) {
                    showErrorToast(
                        `${failed} lectures could not be deleted, see console for details`,
                    );
                }
            })
            .then(reloadFunc)
            .finally(() => {
                setLockedIds([]);
            });
    };

    return (
        <div>
            <table className="table table-bordered w-100 table-sm text-center">
                <thead>
                    <tr>
                        <th scope="col">Nur im Import</th>
                        <th scope="col">In beiden</th>
                        <th scope="col">Nur bei uns</th>
                    </tr>
                </thead>
                <tbody>
                    <tr>
                        <td className="text-warning">{new_events.length}</td>
                        <td className="text-success">{full_matches}</td>
                        <td className="text-warning">{not_imported_events.length}</td>
                    </tr>
                </tbody>
            </table>

            {lockedIds === undefined || lockedIds.length > 0 ? (
                <div className="spinner-border mb-2" role="status" />
            ) : null}

            <div className="card mb-2">
                <div className="card-header d-flex align-items-center">
                    <span className="flex-grow-1">Im Import, nicht bei uns</span>

                    <button
                        className="btn btn-primary"
                        onClick={importAllEvents}
                        disabled={lockedIds === undefined || lockedIds.length > 0}
                    >
                        Alle anlegen
                    </button>
                </div>
                <div className="card-body">
                    <table className="table">
                        <thead>
                            <tr>
                                <th>Zeit</th>
                                <th>Ort</th>
                                <th>Dauer</th>
                                <th></th>
                            </tr>
                        </thead>
                        <tbody>
                            {new_events.map((event, index) => (
                                <tr key={index}>
                                    <td>
                                        {event.startDate
                                            ? event.startDate.toFormat("yyyy-MM-dd HH:mm:ss")
                                            : "kein datum"}
                                    </td>
                                    <td>{event.location}</td>
                                    <td>{event.duration}</td>
                                    <td>
                                        <button
                                            className="btn btn-primary"
                                            onClick={() => importEvent(event)}
                                            disabled={
                                                lockedIds === undefined ||
                                                lockedIds.includes(event.locally_unique_id)
                                            }
                                        >
                                            Anlegen
                                        </button>
                                    </td>
                                </tr>
                            ))}
                        </tbody>
                    </table>
                </div>
            </div>
            <div className="card">
                <div className="card-header d-flex align-items-center">
                    <span className="flex-grow-1">Bei uns, nicht im Import</span>

                    <button
                        className="btn btn-primary"
                        onClick={removeAllEvents}
                        disabled={lockedIds === undefined || lockedIds.length > 0}
                    >
                        Alle entfernen
                    </button>
                </div>
                <div className="card-body">
                    <table className="table">
                        <thead>
                            <tr>
                                <th>Zeit</th>
                                <th>Ort</th>
                                <th>Dauer</th>
                                <th></th>
                            </tr>
                        </thead>
                        <tbody>
                            {not_imported_events.map((event, index) => (
                                <tr key={index}>
                                    <td>
                                        {event.startDate
                                            ? event.startDate.toFormat("yyyy-MM-dd HH:mm:ss")
                                            : "kein datum"}
                                    </td>
                                    <td>{event.location}</td>
                                    <td>{event.duration}</td>
                                    <td>
                                        <button
                                            className="btn btn-primary"
                                            onClick={() => removeEvent(event)}
                                            disabled={
                                                lockedIds === undefined ||
                                                lockedIds.includes(event.locally_unique_id)
                                            }
                                        >
                                            Entfernen
                                        </button>
                                    </td>
                                </tr>
                            ))}
                        </tbody>
                    </table>
                </div>
            </div>
        </div>
    );
}

function RWTHOnlineImportImpl() {
    const api = useBackendContext();
    const searchParam = useSearchParams();
    const course_id = searchParam.get("course_id");
    const [course, setCourse] = useState<course>();
    const imported_events = useRef<ICalEvent[]>([]);
    const imported_courses = useRef<number[]>([]);
    const inputRef = useRef<HTMLInputElement>(null);
    const [isImporting, setIsImporting] = useState(false);
    const [_, doReRender] = useState(0);
    const exampleImportUrl =
        "https://online.rwth-aachen.de/RWTHonline/pl/ui/%24ctx/wbTvw_List.lehrveranstaltung?pStpSpNr=...";

    const reloadData = () => {
        if (course_id === null) return;

        api.getCourse(course_id, true)
            .then(setCourse)
            .catch((e) => showError(e, "Unable to load course"));
    };

    useEffect(reloadData, [course_id, api]);

    if (course_id === null) {
        return (
            <div>
                <p>Invalid ?course_id</p>
            </div>
        );
    }

    if (course === undefined) {
        return (
            <div>
                <p>Loading...</p>
            </div>
        );
    }

    const onImport = () => {
        if (inputRef.current === null) return;

        const url = inputRef.current.value;

        let courseIdNr: number | undefined;
        if (url.includes("pStpSpNr=")) {
            courseIdNr = parseInt(url.split("pStpSpNr=")[1].split("&")[0]);
        } else {
            // https://online.rwth-aachen.de/RWTHonline/ee/ui/ca2/app/desktop/#/slc.tm.cp/student/courses/487373?$ctx=lang=de&$scrollTo=toc_overview
            const regex = /online\.rwth-aachen\.de\/.*\/courses\/(\d+)/;
            const match = url.match(regex);
            if (match) {
                courseIdNr = parseInt(match[1]);
            } else {
                showErrorToast(
                    `The provided URL is invalid. Must be of format: ${exampleImportUrl}`,
                );
                return;
            }
        }
        if (courseIdNr === undefined) {
            showErrorToast(`The provided URL is invalid. Must be of format: ${exampleImportUrl}`);
            return;
        }
        if (imported_courses.current.includes(courseIdNr)) {
            showErrorToast("The specified course is already in the import");
            return;
        }

        setIsImporting(true);
        api.importRWTHOnlineCourse(courseIdNr)
            .then((ical) => {
                imported_courses.current.push(courseIdNr!);
                const parsed = parseICal(ical);
                // add id to events
                parsed.forEach((event, idx) => {
                    event.id_in_foreign_system = courseIdNr;
                    event.locally_unique_id = "rwthonline_" + courseIdNr + "_" + idx;
                });
                // append
                imported_events.current = imported_events.current.concat(parsed);
            })
            .catch((e) => {
                showError(e, "Unable to import lectures from RWTHonline");
            })
            .then(() => {
                setIsImporting(false);
            });
    };

    const removeImportedCourse = (courseId: number) => {
        if (!confirm("Wirklich löschen?")) return;

        imported_courses.current = imported_courses.current.filter((id) => id !== courseId);
        imported_events.current = imported_events.current.filter(
            (event) => event.id_in_foreign_system !== courseId,
        );
        doReRender((r) => r + 1);
    };

    return (
        <ReloadBoundary reloadFunc={reloadData}>
            <Link href={`/${course.id_string}`} className="btn btn-primary mb-2">
                <span className="bi bi-chevron-left" /> Zum Kurs
            </Link>
            <div className="card mb-2">
                <div className="card-header d-flex">Import von RWTHOnline</div>
                <div className="card-body">
                    <a
                        href="https://online.rwth-aachen.de/RWTHonline/pl/ui/%24ctx/wbSuche.LVSuche"
                        target="_blank"
                    >
                        Suche den Kurs auf RWTHOnline
                    </a>
                    , und kopiere die URL hier rein:
                    <div>
                        <input
                            type="text"
                            className="w-100"
                            placeholder={exampleImportUrl}
                            ref={inputRef}
                        />
                        {isImporting ? (
                            <div className="spinner-border mt-2" role="status" />
                        ) : (
                            <button
                                className="btn btn-primary mt-2"
                                type="button"
                                onClick={onImport}
                            >
                                Zum Import hinzufügen
                            </button>
                        )}
                    </div>
                    <div>
                        Bereits importierte Termine: {imported_events.current.length}
                        <table className="table table-bordered">
                            <tbody>
                                {imported_courses.current.map((courseId, index) => (
                                    <tr key={index}>
                                        <td>
                                            {courseId}

                                            <button
                                                type="button"
                                                className="ms-2 btn btn-danger"
                                                onClick={() => removeImportedCourse(courseId)}
                                            >
                                                <i className="bi bi-trash-fill" />
                                            </button>
                                        </td>
                                    </tr>
                                ))}
                            </tbody>
                        </table>
                    </div>
                </div>
            </div>
            <TerminBody course={course} imported_events={imported_events} />
        </ReloadBoundary>
    );
}

export default function RWTHOnlineImport() {
    return (
        <ModeratorBarrier>
            <RWTHOnlineImportImpl />
        </ModeratorBarrier>
    );
}
