import { useBackendContext } from "@/components/BackendProvider";
import { JOB_STATES, JOB_TYPES } from "@/api/Backend";
import ModeratorBarrier from "@/components/ModeratorBarrier";
import { ChooserEditor, StringEditor } from "@/components/TypeEditor";
import { useFilteredDataView, PagingNavigation, FilterInput } from "@/components/FilteredData";
import { ReloadBoundary, useReloadBoundary } from "@/components/ReloadBoundary";
import { showError, showErrorToast, ErrorPage } from "@/misc/ErrorHandlers";

import type { GetJobsResponse, job, job_state, job_type, int } from "@/api/api_v1_types";

function Worker({ workers }: { workers: GetJobsResponse["workers"] }) {
    let now = new Date();
    // figure out which worker is online
    const dwork = workers.map((i) => {
        let last_ping = new Date(i.last_ping);
        let td = now.getTime() - last_ping.getTime();
        let className = "";
        if (td < 30) {
            className = "bg-success-subtle";
        } else if (td < 60) {
            className = "bg-warning-subtle";
        } else {
            className = "bg-danger-subtle";
        }
        return (
            <tr key={i.hostname}>
                <td className={className}>{i.hostname}</td>
                <td className={className}>{i.last_ping}</td>
            </tr>
        );
    });

    return (
        <div className="card mb-2">
            <div className="card-header d-flex">Worker</div>
            <div className="card-body">
                <table className="table table-hover">
                    <thead>
                        <tr>
                            <th scope="col">Hostname</th>
                            <th scope="col">letzter Ping</th>
                        </tr>
                    </thead>
                    <tbody>{dwork}</tbody>
                </table>
            </div>
        </div>
    );
}

function Jobs({ jobs }: { jobs: job[] }) {
    const api = useBackendContext();
    const reloadFunc = useReloadBoundary();

    return (
        <div className="table-responsive">
            <table className="table table-hover">
                <thead>
                    <tr>
                        <th>ID</th>
                        <th>Type</th>
                        <th>Priority</th>
                        <th>Worker</th>
                        <th>letzter Ping</th>
                        <th>State</th>
                        <th>Eingereiht am</th>
                        <th>Fertig geworden am</th>
                        <th>Gescheduled am</th>
                        <th>Daten</th>
                        <th>Status</th>
                        <th></th>
                    </tr>
                </thead>
                <tbody>
                    {jobs.map((i) => {
                        let className = "warning";
                        if (i.last_worker_ping) {
                            let last_ping = new Date(i.last_worker_ping);
                            let td = new Date().getTime() - last_ping.getTime();

                            if (td < 20) {
                                className = "success";
                            } else if (td < 60) {
                                className = "warning";
                            } else {
                                className = "danger";
                            }
                        }

                        const action = (action: "restart" | "cancel" | "copy" | "delete") => {
                            if (action === "delete") {
                                if (
                                    !confirm("Sind Sie sicher, dass Sie den Job löschen möchten?")
                                ) {
                                    return;
                                }
                                api.deleteJob(i.id)
                                    .then(reloadFunc)
                                    .catch((e) => {
                                        showError(e, "Exception while deleting job");
                                    });
                            } else {
                                api.jobAction(i.id, action)
                                    .then(reloadFunc)
                                    .catch((e) => {
                                        showError(e, "Exception while executing action");
                                    });
                            }
                        };

                        return (
                            <tr key={i.id} className={className}>
                                <td>{i.id}</td>
                                <td>{i.type}</td>
                                <td>{i.priority}</td>
                                <td>{i.worker_id}</td>
                                <td>{i.last_worker_ping}</td>
                                <td>
                                    {i.state}
                                    {i.is_canceled ? " (canceled)" : ""}
                                </td>
                                <td>{i.creation_time}</td>
                                <td>{i.finish_time}</td>
                                <td>{i.start_time}</td>
                                <td>{i.data}</td>
                                <td>{i.status}</td>
                                <td>
                                    <ul className="list-inline" style={{ whiteSpace: "nowrap" }}>
                                        {i.state === "failed" && (
                                            <li>
                                                <button
                                                    className="btn btn-primary"
                                                    type="button"
                                                    onClick={() => action("restart")}
                                                    title="Neustarten"
                                                >
                                                    <i className="bi bi-arrow-repeat" />
                                                </button>
                                            </li>
                                        )}
                                        {i.state === "running" && (
                                            <li>
                                                <button
                                                    className="btn btn-danger"
                                                    type="button"
                                                    onClick={() => action("cancel")}
                                                    title="Abbrechen"
                                                >
                                                    <i className="bi bi-stop" />
                                                </button>
                                            </li>
                                        )}
                                        {(i.state === "finished" || i.state === "deleted") && (
                                            <li>
                                                <button
                                                    className="btn btn-primary"
                                                    type="button"
                                                    onClick={() => action("copy")}
                                                    title="Kopie einreihen"
                                                >
                                                    <i className="bi bi-copy" />
                                                </button>
                                            </li>
                                        )}
                                        {(i.state === "ready" || i.state === "failed") && (
                                            <li>
                                                <button
                                                    className="btn btn-danger"
                                                    type="button"
                                                    onClick={() => action("delete")}
                                                    title="Löschen"
                                                >
                                                    <i className="bi bi-trash" />
                                                </button>
                                            </li>
                                        )}
                                    </ul>
                                </td>
                            </tr>
                        );
                    })}
                </tbody>
            </table>
        </div>
    );
}

function JobsOverviewImpl() {
    const api = useBackendContext();
    const { filters, setFilter, reloadData, isLoading, data, error } = useFilteredDataView(
        (filters) =>
            api.getJobs({
                entries_per_page: filters.get("entries_per_page") as int | undefined,
                page: filters.get("page") as int | undefined,
                type: filters.get("type") as job_type | undefined,
                worker_id: filters.get("worker_id") as string | undefined,
                state: filters.get("state") as job_state | undefined,
            }),
    );
    const createSubSetValue =
        (filterId: string) => (val: string | undefined, isValid: boolean, updateNow: boolean) =>
            setFilter(filterId, val, isValid, updateNow);

    const newJob = () => {
        alert("TODO");
    };
    const failed = (action: "delete" | "restart") => {
        if (!data) {
            showErrorToast("No jobs loaded, please try again later");
            return;
        }
        if (!confirm("Sind Sie sicher, dass Sie alle fehlgeschlagenen Jobs löschen möchten?")) {
            return;
        }
        Promise.allSettled(
            data.page.flatMap((i) => {
                if (i.state === "failed") {
                    if (action === "restart") {
                        return [api.jobAction(i.id, "restart")];
                    } else {
                        return [api.deleteJob(i.id)];
                    }
                }
                return [];
            }),
        )
            .then(reloadData)
            .catch((e) => {
                showError(
                    e,
                    `Exception while ${action === "delete" ? "deleting" : "restarting"} a job`,
                );
            });
    };

    return (
        <ReloadBoundary reloadFunc={reloadData}>
            <Worker workers={data ? data.workers : []} />
            <div className="card">
                <div className="card-header d-flex">Jobs</div>
                <div className="card-body">
                    <div className="d-flex mb-2">
                        <button type="button" className="btn btn-primary me-1" onClick={newJob}>
                            Job(s) erzeugen
                        </button>
                        <button
                            type="button"
                            className="btn btn-primary me-1"
                            onClick={() => failed("delete")}
                            disabled={data === undefined}
                        >
                            Alle fehlgeschlagenen Jobs löschen
                        </button>
                        <button
                            type="button"
                            className="btn btn-primary me-1"
                            onClick={() => failed("restart")}
                            disabled={data === undefined}
                        >
                            Alle fehlgeschlagenen Jobs neustarten
                        </button>
                    </div>
                    <div className="d-flex">
                        <FilterInput
                            component={(props) => (
                                <ChooserEditor
                                    unchosenLangKey="ui.generic.filter.any"
                                    possibleValues={JOB_TYPES}
                                    {...props}
                                />
                            )}
                            label="Type"
                            filterId="type"
                            setFilter={setFilter}
                            filters={filters}
                        />
                        <FilterInput
                            component={(props) => (
                                <StringEditor
                                    maxLength={100}
                                    allowUndefined={true}
                                    placeholder="'none' für keinen Worker"
                                    {...props}
                                />
                            )}
                            label="Worker"
                            filterId="worker_id"
                            setFilter={setFilter}
                            filters={filters}
                        />
                        <FilterInput
                            component={(props) => (
                                <ChooserEditor
                                    unchosenLangKey="ui.generic.filter.any"
                                    possibleValues={JOB_STATES}
                                    {...props}
                                />
                            )}
                            label="State"
                            filterId="state"
                            setFilter={setFilter}
                            filters={filters}
                        />
                    </div>
                    <div className="d-flex">
                        <PagingNavigation
                            setFilter={setFilter}
                            filters={filters}
                            pageCount={data?.page_count}
                        />
                        {isLoading && <div className="spinner-border ms-3 mt-1" />}
                    </div>
                    {error ? (
                        <ErrorPage
                            error={error}
                            objectName="Jobs"
                            objectPlural={true}
                            showButtons={false}
                        />
                    ) : (
                        <Jobs jobs={data ? data.page : []} />
                    )}
                </div>
            </div>
        </ReloadBoundary>
    );
}

export default function JobsOverview() {
    return (
        <ModeratorBarrier>
            <JobsOverviewImpl />
        </ModeratorBarrier>
    );
}
