import { useBackendContext } from "@/components/BackendProvider";
import ModeratorBarrier from "@/components/ModeratorBarrier";
import { DateTime } from "luxon";
import Link from "next/link";
import { useEffect, useState } from "react";
import { showError, ErrorPage } from "@/misc/ErrorHandlers";
import { ReloadBoundary } from "@/components/ReloadBoundary";

import type { GetCoursesResponse, GetSorterLogResponse } from "@/api/api_v1_types";

function SortLogImpl() {
    const api = useBackendContext();
    const [sortlog, setSortlog] = useState<GetSorterLogResponse>();
    const [error, setError] = useState<any>();
    const [cachedCourses, setCachedCourses] = useState<GetCoursesResponse>();
    const [cachedLectures, setCachedLectures] = useState<Record<string, string>>({});

    const refreshData = () => {
        api.getSorterLog()
            .then((data) => {
                setError(undefined);
                setSortlog(data);
            })
            .catch((e) => {
                setError(e);
            });
    };

    useEffect(() => {
        if (!sortlog) return;

        if (cachedCourses === undefined) {
            api.getCourses()
                .then((data) => {
                    setCachedCourses(data);
                })
                .catch((e) => {
                    showError(e, "Exception while fetching courses");
                });
        }

        const lectureIds = new Set<number>();

        sortlog.error_log.forEach((i) => {
            i.matched_lectures_ids?.forEach((j) => lectureIds.add(j));
        });

        for (const i of Object.keys(cachedLectures)) {
            if (lectureIds.has(parseInt(i))) lectureIds.delete(parseInt(i));
        }

        let iter = lectureIds.values();
        let first = iter.next();
        if (first.done) return;

        const fetchLecture = (id: number) => {
            return api
                .getCourseLecture(id)
                .then((data) => {
                    setCachedLectures((prev) => {
                        return { ...prev, [id]: data.lecture.title };
                    });
                })
                .catch((e) => {
                    showError(e, "Exception while fetching lecture");
                });
        };
        let promise = fetchLecture(first.value);
        while (true) {
            const next = iter.next();
            if (next.done) break;
            promise = promise.then(() => fetchLecture(next.value));
        }
        // eslint-disable-next-line react-hooks/exhaustive-deps
    }, [api, sortlog]);

    useEffect(refreshData, [api]);

    const removeAllErrors = () => {
        if (!sortlog) return;

        Promise.allSettled(
            sortlog.error_log.map((i) => {
                return api.deleteSorterErrorLog(i.id);
            }),
        )
            .catch((e) => {
                showError(e, "Exception while deleting error log entry");
            })
            .then(refreshData);
    };

    const getCourseName = (id: number) => {
        if (cachedCourses === undefined) return "...";
        const course = Object.values(cachedCourses.courses).find((v) => v.id === id);
        return course ? course.full_name : "???";
    };

    const getLectureName = (id: number) => {
        return cachedLectures[id.toString()] || "???";
    };

    if (error) {
        return (
            <ReloadBoundary reloadFunc={refreshData}>
                <ErrorPage error={error} objectName="Sortierlog" />
            </ReloadBoundary>
        );
    }

    return (
        <div className="card">
            <div className="card-header d-flex">Sortierlog</div>
            <div className="card-body">
                <button
                    className="btn btn-primary"
                    type="button"
                    onClick={removeAllErrors}
                    disabled={!sortlog || sortlog.error_log.length === 0}
                >
                    Alle Fehler entfernen
                </button>

                <p>
                    Hier werden die hochgeladenen Videos einsortiert und geloggt wo jede Datei
                    einsortiert wurde.
                </p>

                <div className="table-responsive" style={{ maxHeight: "250px" }}>
                    <table className="table table-hover table-bordered">
                        <thead>
                            <tr>
                                <th scope="col"></th>
                                <th scope="col">Zeit</th>
                                <th scope="col">Pfad</th>
                                <th scope="col">Course</th>
                                <th scope="col">Matches</th>
                            </tr>
                        </thead>
                        <tbody>
                            {sortlog?.error_log.map((i) => {
                                const formattedDate = DateTime.fromISO(i.time).toFormat(
                                    "yyyy-MM-dd HH:mm:ss",
                                );
                                const deleteObject = () => {
                                    api.deleteSorterErrorLog(i.id)
                                        .catch((e) => {
                                            showError(
                                                e,
                                                "Exception while deleting error log entry",
                                            );
                                        })
                                        .then(refreshData);
                                };
                                return (
                                    <tr className="danger" key={i.id}>
                                        <td>
                                            <button
                                                type="button"
                                                className={"btn btn-outline-danger mx-1"}
                                                onClick={deleteObject}
                                                title="Delete"
                                            >
                                                <i className="bi bi-trash-fill" />
                                            </button>{" "}
                                        </td>
                                        <td>{formattedDate}</td>
                                        <td>
                                            <a href={`${api.assetUrl()}/${i.file_path}`}>
                                                {i.file_path}
                                            </a>
                                        </td>
                                        <td>
                                            <Link href={`/${i.matched_course_id}`}>
                                                {`${getCourseName(i.matched_course_id!)} (${i.matched_course_id})`}
                                            </Link>
                                        </td>
                                        <td>
                                            {i.matched_lectures_ids?.map((j, ind) => (
                                                <span key={j}>
                                                    <Link
                                                        href={`/${i.matched_course_id}#lecture-${j}`}
                                                    >
                                                        {ind === 0 ? "" : ", "}
                                                        {`${getLectureName(j)} (${j})`}
                                                    </Link>
                                                </span>
                                            ))}
                                        </td>
                                    </tr>
                                );
                            })}
                        </tbody>
                    </table>
                </div>

                <div className="table-responsive">
                    <table className="table table-hover table-bordered">
                        <thead>
                            <tr>
                                <th scope="col">Zeit</th>
                                <th scope="col">Pfad</th>
                                <th scope="col">Course</th>
                                <th scope="col">Lecture</th>
                                <th scope="col">Lecture id</th>
                            </tr>
                        </thead>
                        <tbody>
                            {sortlog?.log.map((i, ind) => {
                                const formattedDate = DateTime.fromISO(i.time).toFormat(
                                    "yyyy-MM-dd HH:mm:ss",
                                );
                                return (
                                    <tr key={i.lecture_id + "#" + ind}>
                                        <td>{formattedDate}</td>
                                        <td>
                                            <a href={`${api.assetUrl()}/${i.file_path}`}>
                                                {i.file_path}
                                            </a>
                                        </td>
                                        <td>
                                            <Link
                                                href={`/${i.course_id}`}
                                            >{`${i.course_full_name} (${i.course_id})`}</Link>
                                        </td>
                                        <td>
                                            <Link href={`/${i.course_id}#lecture-${i.lecture_id}`}>
                                                {i.lecture_title}
                                            </Link>
                                        </td>
                                        <td>{i.lecture_id}</td>
                                    </tr>
                                );
                            })}
                        </tbody>
                    </table>
                </div>
            </div>
        </div>
    );
}

export default function SortLog() {
    return (
        <ModeratorBarrier>
            <SortLogImpl />
        </ModeratorBarrier>
    );
}
