import { useBackendContext } from "@/components/BackendProvider";
import { useLanguage } from "@/components/LanguageProvider";
import ModeratorBarrier from "@/components/ModeratorBarrier";
import { ReloadBoundary, useReloadBoundary } from "@/components/ReloadBoundary";
import { useEffect, useState } from "react";
import { showError } from "@/misc/ErrorHandlers";

import type React from "react";
import type { GetMySettingsResponse, settings_entry, settings_section } from "@/api/api_v1_types";

function BoolEntry({ entry, disabled }: { entry: settings_entry; disabled: boolean }) {
    const reloadFunc = useReloadBoundary();
    const api = useBackendContext();
    const [isChanging, setIsChanging] = useState(false);
    const { language } = useLanguage();

    const onChange = (e: React.ChangeEvent<HTMLInputElement>) => {
        setIsChanging(true);
        api.updateMySettings({ updates: { [entry.id]: e.target.checked } })
            .then(reloadFunc)
            .catch((e) => {
                showError(e, "Exception while updating settings");
            })
            .then(() => setIsChanging(false));
    };

    return (
        <div className="d-flex align-items-center">
            {language.get(`setting.entry.${entry.id}`)}
            <input
                type="checkbox"
                className="ms-2 form-check-input"
                checked={entry.value}
                onChange={onChange}
                disabled={isChanging || disabled}
            />
        </div>
    );
}

function SettingsEntry({ entry, disabled }: { entry: settings_entry; disabled: boolean }) {
    const { language } = useLanguage();

    if (entry.type === "boolean") {
        return <BoolEntry entry={entry} disabled={disabled} />;
    }

    if (entry.type === "subsection") {
        return (
            <div>
                <h6>{language.getStyled(`setting.entry.${entry.id}`, true)}</h6>
                <ul>
                    {entry.entries?.map((setting, ind) => (
                        <li key={ind}>
                            <SettingsEntry entry={setting} disabled={disabled} />
                        </li>
                    ))}
                </ul>
            </div>
        );
    }
    throw new Error(`unknown entry type ${entry.type}`);
}

function SettingsSection({ section, disabled }: { section: settings_section; disabled: boolean }) {
    const { language } = useLanguage();

    return (
        <div>
            <h5>{language.get(`setting.section.${section.name}`)}</h5>
            <ul className="list-unstyled">
                {section.entries.map((setting, ind) => (
                    <li key={ind}>
                        <SettingsEntry entry={setting} disabled={disabled} />
                    </li>
                ))}
            </ul>
        </div>
    );
}

function UserImpl() {
    const api = useBackendContext();
    const [settings, setSettings] = useState<GetMySettingsResponse>();
    const [reload, setReload] = useState(false);
    const [isLoading, setIsLoading] = useState(true);

    useEffect(() => {
        setIsLoading(true);
        api.getMySettings()
            .then(
                (settings) => {
                    setSettings(settings);
                },
                (error) => {
                    showError(error, "Unable to load settings");
                    setSettings(undefined);
                },
            )
            .then(() => setIsLoading(false));
    }, [api, reload]);

    return (
        <div>
            <h1>Settings</h1>
            <ReloadBoundary reloadFunc={() => setReload(!reload)}>
                {settings &&
                    settings.sections.map((section, ind) => (
                        <SettingsSection key={ind} section={section} disabled={isLoading} />
                    ))}
                {(!settings || isLoading) && <div className="spinner-border" role="status" />}
            </ReloadBoundary>
        </div>
    );
}

export default function User() {
    return (
        <ModeratorBarrier>
            <UserImpl />
        </ModeratorBarrier>
    );
}
