import { useBackendContext } from "@/components/BackendProvider";
import { useSearchParams } from "next/navigation";
import { useEffect, useRef, useState } from "react";
import { CourseItem } from "./courses";
import { LectureListItem } from "@/components/CourseListing";
import { ReloadBoundary } from "@/components/ReloadBoundary";
import { useDebounce } from "@/misc/PromiseHelpers";
import { useRouter } from "next/router";
import { ErrorPage } from "@/misc/ErrorHandlers";

import type { SearchResponse, course, lecture } from "@/api/api_v1_types";
import type React from "react";
import { useLanguage } from "@/components/LanguageProvider";

function CourseResults({ courses }: { courses: course[] }) {
    return (
        <div className="mb-3">
            <h3>Kurse</h3>
            <div className="courses-list">
                {courses.map((course, ind) => (
                    <CourseItem key={course.id} course={course} tabIndex={ind + 1} />
                ))}
            </div>
        </div>
    );
}

function LectureResults({
    lectures,
    tabIndexBegin,
}: {
    lectures: lecture[];
    tabIndexBegin: number;
}) {
    return (
        <>
            <h3>Vorlesungen</h3>
            <div className="list-group lectureslist">
                {lectures.map((lecture, ind) => (
                    <LectureListItem
                        key={lecture.id}
                        lecture={lecture}
                        tabIndex={tabIndexBegin + ind}
                    />
                ))}
            </div>
        </>
    );
}

export default function Search() {
    const api = useBackendContext();
    const params = useSearchParams();
    const [query, setQuery] = useState(params.get("q"));
    const [searchResults, setSearchResults] = useState<SearchResponse>();
    const workingQuery = useRef(query);
    const [_, doForceReRender] = useState(0);
    const [error, setError] = useState<any>();
    const router = useRouter();
    const { language } = useLanguage();

    // TODO: this code needs more cleanup, variable names are not very descriptive

    useEffect(() => {
        if (params.has("q")) setQuery(params.get("q"));
        else if (query !== undefined && query !== null && query.length > 0) setQuery("");
        // eslint-disable-next-line react-hooks/exhaustive-deps
    }, [params]);

    const updateSearchResults = () => {
        setSearchResults(undefined);
        if (query === undefined || query === null || query === "") {
            if (params.has("q")) router.push({ search: new URLSearchParams().toString() });
            return;
        }
        if (query !== params.get("q")) {
            const newParams = new URLSearchParams(params.toString());
            newParams.set("q", query);

            router.push({ search: newParams.toString() });
        }

        workingQuery.current = query;
        api.search(query)
            .then((results) => {
                if (query !== workingQuery.current) return; // if the query has changed, don't update the results
                setSearchResults(results);
                setError(undefined);
            })
            .catch((e) => {
                setError(e);
            });
    };

    // eslint-disable-next-line react-hooks/exhaustive-deps
    useEffect(updateSearchResults, [query, api]);

    const [updateQueryDeferred, updateQueryNow] = useDebounce(() => {
        setQuery(workingQuery.current);
    }, 500);

    let results = (
        <div className="spinner-border" role="status">
            <span className="visually-hidden">Loading...</span>
        </div>
    );
    if (error) {
        results = <ErrorPage error={error} objectName="Suchergebnisse" objectPlural={true} />;
    } else if (
        query !== workingQuery.current &&
        query !== undefined &&
        query !== null &&
        query !== ""
    ) {
        results = (
            <div className="spinner-border text-secondary" role="status">
                <span className="visually-hidden">Loading...</span>
            </div>
        );
    } else if (searchResults) {
        results = (
            <>
                {searchResults.courses.length > 0 && (
                    <CourseResults
                        courses={searchResults.courses.map(
                            (courseId) => searchResults.courses_context[courseId],
                        )}
                    />
                )}
                {searchResults.lectures.length > 0 && (
                    <LectureResults
                        lectures={searchResults.lectures}
                        tabIndexBegin={searchResults.courses.length + 1}
                    />
                )}
                {searchResults.courses.length === 0 && searchResults.lectures.length === 0 && (
                    <h4 className="text-muted">{language.get("ui.search.no_results")}</h4>
                )}
            </>
        );
    } else if (query === undefined || query === null || query === "") {
        results = <h4 className="text-muted">{language.get("ui.search.no_query")}</h4>;
    }

    const onChange = (e: React.ChangeEvent<HTMLInputElement>) => {
        workingQuery.current = e.target.value;
        updateQueryDeferred();
        doForceReRender((r) => r + 1);
    };

    return (
        <ReloadBoundary reloadFunc={updateSearchResults}>
            <div>
                <h1 className="d-flex align-items-center w-100">
                    <span className="me-2">{language.get("ui.generic.search")}:</span>
                    <form
                        className="d-inline flex-fill"
                        onSubmit={(e) => {
                            e.preventDefault();
                            updateQueryNow();
                        }}
                    >
                        <input
                            className="sneak-100-input fst-italic text-muted w-100 border-bottom"
                            type="text"
                            value={workingQuery.current || ""}
                            onChange={onChange}
                            autoFocus
                            tabIndex={0}
                        />
                    </form>
                </h1>
                {results}
            </div>
        </ReloadBoundary>
    );
}
